
#ifdef MODULE
#ifndef __KERNEL__
#define __KERNEL__
#endif
#endif

#ifndef MODULE
#include "config.h"
#include "ltdl.h"
#include <stdio.h>
#include <string.h>
#include <assert.h>
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdlib.h>
#ifdef HAVE_WINDOWS_H
#include <winsock.h>
#define __ANSI_CPP__
#else
#include <netinet/in.h>
#include <arpa/inet.h>
#endif
#include <io.h>
#include <atl.h>
#include "cm.h"
#include "gen_thread.h"
#include "cm_internal.h"
#include "cm_transport.h"
#else
#include <asm/uaccess.h>
#include <linux/in.h>
#include "config.h"
#include "../include/io.h"
#include "../include/atl.h"
#include "../include/utils.h"
#include "cm.h"
#include "cm_internal.h"
#include "cm_transport.h"
#include "kcm.h"
#include "library.h"
#include "assert.h"

#define printf         printk
#define perror         printk
#define fprintf(fmt, args...) printk (args) 
#define malloc         (void *)DAllocMM
#define free(a)        DFreeMM((addrs_t)a)
#define realloc(a,b)   (void *)DReallocMM((addrs_t)a,b)
#define exit           sys_exit
#define qsort          _quicksort
#define getpid()       current->pid
#define signal(a,b)    
#define gethostname    sys_gethostname
#define gethostbyname  lookup_name
#define select         sys_select
#define thr_mutex_free(a)

#define MAX_THREADS 1000

int resubmit_thread();

extern int cm_identifier;
extern kthread_t *resubmit_threads[MAX_THREADS];
extern int terminate_threads;

static char *
inet_ntoa(struct in_addr ina)
{
    static char buf[4*sizeof "123"];
    unsigned char *ucp = (unsigned char *)&ina;

    sprintf(buf, "%d.%d.%d.%d",
	ucp[0] & 0xff,
	ucp[1] & 0xff,
	ucp[2] & 0xff,
	ucp[3] & 0xff);
    return buf;
}

typedef struct func_list_item {
    select_list_func func;
    void *arg1;
    void *arg2;
} FunctionListElement;

typedef struct select_data {
    thr_thread_t server_thread;
#ifndef MODULE
    void *fdset;
#endif
    int sel_item_max;
    FunctionListElement *select_items;
    periodic_task_handle periodic_task_list;

    int closed;
    int select_consistency_number;
    int wake_read_fd;
    int wake_write_fd;
    char *hostname;
    int listen_port;
#ifdef MODULE
    int loop;
    int event;
    int listen_fd;
    char fdset[FD_SETSIZE];
    CManager cm;
#endif
} *select_data_ptr;

#ifdef MODULE
void resubmit_timer_func(unsigned long data);
#endif

extern
int
CMcontrol_list_poll(CMControlList cl);

extern
int
CMcontrol_list_wait(CMControlList cl);

static void CMControlList_close ARGS((CMControlList cl));
static int CMdo_non_CM_handler ARGS((CMConnection conn, int header));

extern int load_transport(CManager cm, const char *trans_name);

#ifdef MODULE
extern void libcmselect_LTX_add_select(CManager cm, CMtrans_services svc, 
	select_data_ptr *sdp, int fd, select_list_func func, 
	void *arg1,
	void *arg2);  
#else
extern void libcmselect_LTX_add_select(CMtrans_services svc, 
	select_data_ptr *sdp, int fd, select_list_func func, 
	void *arg1,
	void *arg2);  
#endif
#ifdef MODULE
extern void libcmselect_LTX_remove_select(CManager cm, CMtrans_services svc,
	select_data_ptr *sdp, int fd);
#else
extern void libcmselect_LTX_remove_select(CMtrans_services svc,
	select_data_ptr *sdp, int fd);
#endif
extern void libcmselect_LTX_blocking_function(CMtrans_services svc,
	void *client_data);
extern void libcmselect_LTX_polling_function(CMtrans_services svc,
	void *client_data);
#ifdef MODULE
extern periodic_task_handle libcmselect_LTX_add_periodic(CManager cm,
	CMtrans_services svc,
	select_data_ptr *sdp, int interval_sec, int interval_usec,
	select_list_func func, void *arg1, void *arg2);
#else
extern periodic_task_handle libcmselect_LTX_add_periodic(CMtrans_services svc,
	select_data_ptr *sdp, int interval_sec, int interval_usec,
	select_list_func func, void *arg1, void *arg2);
#endif
#ifdef MODULE
extern periodic_task_handle libcmselect_LTX_add_delayed_task(
	CManager cm,
	CMtrans_services svc,
	select_data_ptr *sdp, int delay_sec, int delay_usec,
	select_list_func func, void *arg1, void *arg2);
#else
extern periodic_task_handle libcmselect_LTX_add_delayed_task(
	CMtrans_services svc,
	select_data_ptr *sdp, int delay_sec, int delay_usec,
	select_list_func func, void *arg1, void *arg2);
#endif
extern void libcmselect_LTX_remove_periodic(CMtrans_services svc,
	select_data_ptr *sdp, periodic_task_handle handle);
#ifdef MODULE
extern void libcmselect_LTX_select_initialize(CMtrans_services svc,
	CManager cm, void *client_data);
#else
extern void libcmselect_LTX_select_initialize(CMtrans_services svc,
	void *client_data);
#endif

static DECLARE_WAIT_QUEUE_HEAD(sleep_queue);
struct timer_list timer;

static void 
wake_up_poll(data)
unsigned long data;
{
    wake_up_interruptible (&sleep_queue);
}

#endif

static void CMinitialize ARGS((CManager cm));

struct CMtrans_services CMstatic_trans_svcs = {CMmalloc, CMrealloc, CMfree, 
					       CM_fd_add_select, 
					       CM_fd_remove_select, 
					       CMtransport_trace,
					       CMConnection_create};
static void CMControlList_close ARGS((CMControlList cl));
static int CMdo_non_CM_handler ARGS((CMConnection conn, int header));

static void
CMpoll_forever(cm)
CManager cm;
{
    CMControlList cl = cm->control_list;
#ifdef MODULE
    DECLARE_WAITQUEUE(mywait, current);

    init_timer(&timer);
#endif
    while(!cl->closed) { 
#ifndef MODULE
	CMtrace_out(cm, CMLowLevelVerbose, "CM Poll Forever - thread %lx doing wait", (long)thr_thread_self());
	if (CMcontrol_list_wait(cl) == -1) {
	    exit(1);
	}
	thr_thread_yield();
#else
      timer.expires = jiffies + PERIOD;
      timer.data = 0;
      timer.function = wake_up_poll;

      add_timer (&timer);
      add_wait_queue(&sleep_queue,&mywait);
      current->state = TASK_INTERRUPTIBLE;
      schedule();
      current->state = TASK_RUNNING;
      remove_wait_queue(&sleep_queue,&mywait);
#endif
    }
    CMtrace_out(cm, CMLowLevelVerbose, "CM Poll Forever - doing close");
    CManager_close(cm);
}

extern void
CMrun_network(cm)
CManager cm;
{
    CMpoll_forever(cm);
}

#ifndef MODULE
static int
CM_test_thread_func()
{
    return 1;
}
#endif

int
CMfork_comm_thread(cm)
CManager cm;
{
    /* if we're on a kernel-level-threads package, for the thread and 
       return 1, else return 0; */
#ifndef MODULE
    if (gen_thr_is_kernel()) {
	CMControlList_lock(cm->control_list);
	if (cm->control_list->has_thread == 0) {
	    if (cm->control_list->network_blocking_function.func) {
		thr_thread_t server_thread = 
		    thr_fork((void_arg_func)CMpoll_forever, 
			     (void*)cm);
		if (server_thread == NULL) {
		    return 0;
		}
		cm->control_list->server_thread = server_thread;
		cm->control_list->has_thread = 1;
		cm->reference_count++;
	    } else {
		/*
		 *  Can't start a server thread yet, but lets see 
		 *  if we can fork anything successfully.
		 */
		thr_thread_t test_thread = 
		    thr_fork((void_arg_func)CM_test_thread_func, 
			     (void*)cm);
		if (test_thread == NULL) {
		    /* No.  Say we can't. */
		    return 0;
		}
		/* OK, we'll fork it later. */
		cm->control_list->has_thread = -1; /* should fork one */
	    }
	}
	CMControlList_unlock(cm->control_list);
	return 1;
    }
#endif
    return 0;
}

extern
void
CMControlList_set_blocking_func(cl, cm, bfunc, pfunc, client_data)
CMControlList cl;
CManager cm;
CMPollFunc bfunc;
CMPollFunc pfunc;
void *client_data;
{
#ifndef MODULE
    assert(CMControlList_locked(cl));
#endif
    assert(cl->network_blocking_function.func == NULL);
    cl->network_blocking_function.func = bfunc;
    cl->network_blocking_function.client_data = client_data;
    cl->network_blocking_function.cm = NULL;
    cl->network_polling_function.func = pfunc;
    cl->network_polling_function.client_data = client_data;
    cl->network_polling_function.cm = NULL;
    if (cl->has_thread == -1) {
#ifndef MODULE
	if (thr_fork((void_arg_func)CMpoll_forever, (void*)cm) == NULL) {
	    return;
	}
#endif
	cl->has_thread = 1;
	cm->reference_count++;
    }
}

#ifdef MODULE
extern
int
CMcontrol_list_poll_network(cl)
CMControlList cl;
{
    cl->network_polling_function.func((void*)&CMstatic_trans_svcs,
				      cl->network_polling_function.client_data);
    return CMcontrol_list_poll(cl);
}
#else
extern
void
CMcontrol_list_poll_network(cl)
CMControlList cl;
{
    cl->network_polling_function.func((void*)&CMstatic_trans_svcs,
				      cl->network_polling_function.client_data);
    CMcontrol_list_poll(cl);
}
#endif

#ifdef MODULE
extern int 
CMpoll_network(cm)
CManager cm;
{
    CMtrace_out(cm, CMLowLevelVerbose, "CM Poll Network");
    return CMcontrol_list_poll_network(cm->control_list);
}
#else
extern void
CMpoll_network(cm)
CManager cm;
{
    CMtrace_out(cm, CMLowLevelVerbose, "CM Poll Network");
    CMcontrol_list_poll_network(cm->control_list);
}
#endif

static void
add_contact_list(cm, attrs)
CManager cm;
attr_list attrs;
{
    int list_size = 0;
    if (cm->contact_lists == NULL) {
	cm->contact_lists = CMmalloc(sizeof(attr_list) *2);
	list_size = 0;
    } else {
	while(cm->contact_lists[list_size] != NULL) list_size++;
	cm->contact_lists = CMrealloc(cm->contact_lists, 
				      sizeof(attr_list) * (list_size + 2));
    }
    cm->contact_lists[list_size] = attrs;
    cm->contact_lists[list_size+1] = NULL;
}

attr_list
CMget_contact_list(cm)
CManager cm;
{
    if (cm->contact_lists == NULL) return NULL;
    add_ref_attr_list(cm->contact_lists[0]);
    return (cm->contact_lists[0]);
}

int
CMlisten(cm)
     CManager cm;
{
  return CMlisten_specific (cm, NULL);
}

extern int
CMinternal_listen(cm, listen_info)
CManager cm;
attr_list listen_info;
{
    int success = 0;
    transport_entry *trans_list;

    trans_list = cm->transports;
    while ((trans_list != NULL) && (*trans_list != NULL)) {
	attr_list attrs;
	attrs = (*trans_list)->listen(cm, &CMstatic_trans_svcs,
				      *trans_list,
                                      listen_info);
	add_contact_list(cm, attrs);
	if (attrs != NULL) {
	    success++;
	}
	trans_list++;
    }
    return success;
}

int
CMlisten_specific(cm, listen_info)
CManager cm;
attr_list listen_info;
{
    int success = 0;
    if (!cm->initialized) CMinitialize(cm);
    CManager_lock(cm);
    success = CMinternal_listen(cm, listen_info);
    CManager_unlock(cm);
    return (success != 0);
}

#ifdef MODULE
#ifdef RUDP
static char *CMglobal_default_transport = "rudp";
#else
static char *CMglobal_default_transport = "sockets";
#endif
#else
#ifndef DONT_USE_SOCKETS
static char *CMglobal_default_transport = "sockets";
#else 
static char *CMglobal_default_transport = "rudp";
#endif
#endif

static char **CMglobal_alternate_transports = NULL;

static void 
CMinitialize(cm)
CManager cm;
{
    char **transport_names = CMglobal_alternate_transports;
    CMglobal_data_lock();
    if (CMglobal_default_transport) {
	if (load_transport(cm, CMglobal_default_transport) == 0) {
	    fprintf(stderr, "Failed to initialize default transport.  Exiting.\n");
	    exit(1);
	}
    }
    while ((transport_names != NULL) && (transport_names[0] != NULL)) {
	load_transport(cm, transport_names[0]);
	transport_names++;
    }
    cm->initialized++;
    CMglobal_data_unlock();
}

#ifdef MODULE
extern void
CMevent_delivered(cm)
CManager cm;
{
    cm->event_delivered = 1;
}

extern void
CMdeliver_event(cm, func, cm_event, cm_event_len, client_data, event_attrs)
CManager cm;
void *func;
void *cm_event;
int cm_event_len;
void *client_data;
attr_list event_attrs;
{
  cm->func = func;
  cm->event = cm_event;
  cm->event_len = cm_event_len;
  cm->client_data = client_data;
  cm->event_attrs = event_attrs;
}

extern int 
CMretrieve_event(cm)
CManager cm;
{
    CMControlList cl = cm->control_list;
    func_entry *poll_list = cl->polling_function_list;

    CMControlList_lock(cl);
    while ((poll_list != NULL) && (poll_list->func != NULL)){
	int consistency_number = cl->cl_consistency_number;

	CMControlList_unlock(cl);
	poll_list->cm->event_delivered = 0;
	poll_list->func(poll_list->cm, poll_list->client_data);
	if (poll_list->cm->event_delivered == 1) return 0;
	/* do function */
	CMControlList_lock(cl);
	if (consistency_number != cl->cl_consistency_number) {
	    CMControlList_unlock(cl);
	    return 1;
	}
	poll_list++;
    }
    CMControlList_unlock(cl);
    return 1;
}
#endif

extern
int
CMcontrol_list_poll(cl)
CMControlList cl;
{
    func_entry *poll_list = cl->polling_function_list;
#ifdef MODULE
    if (cl->tsk != NULL) return 2;
    if (current->kthread != 1) return 2;
#endif
    CMControlList_lock(cl);
    while ((poll_list != NULL) && (poll_list->func != NULL)){
	int consistency_number = cl->cl_consistency_number;

	CMControlList_unlock(cl);
	poll_list->func(poll_list->cm, poll_list->client_data);
	/* do function */
	CMControlList_lock(cl);
	if (consistency_number != cl->cl_consistency_number) {
	    CMControlList_unlock(cl);
	    return 1;
	}
	poll_list++;
    }
    CMControlList_unlock(cl);
    return 1;
}

static 
void
CMControlList_add_poll(cl, cm, func, client_data)
CMControlList cl;
CManager cm;
CMPollFunc func;
void *client_data;
{
    func_entry *poll_list;
    int count = 0;
    CMControlList_lock(cl);
    poll_list = cl->polling_function_list;
    while ((poll_list != NULL) && (poll_list->func != NULL)) {
	count++;
    }
    if (poll_list != NULL) {
	poll_list = CMrealloc(poll_list, sizeof(func_entry) * (count+2));
    } else {
	poll_list = CMmalloc(sizeof(func_entry)*2);
    }
    poll_list[count].cm = cm;
    poll_list[count].func = func;
    poll_list[count].client_data = client_data;
    poll_list[count+1].func = NULL;
    cl->polling_function_list = poll_list;
    CMControlList_unlock(cl);
}
    
extern
void
CMadd_poll(cm, func, client_data)
CManager cm;
CMPollFunc func;
void *client_data;
{
    CMControlList_add_poll(cm->control_list, cm, func, client_data);
}

struct _CMTaskHandle {
    CManager cm;
    periodic_task_handle task;
};

extern
int
CMcontrol_list_wait(cl)
CMControlList cl;
{
#ifndef MODULE
    if ((cl->server_thread != 0) &&
	(cl->server_thread != thr_thread_self())) {
	/* What?  We're polling, but we're not the server thread? */
	fprintf(stderr, "Warning:  Multiple threads calling CMnetwork_wait\n");
	fprintf(stderr, "          This situation may result in unexpected I/O blocking.\n");
	fprintf(stderr, "          Server thread set to %lx.\n", (long) thr_thread_self());
    }
    cl->server_thread = thr_thread_self();
#endif
    if (cl->network_blocking_function.func != NULL) {
	cl->network_blocking_function.func((void*)&CMstatic_trans_svcs,
					   cl->network_blocking_function.client_data);
    }
    CMcontrol_list_poll(cl);
    return 1;
}
extern CMControlList CMControlList_create();

extern
CManager
CManager_create()
{
    CManager cm = (CManager) CMmalloc(sizeof(CManager_s));

    if (cm == NULL)
	return NULL;

    /* initialize data structs */
    cm->transports = NULL;
    cm->initialized = 0;
    cm->reference_count = 1;

    cm->control_list = CMControlList_create();
#ifdef MODULE
    sema_init(&cm->exchange_lock, 1);
#else
    cm->exchange_lock = thr_mutex_alloc();
#endif
    cm->locked = 0;
    cm->closed = 0;
    CMinit_local_formats(cm);
#ifdef MODULE
    sema_init(&cm->context_lock, 1);
#else
    cm->context_lock = thr_mutex_alloc();
#endif

    cm->in_format_count = 0;
    cm->in_formats = CMmalloc(1);

    cm->reg_format_count = 0;
    cm->reg_formats = CMmalloc(1);

    cm->pending_request_max = 1;
    cm->pbio_requests = CMmalloc(sizeof(struct _pending_format_requests));
    cm->pbio_requests[0].server_id = NULL;
    cm->pbio_requests[0].id_length = 0;
    cm->pbio_requests[0].condition = 0;
    cm->pbio_requests[0].top_request = 0;

    cm->connection_count = 0;
    cm->connections = CMmalloc(1);
    cm->reg_user_format_count = 0;
    cm->reg_user_formats = CMmalloc(1);
    cm->taken_buffer_list = NULL;
    cm->cm_buffer_list = NULL;

    cm->contact_lists = NULL;
#ifdef MODULE
    if (current->qfabric == NULL)
      initialize_qfabric();
    init_timer(&cm->cm_timer);
#endif
    return cm;
}

static void
CManager_free(cm)
CManager cm;
{
    int i;

    if (cm->transports) CMfree(cm->transports);
    for (i=0 ; i < cm->reg_format_count; i++) {
	free_IOsubcontext(cm->reg_formats[i]->IOsubcontext);
	CMfree(cm->reg_formats[i]);
    }
    CMfree(cm->reg_formats);

    CMfree(cm->reg_user_formats);
    thr_mutex_free(cm->exchange_lock);
    free_IOcontext(cm->IOcontext);

    i = 0;
    if (cm->contact_lists != NULL) {
	while(cm->contact_lists[i] != NULL) {
	    free_attr_list(cm->contact_lists[i]);
	    i++;
	}
    }
    CMfree(cm);
}

extern void
CManager_close(cm)
CManager cm;
{
    CMControlList_close(cm->control_list);
    cm->reference_count--;
    if (cm->reference_count == 0) {
	CManager_free(cm);
    }
}

static void
add_conn_to_CM(CManager cm, CMConnection conn)
{
    CManager_lock(cm);
    cm->connections = 
	CMrealloc(cm->connections, 
		  (cm->connection_count + 1) * sizeof(cm->connections[0]));
    cm->connections[cm->connection_count] = conn;
    cm->connection_count++;
    CManager_unlock(cm);
}

static void
remove_conn_from_CM(CManager cm, CMConnection conn)
{
    int i;
    int found = 0;
    CManager_lock(cm);
    for (i=0; i < cm->connection_count; i++) {
	if (cm->connections[i] == conn) {
	    found++;
	} else if (found) {
	    /* copy down */
	    cm->connections[i-1] = cm->connections[i];
	}
    }
    if (found == 0) {
	fprintf(stderr, "Internal error, remove_conn_from_CM.  Not found\n");
    } else {
	cm->connection_count--;
    }
    CManager_unlock(cm);
}

extern CMControlList
CMControlList_create()
{
    CMControlList new_list = (CMControlList) CMmalloc(sizeof(CMControlList_s));
    new_list->select_initialized = 0;
    new_list->select_data = NULL;
    new_list->add_select = NULL;
    new_list->remove_select = NULL;
    new_list->server_thread = NULL;
    new_list->network_blocking_function.func = NULL;
    new_list->network_polling_function.func = NULL;
    new_list->polling_function_list = NULL;
    new_list->cl_consistency_number = 1;
    new_list->reference_count = 1;
    new_list->free_reference_count = 1;
#ifdef MODULE
    new_list->tsk = NULL;
    sema_init(&new_list->list_mutex, 1);
#else
    new_list->list_mutex = thr_mutex_alloc();
#endif
    new_list->condition_list = NULL;
    new_list->next_condition_num = 1;
    new_list->closed = 0;
    new_list->locked = 0;
    new_list->has_thread = 0;
    return new_list;
}

CMConnection
CMConnection_create(trans, transport_data, conn_attrs)
transport_entry trans;
void *transport_data;
attr_list conn_attrs;
{
    CMConnection conn = CMmalloc(sizeof(struct _CMConnection));
    conn->cm = trans->cm;
    conn->trans = trans;
    conn->transport_data = transport_data;
    conn->ref_count = 1;
#ifdef MODULE
    sema_init(&conn->write_lock, 1);
    sema_init(&conn->read_lock, 1);
#else
    conn->write_lock = thr_mutex_alloc();
    conn->read_lock = thr_mutex_alloc();
#endif
    conn->closed = 0;
    conn->downloaded_formats = NULL;
    conn->IOsubcontext = create_IOsubcontext(trans->cm->IOcontext);
    conn->foreign_data_handler = NULL;
    conn->io_out_buffer = create_IOBuffer();
    conn->close_handler = NULL;
    conn->attrs = conn_attrs;
    add_conn_to_CM(trans->cm, conn);
//    init_timer(&conn->resubmit_timer);
#ifdef MODULE
    conn->send_thread.used = -1;
    conn->thread_used = -1;
#endif
    return conn;
}

extern attr_list
CMConnection_get_attrs(conn)
CMConnection conn;
{
    return conn->attrs;
}

void
CMConnection_close(conn)
CMConnection conn;
{
    CManager_lock(conn->cm);
    conn->ref_count--;
    if (conn->ref_count > 0) {
	CMtrace_out(conn->cm, CMConnectionVerbose, "CM - Dereference connection %lx\n",
		    (void*)conn);
	CManager_unlock(conn->cm);
	return;
    }
    CManager_unlock(conn->cm);
    CMconn_fail_conditions(conn);
    CMtrace_out(conn->cm, CMConnectionVerbose, "CM - Shut down connection %lx\n",
		(void*)conn);
    remove_conn_from_CM(conn->cm, conn);
    conn->trans->shutdown_conn(&CMstatic_trans_svcs, conn->transport_data);
    thr_mutex_free(conn->write_lock);
    thr_mutex_free(conn->read_lock);
    free_IOsubcontext(conn->IOsubcontext);
    conn->closed = 1;
    CMfree(conn->downloaded_formats);
    conn->foreign_data_handler = NULL;
    free_IOBuffer(conn->io_out_buffer);
    if (conn->close_handler) {
	CMtrace_out(conn->cm, CMConnectionVerbose, 
		    "CM - Calling close handler %lx for connection %lx\n",
		    (void*) conn->close_handler, (void*)conn);
	conn->close_handler(conn->cm, conn, conn->close_client_data);
    }
}

void
CMconn_register_close_handler(conn, func, client_data)
CMConnection conn;
CMCloseHandlerFunc func;
void *client_data;
{
    conn->close_handler = func;
    conn->close_client_data = client_data;
}

static void
CMControlList_close(cl)
CMControlList cl;
{
    cl->reference_count--;
    cl->closed = 1;
    if (cl->reference_count == 0) {
/*	shutdown_wake_mechanism(cl);*/
	thr_mutex_free(cl->list_mutex);
	cl->closed = 1;
    }
}

extern void
CMControlList_free(cl)
CMControlList cl;
{
    cl->free_reference_count--;
    if(cl->free_reference_count == 0) {
	CMfree(cl);
    }
}

#ifndef GETDOMAINNAME_DEFINED
extern int getdomainname ARGS((char*name, int namelen));
#endif

extern void
CMget_qual_hostname(char *buf, int len)
{
    struct hostent *host;
#ifdef MODULE
    mm_segment_t fs;
    struct hostent *host_addr;
#endif

    char *network_string = getenv("CM_NETWORK");
#ifdef MODULE
    fs = get_fs();
    set_fs(get_ds());
    sys_gethostname(buf, len);
    set_fs(fs);
#else
    gethostname(buf, len);
#endif
    buf[len - 1] = '\0';
    if (memchr(buf, '.', strlen(buf)) == NULL) {
	/* no dots, probably not fully qualified */
	int end = strlen(buf);
	buf[end] = '.';
#ifdef HAVE_GETDOMAINNAME
#ifdef MODULE
    	fs = get_fs();
    	set_fs(get_ds());
	sys_getdomainname((&buf[end]) + 1, len - strlen(buf));
    	set_fs(fs);
#else
	getdomainname((&buf[end]) + 1, len - strlen(buf));
#endif
	if (buf[end+1] == 0) {
	    char *tmp_name;
	    buf[end] = 0;
	    /* getdomainname was useless, hope that gethostbyname helps */

#ifdef MODULE
	    host_addr = gethostbyname(buf);
	    if (host_addr == NULL) return;
	    tmp_name = host_addr->h_name;
	    strncpy(buf, tmp_name, len);
	    DFreeMM((addrs_t)tmp_name);
	    DFreeMM((addrs_t)host_addr->h_addr_list);
	    DFreeMM((addrs_t)host_addr);
#else
	    tmp_name = (gethostbyname(buf))->h_name;
	    strncpy(buf, tmp_name, len);
#endif
	}
#else
	{
	    /* no getdomainname, hope that gethostbyname will help */
#ifdef MODULE
	    char *tmp_name;
	    host_addr = gethostbyname(buf);
	    if (host_addr == NULL) return;
	    tmp_name = host_addr->h_name;
	    strncpy(buf, tmp_name, len);
	    DFreeMM(tmp_name);
	    DFreeMM(host_addr->h_addr_list);
	    DFreeMM(host_addr);
#else
	    char *tmp_name = (gethostbyname(buf))->h_name;
	    strncpy(buf, tmp_name, len);
#endif
	}
#endif
	buf[len - 1] = '\0';
    }
    if (((host = gethostbyname(buf)) == NULL) ||
	(memchr(buf, '.', strlen(buf)) == NULL)){
	/* bloody hell, what do you have to do? */
#ifndef MODULE
	gethostname(buf, len);
	host  = gethostbyname(buf);
	host = gethostbyaddr(host->h_addr_list[0], host->h_length, host->h_addrtype);
	if (host != NULL) {
	  strncpy(buf, host->h_name, len);
	}
#endif
    }
#ifdef MODULE
    DFreeMM((addrs_t)host->h_name);
    DFreeMM((addrs_t)host->h_addr_list);
    DFreeMM((addrs_t)host);
#endif
    
    if (network_string != NULL) {
	int name_len = strlen(buf) + 2 + strlen(network_string);
	char *new_name_str = CMmalloc(name_len);
	char *first_dot = strchr(buf, '.');

	/* stick the CM_NETWORK value in there */
	memset(new_name_str, 0, name_len);
	*first_dot = 0;
	first_dot++;
	sprintf(new_name_str, "%s-%s.%s", buf, network_string, first_dot);
#ifdef MODULE
	host_addr = gethostbyname(new_name_str);
	if (host_addr != NULL) {
	    strcpy(buf, new_name_str);
	    DFreeMM((addrs_t)host_addr->h_name);
	    DFreeMM((addrs_t)host_addr->h_addr_list);
	    DFreeMM((addrs_t)host_addr);
	}
#else
	if (gethostbyname(new_name_str) != NULL) {
	    /* host has no NETWORK interface */
	    strcpy(buf, new_name_str);
	}
#endif
	CMfree(new_name_str);
    }

    if (((host = gethostbyname(buf)) == NULL) ||
	(memchr(buf, '.', strlen(buf)) == NULL)){
	/* just use the bloody IP address */
	gethostname(buf, len);
	host  = gethostbyname(buf);
	if (host != NULL) {
	    char *tmp = inet_ntoa(*(struct in_addr*)(host->h_addr_list[0]));
	    strncpy(buf, tmp, len);
#ifdef MODULE
	    DFreeMM((addrs_t)host->h_name);
	    DFreeMM((addrs_t)host->h_addr_list);
	    DFreeMM((addrs_t)host);
#endif
	} else {
	    static int warn_once = 0;
	    if (warn_once == 0) {
		warn_once++;
		printf("Attempts to establish your fully qualified hostname, or indeed any\nuseful network name, have failed horribly.  Sorry.\n");
	    }
	    strncpy(buf, "Unknown", len);
	}
    }
}

CMConnection
CMinternal_initiate_conn(cm, attrs)
CManager cm;
attr_list attrs;
{
    transport_entry *trans_list;
    trans_list = cm->transports;
#ifndef MODULE
    assert(CManager_locked(cm));
#endif
    while ((trans_list != NULL) && (*trans_list != NULL)) {
	CMConnection conn;
	CManager_unlock(cm);
	conn = (*trans_list)->initiate_conn(cm, &CMstatic_trans_svcs,
					    *trans_list, attrs);
#ifdef MODULE
	conn->resubmit_buf = NULL;
	conn->resubmit_len = 0;
#endif
	CManager_lock(cm);
	if (conn != NULL) {
	    if (CMtrace_on(conn->cm, CMConnectionVerbose)) {
		char *attr_str = attr_list_to_string(attrs);
		CMtrace_out(conn->cm, CMConnectionVerbose, 
			    "CM - Establish connection %lx - %s", (void*)conn,
			    attr_str);
		CMfree(attr_str);
	    }
	    return conn;
	}
	trans_list++;
    }
    return NULL;
}

CMConnection
CMinitiate_conn(cm, attrs)
CManager cm;
attr_list attrs;
{
    CMConnection conn;
    if (!cm->initialized) CMinitialize(cm);
    CManager_lock(cm);
    conn = CMinternal_initiate_conn(cm, attrs);
    CManager_unlock(cm);
    return conn;
}

void
CMConnection_add_reference(conn)
CMConnection conn;
{
    conn->ref_count++;
}

CMConnection
CMinternal_get_conn(cm, attrs)
CManager cm;
attr_list attrs;
{
    int i;
    CMConnection conn;
#ifndef MODULE
    assert(CManager_locked(cm));
#endif
    for (i=0; i<cm->connection_count; i++) {
	conn = cm->connections[i];
	if (conn->trans->connection_eq(cm, &CMstatic_trans_svcs,
				       conn->trans, attrs,
				       conn->transport_data)) {
	    conn->ref_count++;
	    return conn;
	}
    }
    conn = CMinternal_initiate_conn(cm, attrs);
    return conn;
}

CMConnection
CMget_conn(cm, attrs)
CManager cm;
attr_list attrs;
{
    CMConnection conn;
    if (!cm->initialized) CMinitialize(cm);
    CManager_lock(cm);
    conn = CMinternal_get_conn(cm, attrs);
    CManager_unlock(cm);
    return conn;
}

int
CMcontact_self_check(cm, attrs)
CManager cm;
attr_list attrs;
{
    transport_entry *trans_list;
    if (!cm->initialized) CMinitialize(cm);
    CManager_lock(cm);
    trans_list = cm->transports;
    while ((trans_list != NULL) && (*trans_list != NULL)) {
	int result = 0;
	CManager_unlock(cm);
	result = (*trans_list)->self_check(cm, &CMstatic_trans_svcs, 
					   *trans_list, attrs);
	if (result) return result;
	trans_list++;
	CManager_lock(cm);
    }
    CManager_unlock(cm);
    return 0;
}

/* alloc temporary buffer for CM use */
static CMbuffer
cm_get_data_buf(CManager cm, int length)  
{
    CMbuffer tmp = cm->cm_buffer_list;
    while (tmp != NULL) {
	if (!tmp->in_use_by_cm) {
	    if (tmp->size < length) {
		tmp->buffer = CMrealloc(tmp->buffer, length);
		tmp->size = length;
	    }
	    tmp->in_use_by_cm++;
	    return tmp;
	}
	tmp = tmp->next;
    }
    tmp = CMmalloc(sizeof(*tmp));
    tmp->buffer = CMmalloc(length);
    tmp->size = length;
    tmp->in_use_by_cm = 1;
    tmp->next = cm->cm_buffer_list;
    cm->cm_buffer_list = tmp;
    return tmp;
}

/* CM says that it is done with temporary buffer */
static void
cm_return_data_buf(CMbuffer cmb)
{
    cmb->in_use_by_cm = 0;
}

/* user wants CM temporary buffer */
static int
cm_user_take_data_buf(CManager cm, void *buffer)
{
    CMbuffer tmp = cm->cm_buffer_list;
    CMbuffer last = NULL;
    while (tmp != NULL) {
	if ((tmp->buffer <= buffer) && 
	    ((char*)buffer < ((char*)tmp->buffer + tmp->size))){
	    /* remove the buffer from CM's list */
	    if (last == NULL) {
		cm->cm_buffer_list = tmp->next;
	    } else {
		last->next = tmp->next;
	    }
	    /* add the buffer to the taken list */
	    tmp->next = cm->taken_buffer_list;
	    cm->taken_buffer_list = tmp;
	    return 1;
	}
	last = tmp;
	tmp = tmp->next;
    }
    return 0;
}
    
/* user is done with CM temporary buffer */
static int
cm_user_return_data_buf(CManager cm, void *buffer)
{
    CMbuffer tmp = cm->taken_buffer_list;
    CMbuffer last = NULL;
    while (tmp != NULL) {
	if ((tmp->buffer <= buffer) && 
	    ((char*)buffer < ((char*)tmp->buffer + tmp->size))){
	    /* remove the buffer from taken list */
	    if (last == NULL) {
		cm->taken_buffer_list = tmp->next;
	    } else {
		last->next = tmp->next;
	    }
	    /* add the buffer to the CMs list */
	    tmp->next = cm->cm_buffer_list;
	    cm->cm_buffer_list = tmp;
	    return 1;
	}
	last = tmp;
	tmp = tmp->next;
    }
    return 0;
}

int (*cm_write_hook)(int) = (int (*)(int)) NULL;
int (*cm_preread_hook)(int,char*) = (int (*)(int, char*)) NULL;
void (*cm_postread_hook)(int,char*) = (void (*)(int, char*)) NULL;
void (*cm_last_postread_hook)() = (void (*)()) NULL;

extern void CMDataAvailable(trans, conn)
transport_entry trans;
CMConnection conn;
{
    int byte_swap = 0;
    int length, actual, i, decoded_length;
    char *buffer = NULL, *decode_buffer = NULL;
    IOFormat format;
    CManager cm = conn->cm;
    CMincoming_format_list cm_format = NULL;
    int tmp, do_read = 1;
    CMbuffer cm_buffer = NULL, cm_decode_buf = NULL;
#ifdef MODULE
    select_data_ptr sd = *((select_data_ptr *)
	cm->control_list->network_polling_function.client_data);
    int iget;

    sd->loop = 0;
    sd->event = 0;
#endif

    if (conn->closed) {
	printf("data available on closed connection\n");
	return;
    }
    if (conn->foreign_data_handler != NULL) {
	((void(*)())conn->foreign_data_handler)(conn);
	return;
    }
    /* read first 4 bytes */
    if (cm_preread_hook) {
	buffer = (char*)&tmp;
	do_read = cm_preread_hook(4, buffer);
    }
    if (do_read) {
	CMtrace_out(cm, CMLowLevelVerbose, "CMdata reading 4 header bytes");
	if (trans->read_to_buffer_func) {
#ifndef MODULE
	    if (trans->read_to_buffer_func(&CMstatic_trans_svcs, 
					   conn->transport_data, &tmp, 
					   4) != 4) {
		CMConnection_close(conn);
		return;
	    }
#else
	if ((iget = trans->read_to_buffer_func(&CMstatic_trans_svcs, 
				       conn->transport_data, &tmp, 4, 
				       cm->control_list->tsk, 0)) != 4) {
	    if (iget == 0) 
  		return;
	    CMConnection_close(conn);
	    return;
	}
#endif
	    buffer = (char*)&tmp;
	} else {
	    buffer = trans->read_func(&CMstatic_trans_svcs, 
				      conn->transport_data,
				      4, &actual);
	    if (actual != 4) {
#ifdef MODULE
	        if (actual == 0) 
  		    return;
#endif
		CMConnection_close(conn);
		return;
	    }
	}
	if (cm_postread_hook) {
	    cm_postread_hook(4, buffer);
	}
    }
    switch (*((int*)buffer)) {  /* assume 4-byte int */
    case 0x00444d43: /* \0DMC reversed byte order */
	byte_swap = 1;
    case 0x434d4400:  /* CMD\0 */
	break;
    default:
	/*  non CM message */
	/*  lookup registered message prefixes and try to find handler */
	/*  otherwise give up */
	if (CMdo_non_CM_handler(conn, *(int*)buffer) == 0) {
	    printf("Unknown message on connection %lx\n", (long) conn);
	    CMConnection_close(conn);
	}
	return;
    }

    if (cm_preread_hook) {
	buffer = (char*)&tmp;
	do_read = cm_preread_hook(4, buffer);
    }
    if (do_read) {
	CMtrace_out(cm, CMLowLevelVerbose, "CMdata reading 4 length bytes");
	if (trans->read_to_buffer_func) {
#ifdef MODULE
	    if (trans->read_to_buffer_func(&CMstatic_trans_svcs, 
					   conn->transport_data, &tmp, 
					   4, cm->control_list->tsk, 0) != 4) {
#else
	    if (trans->read_to_buffer_func(&CMstatic_trans_svcs, 
					   conn->transport_data, &tmp, 
					   4) != 4) {
#endif
		CMConnection_close(conn);
		return;
	    }
	    buffer = (char*)&tmp;
	} else {
	    buffer = trans->read_func(&CMstatic_trans_svcs, 
				      conn->transport_data,
				      4, &actual);
	    if (actual != 4) {
		CMConnection_close(conn);
		return;
	    }
	}
	if (cm_postread_hook) {
	    cm_postread_hook(4, buffer);
	}
    }
    if (byte_swap) {
	((char*)&length)[0] = buffer[3];
	((char*)&length)[1] = buffer[2];
	((char*)&length)[2] = buffer[1];
	((char*)&length)[3] = buffer[0];
    } else {
	length = *(int *) buffer;
    }
    CManager_lock(conn->cm);
    cm_buffer = cm_get_data_buf(cm, length);
    if (cm_preread_hook) {
	buffer = cm_buffer->buffer;
	do_read = cm_preread_hook(length, buffer);
    }
    if (do_read) {
	CManager_unlock(conn->cm);
	CMtrace_out(cm, CMLowLevelVerbose, "CMdata reading %d message bytes",
		    length);
	if (trans->read_to_buffer_func) {
#ifdef MODULE
	    if (trans->read_to_buffer_func(&CMstatic_trans_svcs,
					   conn->transport_data, 
					   cm_buffer->buffer,
					   length, cm->control_list->tsk, 1) 
			    			!= length) {
#else
	    if (trans->read_to_buffer_func(&CMstatic_trans_svcs,
					   conn->transport_data, 
					   cm_buffer->buffer,
					   length) != length) {
#endif
		CMConnection_close(conn);
		cm_return_data_buf(cm_buffer);
		return;
	    }
	    buffer = cm_buffer->buffer;
	} else {
	    buffer = trans->read_func(&CMstatic_trans_svcs, 
				      conn->transport_data,
				      length, &actual);
	    cm_return_data_buf(cm_buffer);
	    cm_buffer = NULL;
	    if (actual != length) {
		CMConnection_close(conn);
		return;
	    }
	}
	if (cm_postread_hook) {
	    cm_postread_hook(length, buffer);
	}
	CManager_lock(conn->cm);
    }
    if (cm_last_postread_hook) {
	cm_last_postread_hook();
    }
    format = get_format_app_IOcontext(conn->cm->IOcontext, buffer, conn);
    if (format == NULL) {
	fprintf(stderr, "invalid format in incoming buffer\n");
	cm_return_data_buf(cm_buffer);
	CManager_unlock(conn->cm);
	return;
    }
    CMtrace_out(cm, CMDataVerbose, "CM - Receiving record of type %s", 
		name_of_IOformat(format));
    for (i=0; i< cm->in_format_count; i++) {
	if (cm->in_formats[i].format == format) {
	    cm_format = &cm->in_formats[i];
	}
    }
    if (cm_format == NULL) {
	cm_format = CMidentify_CMformat(cm, format);
    }

    if ((cm_format == NULL) || (cm_format->handler == NULL)) {
	fprintf(stderr, "CM - No handler for incoming data of this version of format \"%s\"\n",
		name_of_IOformat(format));
	cm_return_data_buf(cm_buffer);
	CManager_unlock(conn->cm);
	return;
    }
    assert(has_conversion_IOformat(format));
    decoded_length = this_IOrecord_length(cm->IOcontext, buffer, length);

    if (decode_in_place_possible(format)) {
	if (!decode_in_place_IOcontext(cm->IOcontext, buffer, 
				       (void**)&decode_buffer)) {
	    printf("Decode failed\n");
	    cm_return_data_buf(cm_buffer);
	    CManager_unlock(conn->cm);
	    return;
	}
    } else {
	cm_decode_buf = cm_get_data_buf(cm, decoded_length);
	decode_buffer = cm_decode_buf->buffer;
	decode_to_buffer_IOcontext(cm->IOcontext, buffer, decode_buffer);
	cm_return_data_buf(cm_buffer);
	cm_buffer = NULL;
    }
    if (CMtrace_on(conn->cm, CMDataVerbose)) {
	static int dump_char_limit = 256;
	static int warned = 0;
	static int size_set = 0;
	int r;
	if (size_set == 0) {
	    char *size_str = getenv("CMDumpSize");
	    size_set++;
	    if (size_str != NULL) {
		dump_char_limit = atoi(size_str);
	    }
	}
	printf("CM - record contents are:\n  ");
	r = dump_limited_unencoded_IOrecord((IOFile)cm->IOcontext, format,
					    decode_buffer, dump_char_limit);
	if (r && !warned) {
	    printf("\n\n  ****  Warning **** CM record dump truncated\n");
	    printf("  To change size limits, set CMDumpSize environment variable.\n\n\n");
	    warned++;
	}
    }
    CManager_unlock(conn->cm);
    cm_format->handler(cm, conn, decode_buffer, cm_format->client_data);
    if (cm_buffer) cm_return_data_buf(cm_buffer);
    if (cm_decode_buf) cm_return_data_buf(cm_decode_buf);
    CMtrace_out(cm, CMDataVerbose, "CM - Finish processing - record of type %s", 
		name_of_IOformat(format));
#ifdef MODULE
    sd->loop = 1;
    if ((strcmp(name_of_IOformat(format), "Event Message") == 0) ||
        (strcmp(name_of_IOformat(format), "EventV Message") == 0))
      sd->event = 1;
#endif
    return;
}


void *
CMtake_buffer(cm, data)
CManager cm;
void *data;
{
    if (cm_user_take_data_buf(cm, data) == 0) {
	/*
	 * someday we may get here if we have transports that allocate 
	 * their own buffers.  How to support?  Maybe do a copy and 
	 * return the copy...?
	 */
	fprintf(stderr, "Error: CMtake_buffer called with record not associated with cm\n");
	return NULL;
    } else {
	return data;
    }
}

extern void
CMreturn_buffer(cm, data)
CManager cm;
void *data;
{
    if (cm_user_return_data_buf(cm, data) == 0) {
	/*
	 * someday we may get here if we have transports that allocate 
	 * their own buffers.  How to support?  Maybe do a copy and 
	 * return the copy...?
	 */
	fprintf(stderr, "Error: CMtake_buffer called with record not associated with cm\n");
    }
}

void
CMregister_handler(format, handler, client_data)
CMFormat format;
CMHandlerFunc handler;
void *client_data;
{
    format->handler = handler;
    format->client_data = client_data;
}

extern int
CMwrite(conn, format, data)
CMConnection conn;
CMFormat format;
void *data;
{
    return CMwrite_attr(conn, format, data, NULL);
}

extern int
CMwrite_attr(conn, format, data, attrs)
CMConnection conn;
CMFormat format;
void *data;
attr_list attrs;
{
    int header[2] = {0x434d4400, 0};  /* CMD\0 in first entry */
    IOEncodeVector vec;
    IOEncodeVector tmp_vec2;
    int length = 0, vec_count = 0, actual;
    int do_write = 1;
    int i;

    /* ensure conn is open */
    if (conn->closed != 0) {
	CMtrace_out(conn->cm, CMDataVerbose, "Not writing data to closed connection");
	return 0;
    }

    if (format->registration_pending) {
	CMcomplete_format_registration(format, 1);
    }
    if (format->format == NULL) {
	printf("Format registration has failed for format \"%s\" - write aborted\n",
	       format->format_name);
	return 0;
    }
    CMformat_preload(conn, format);

    if (CMtrace_on(conn->cm, CMDataVerbose)) {
	static int dump_char_limit = 256;
	static int warned = 0;
	static int size_set = 0;
	int r;
	if (size_set == 0) {
	    char *size_str = getenv("CMDumpSize");
	    size_set++;
	    if (size_str != NULL) {
		dump_char_limit = atoi(size_str);
	    }
	}
	printf("CM - Writing record of type %s\n",
	       name_of_IOformat(format->format));
	printf("CM - record contents are:\n  ");
	r = dump_limited_unencoded_IOrecord((IOFile)format->IOsubcontext, 
					    format->format,
					    data, dump_char_limit);
	if (r && !warned) {
	    printf("\n\n  ****  Warning **** CM record dump truncated\n");
	    printf("  To change size limits, set CMDumpSize environment variable.\n\n\n");
	    warned++;
	}
    }

    CMConn_write_lock(conn);
    /* encode data with CM context */
    vec = encode_IOcontext_vectorB(format->IOsubcontext, conn->io_out_buffer,
				   format->format, data);

    while(vec[vec_count].iov_base != NULL) {
	length += vec[vec_count].iov_len;
	vec_count++;
    }
    header[1] = length;
    if (cm_write_hook != NULL) {
	do_write = cm_write_hook(length);
    }
    if (do_write) {
#ifdef MODULE
	struct resubmit_node *p = conn->resubmit_buf, *q;
	IOEncodeVector resubmit_vector, temp;
#endif
	struct _io_encode_vec static_vec[100];
	IOEncodeVector tmp_vec = &static_vec[0];
	if (vec_count >= sizeof(static_vec)/ sizeof(static_vec[0])) {
	    tmp_vec = CMmalloc((vec_count+1) * sizeof(*tmp_vec));
	}
	tmp_vec[0].iov_base = &header;
	tmp_vec[0].iov_len = sizeof(header);
	memcpy(&tmp_vec[1], vec, sizeof(*tmp_vec) * vec_count);

#ifdef MODULE
	/* Added by Feng Li: Should vec be freed? */
	/* First, Check whether there are unsubmitted events */
	if (p != NULL) {
		int count, submitted;
		//printk("Resubmit (%d)...\n", conn->resubmit_len);
		while (p != NULL) {
			/* Submit event again */
			resubmit_vector = p->buf_ptr;
			count = p->count;
			submitted = 0;
			if (conn->trans->writev_attr_func != NULL) {
	 		   	submitted = 
				conn->trans->writev_attr_func(&CMstatic_trans_svcs, 
						conn->transport_data, 
						resubmit_vector,
						count, attrs);
			} else {
				submitted = conn->trans->writev_func(&CMstatic_trans_svcs, 
						conn->transport_data, 
						resubmit_vector, count);
			}
			if (submitted != count) {
				//printk("submitted: %d, count: %d\n", submitted, count);
				/* Fail, adjust vector, add new event */
				temp = CMmalloc((count-submitted) * sizeof(*temp));
				memcpy(temp, &resubmit_vector[submitted], sizeof(*temp) * (count-submitted));
				temp[count-submitted].iov_base = NULL;
				CMfree(resubmit_vector);
				p->buf_ptr = temp;
				p->count = count-submitted;
				q = (struct resubmit_node*)CMmalloc(sizeof(struct resubmit_node));
				q->next = NULL;
				tmp_vec2 = CMmalloc((vec_count+1) * 
					sizeof(*tmp_vec2));
				memcpy(tmp_vec2, &tmp_vec[0], sizeof(*tmp_vec2) * 
					(vec_count+1));
				q->buf_ptr = tmp_vec2;	
				q->count = vec_count+1;
				while (p->next != NULL)
					p = p->next;
				p->next = q;
				conn->resubmit_len++;

				CMConn_write_unlock(conn);
				//printk("Resubmit failed!\n");
				return 0;
			}
			// Success, free the vector and node, then continue to submit next node
			CMfree(p->buf_ptr);
			q = p;
			conn->resubmit_buf = p = p->next;
			CMfree(q);
			conn->resubmit_len--;
			//printk("Resubmit completed!\n");
	       	}
		/* All the remainder events are submitted, resubmit_buf should be NULL */
		/* Continue to submit current event */
	}
#endif
	CMtrace_out(conn->cm, CMLowLevelVerbose, "Writing %d vectors, total %d bytes in writev", vec_count +1, length + sizeof(header));
	if (conn->trans->writev_attr_func != NULL) {
//printk("writev_attr_func\n");
	    actual = conn->trans->writev_attr_func(&CMstatic_trans_svcs, 
						   conn->transport_data, 
						   tmp_vec,
						   vec_count+1, attrs);
	} else {
//printk("write_attr_func\n");
	    actual = conn->trans->writev_func(&CMstatic_trans_svcs, 
					      conn->transport_data, 
					      tmp_vec, vec_count+1);
	}
//	if (vec_count >= sizeof(static_vec)/ sizeof(static_vec[0])) {
//	    CMfree(tmp_vec);
//	}
	//printk("actual=%d, count=%d\n", actual, (vec_count+1));
	if (actual != (vec_count + 1)) {
	    /* fail */
#ifdef MODULE
		/* Added by Feng Li */
		//struct resubmit_timer_data *data;
		//printk("Construct resubmit buf...(alloc:%d)\n",
				//(vec_count+1-actual)*sizeof(*tmp_vec));
		resubmit_vector = CMmalloc((vec_count+1-actual) * sizeof(*tmp_vec));
		//if (resubmit_vector == NULL) printk("malloc error!\n");
		memcpy(resubmit_vector, &tmp_vec[actual], sizeof(*tmp_vec) * 
				(vec_count+1-actual));
		/* Now, resubmit_buf is NULL */
		conn->resubmit_buf = (struct resubmit_node*)CMmalloc(sizeof(struct resubmit_node));
		conn->resubmit_buf->next = NULL;
		conn->resubmit_buf->buf_ptr = resubmit_vector;
		conn->resubmit_buf->count = vec_count+1-actual;
		conn->resubmit_len = 1;
		if (vec_count >= sizeof(static_vec)/ sizeof(static_vec[0])) {
	//printk("HERE!\n");
		    CMfree(tmp_vec);
		}
		/* Need to add a timer */
		/*
		printk("Add timer...\n");
                timer = CMmalloc(sizeof(struct timer_list));
                timer->expires = jiffies + 1000;
		data = CMmalloc(sizeof(struct resubmit_timer_data));
		data->conn = conn;
		data->attrs = attrs;
		timer->data = (unsigned long)data;
                timer->function = resubmit_timer_func;
		*/
		/*
                conn->resubmit_timer.function = resubmit_timer_func;
                conn->resubmit_timer.expires = jiffies + HZ;
		data = CMmalloc(sizeof(struct resubmit_timer_data));
		data->conn = conn;
		data->attrs = attrs;
		conn->resubmit_timer.data = (unsigned long)data;
		add_timer(&conn->resubmit_timer);
		*/
		if (conn->thread_used == -1) {
			conn->send_thread.data = conn;
			//printk("Start thread!\n");
			if (start_kthread_args(resubmit_thread, &conn->send_thread,
						conn) < 0) {
				printk("Failed creating resubmit thread!\n");
			} 
			conn->send_thread.used = 1;
			conn->thread_used = 1;
			//printk("Start thread done!\n");
			i = 0;
			while ((resubmit_threads[i] != NULL)&&(i < MAX_THREADS))
				i++;
			resubmit_threads[i] = &conn->send_thread;
		}
#endif
    		CMConn_write_unlock(conn);
 	    return 0;
 	}
	if (vec_count >= sizeof(static_vec)/ sizeof(static_vec[0])) {
//printk("HERE2\n");
	    CMfree(tmp_vec);
	}
     }
     CMConn_write_unlock(conn);
     return 1;
 }
 
#ifdef MODULE
int resubmit_thread(void *kargs) 
{
	int i;
	int found = -1;
	//IOEncodeVector resubmit_vector, temp;
	struct kthread_args *args = (struct kthread_args *) kargs;
	CMConnection conn = (CMConnection) args->data;
	//struct resubmit_node *p = conn->resubmit_buf, *q;

	init_kthread(&conn->send_thread, "resubmit_thread");
	//printk("Start thread is done!\n");
	for (i = 0; i < MAX_THREADS; i++) {
		if (resubmit_threads[i] != NULL)
			if (resubmit_threads[i]->task == 
				conn->send_thread.task) 
				found = i;
	}
	if (current->qfabric == NULL) 
		initialize_qfabric();
	while (terminate_threads != 1) {
		CMtask_sleep(HZ);
		/*
	     	CMConn_write_lock(conn);
		p = conn->resubmit_buf;
		if (p != NULL) {
		    int count, submitted;
                    while (p != NULL) {
                        resubmit_vector = p->buf_ptr;
                        count = p->count;
                        submitted = 0;
                        if (conn->trans->writev_attr_func != NULL) {
                                submitted = conn->trans->writev_attr_func(&CMstatic_trans_svcs,
                                                conn->transport_data,
                                                resubmit_vector,
                                                count, attrs);
                        } else {
                                submitted = conn->trans->writev_func(&CMstatic_trans_svcs,
                                                conn->transport_data,
                                                resubmit_vector, count);
                        }
			submitted = 0;
                        if (submitted != count) {
                                temp = CMmalloc((count-submitted) * sizeof(*temp));
                                memcpy(temp, &resubmit_vector[submitted], sizeof(*temp) * (count-submitted));
                                CMfree(resubmit_vector);
                                p->buf_ptr = temp;
                                p->count = count-submitted;

                                break;	// Get out of while
                        }
                        // Success, free the vector and node, then continue to submit next node
                        CMfree(p->buf_ptr);
                        q = p;
                        conn->resubmit_buf = p = p->next;
                        CMfree(q);
			conn->resubmit_len--;
                    } // end while
                } // end if
	    	CMConn_write_unlock(conn);
		*/
	}
	exit_kthread(resubmit_threads[found]);
}

void resubmit_timer_func(unsigned long data)
{
	/* First, resubmit the events in queue */
	CMConnection conn = ((struct resubmit_timer_data *)data)->conn;
	attr_list attrs = ((struct resubmit_timer_data *)data)->attrs;
	struct resubmit_node *p = conn->resubmit_buf, *q;
	IOEncodeVector resubmit_vector, temp;

//    	CMConn_write_lock(conn);
	/* First, Check whether there are unsubmitted events */
	if (p != NULL) {
		int count, submitted;
                while (p != NULL) {
                        /* Submit event again */
                        resubmit_vector = p->buf_ptr;
                        count = p->count;
                        submitted = 0;
                        if (conn->trans->writev_attr_func != NULL) {
                                submitted = conn->trans->writev_attr_func(&CMstatic_trans_svcs,
                                                conn->transport_data,
                                                resubmit_vector,
                                                count, attrs);
                        } else {
                                submitted = conn->trans->writev_func(&CMstatic_trans_svcs,
                                                conn->transport_data,
                                                resubmit_vector, count);
                        }
                        if (submitted != count) {
                                /* Fail, adjust vector, add new event */
                                temp = CMmalloc((count-submitted) * sizeof(*temp));
                                memcpy(temp, &resubmit_vector[submitted], sizeof(*temp) * (count-submitted));
                                CMfree(resubmit_vector);
                                p->buf_ptr = temp;
                                p->count = count-submitted;

                                break;	// Get out of while
                        }
                        // Success, free the vector and node, then continue to submit next node
                        CMfree(p->buf_ptr);
                        q = p;
                        conn->resubmit_buf = p = p->next;
                        CMfree(q);
			conn->resubmit_len--;
                } // end while
                /* Now, all the remainder events are submitted, resubmit_buf should be NULL */
        } // end if
	
//	CMConn_write_unlock(conn);

#ifdef ABCD
        /* Second, if there are events unsubmitted, add a new timer */
	if (conn->resubmit_buf != NULL) {
		struct timer_list *timer;
		/* Need to add a timer */
                timer = CMmalloc(sizeof(struct timer_list));
                timer->expires = jiffies + 1000;
		timer->data = data;
                timer->function = resubmit_timer_func;
		add_timer(timer);
	} else {
		CMfree((struct resubmit_timer_data *)data);
	}
#endif
}

void initiate_resubmit_buf(CMConnection conn)
{
	conn->resubmit_buf = NULL;
	conn->resubmit_len = 0;
}
#endif

static void CM_init_select ARGS((CMControlList cl, CManager cm));

extern void
CM_fd_add_select(cm, fd, handler_func, param1, param2)
CManager cm;
int fd;
select_list_func handler_func;
void *param1;
void *param2;
{
    if (!cm->control_list->select_initialized) {
	CM_init_select(cm->control_list, cm);
    }
#ifdef MODULE
    cm->control_list->add_select(cm, &CMstatic_trans_svcs,
				 &cm->control_list->select_data, fd,
				 handler_func, param1, param2);
#else
    cm->control_list->add_select(&CMstatic_trans_svcs,
				 &cm->control_list->select_data, fd,
				 handler_func, param1, param2);
#endif
}

extern void
CM_fd_remove_select(cm, fd)
CManager cm;
int fd;
{
    if (!cm->control_list->select_initialized) {
	CM_init_select(cm->control_list, cm);
    }
#ifdef MODULE
    cm->control_list->remove_select(cm, &CMstatic_trans_svcs,
				    &cm->control_list->select_data, fd);
#else
    cm->control_list->remove_select(&CMstatic_trans_svcs,
				    &cm->control_list->select_data, fd);
#endif
}

extern CMTaskHandle
CMadd_periodic(cm, period, func, client_data)
CManager cm;
long period;
CMPollFunc func;
void *client_data;
{
    CMTaskHandle handle = CMmalloc(sizeof(*handle));
    if (!cm->control_list->select_initialized) {
	CM_init_select(cm->control_list, cm);
    }
    handle->cm = cm;
#ifdef MODULE
    handle->task = 
	cm->control_list->add_periodic(cm, &CMstatic_trans_svcs,
				       &cm->control_list->select_data,
				       0, period, (select_list_func)func, 
				       (void*)cm, client_data);
#else
    handle->task = 
	cm->control_list->add_periodic(&CMstatic_trans_svcs,
				       &cm->control_list->select_data,
				       0, period, (select_list_func)func, 
				       (void*)cm, client_data);
#endif
    if (handle->task == NULL) {
	free(handle);
	return NULL;
    }
    return handle;
}

extern CMTaskHandle
CMadd_periodic_task(cm, period_sec, period_usec, func, client_data)
CManager cm;
int period_sec;
int period_usec;
CMPollFunc func;
void *client_data;
{
    CMTaskHandle handle = CMmalloc(sizeof(*handle));
    if (!cm->control_list->select_initialized) {
	CM_init_select(cm->control_list, cm);
    }
    handle->cm = cm;
#ifdef MODULE
    handle->task = 
	cm->control_list->add_periodic(cm, &CMstatic_trans_svcs,
				       &cm->control_list->select_data,
				       period_sec, period_usec, 
				       (select_list_func)func, 
				       (void*)cm, client_data);
#else
    handle->task = 
	cm->control_list->add_periodic(&CMstatic_trans_svcs,
				       &cm->control_list->select_data,
				       period_sec, period_usec, 
				       (select_list_func)func, 
				       (void*)cm, client_data);
#endif
    if (handle->task == NULL) {
	free(handle);
	return NULL;
    }
    return handle;
}

extern void
CMremove_periodic(handle)
CMTaskHandle handle;
{
    CManager cm = handle->cm;
    cm->control_list->remove_periodic(&CMstatic_trans_svcs,
				      &cm->control_list->select_data, 
				      handle->task);
    free(handle);
}

extern void
CMremove_task(handle)
CMTaskHandle handle;
{
    CManager cm = handle->cm;
    cm->control_list->remove_periodic(&CMstatic_trans_svcs,
				      &cm->control_list->select_data, 
				      handle->task);
    free(handle);
}

extern CMTaskHandle
CMadd_delayed_task(cm, delay_sec, delay_usec, func, client_data)
CManager cm;
int delay_sec;
int delay_usec;
CMPollFunc func;
void *client_data;
{
    CMTaskHandle handle = CMmalloc(sizeof(*handle));
    if (!cm->control_list->select_initialized) {
	CM_init_select(cm->control_list, cm);
    }
    handle->cm = cm;
#ifdef MODULE
    handle->task = 
	cm->control_list->add_delayed_task(cm, &CMstatic_trans_svcs,
					   &cm->control_list->select_data,
					   delay_sec, delay_usec,
					   (select_list_func)func, 
					   (void*)cm, client_data);
#else
    handle->task = 
	cm->control_list->add_delayed_task(&CMstatic_trans_svcs,
					   &cm->control_list->select_data,
					   delay_sec, delay_usec,
					   (select_list_func)func, 
					   (void*)cm, client_data);
#endif
    if (handle->task == NULL) {
	free(handle);
	return NULL;
    }
    return handle;
}

static void
CM_init_select(cl, cm)
CMControlList cl;
CManager cm;
{
    CMPollFunc blocking_function, polling_function;
#ifdef MODULE
    typedef void (*SelectInitFunc)(CManager cm, CMtrans_services svc,  
		    void *client_data);
#else
    typedef void (*SelectInitFunc)(CMtrans_services svc, void *client_data);
#endif
    SelectInitFunc init_function;
#ifndef MODULE
    lt_dlhandle handle;	
    CMControlList_lock(cl);
    if (lt_dlinit() != 0) {
	fprintf (stderr, "error during initialization: %s\n", lt_dlerror());
	return;
    }
    lt_dladdsearchdir(CM_LIBRARY_BUILD_DIR);
    lt_dladdsearchdir(CM_LIBRARY_INSTALL_DIR);
    handle = lt_dlopen("libcmselect.la");
    if (!handle) {
	fprintf(stderr, "Failed to load required select dll.  Error \"%s\".\n",
		lt_dlerror());
	fprintf(stderr, "Search path includes '.', '%s', '%s' and any default search paths supported by ld.so\n", CM_LIBRARY_BUILD_DIR, 
		CM_LIBRARY_INSTALL_DIR);
	exit(1);
    }
    cl->add_select = (CMAddSelectFunc)lt_dlsym(handle, "add_select");  
    cl->remove_select = (CMRemoveSelectFunc)lt_dlsym(handle, "remove_select");  
    cl->add_periodic = (CMAddPeriodicFunc)lt_dlsym(handle, "add_periodic");  
    cl->add_delayed_task = 
	(CMAddPeriodicFunc)lt_dlsym(handle, "add_delayed_task");  
    cl->remove_periodic = (CMRemovePeriodicFunc)lt_dlsym(handle, "remove_periodic");  
    blocking_function = (CMPollFunc)lt_dlsym(handle, "blocking_function");
    polling_function = (CMPollFunc)lt_dlsym(handle, "polling_function");
    init_function = (SelectInitFunc)lt_dlsym(handle, "select_initialize");
#else
    cl->add_select = (CMAddSelectFunc)libcmselect_LTX_add_select;  
    cl->remove_select = (CMRemoveSelectFunc)libcmselect_LTX_remove_select;  
    cl->add_periodic = (CMAddPeriodicFunc)libcmselect_LTX_add_periodic;  
    cl->add_delayed_task = 
	(CMAddPeriodicFunc)libcmselect_LTX_add_delayed_task;  
    cl->remove_periodic = (CMRemovePeriodicFunc)
	libcmselect_LTX_remove_periodic;  
    blocking_function = (CMPollFunc)libcmselect_LTX_blocking_function;
    polling_function = (CMPollFunc)libcmselect_LTX_polling_function;
    init_function = (SelectInitFunc)libcmselect_LTX_select_initialize;
#endif
    if ((cl->add_select == NULL) || (cl->remove_select == NULL) || 
	(blocking_function == NULL) || (cl->add_periodic == NULL) ||
	(cl->remove_periodic == NULL)) {
	printf("Select failed to load properly\n");
	exit(1);
    }
#ifdef MODULE
    init_function(cm, &CMstatic_trans_svcs, &cm->control_list->select_data);
#else
    init_function(&CMstatic_trans_svcs, &cm->control_list->select_data);
#endif
    CMControlList_set_blocking_func(cl, cm, blocking_function, 
				    polling_function,
				    (void*)&(cl->select_data));
    cl->select_initialized = 1;
    CMControlList_unlock(cl);
}

static void
wake_function(cm, cond)
CManager cm;
void *cond;
{
    CMCondition_signal(cm, (int)(long)cond);
}

extern void
CMsleep(cm, sec)
CManager cm;
int sec;
{
    int cond = CMCondition_get(cm, NULL);
    CMTaskHandle handle = 
	CMadd_delayed_task(cm, sec, 0, wake_function, (void*)cond);
    CMfree(handle);
    CMCondition_wait(cm, cond);
}

extern void
CMusleep(cm, usec)
CManager cm;
int usec;
{
    int cond = CMCondition_get(cm, NULL);
    CMTaskHandle handle = 
	CMadd_delayed_task(cm, 0, usec, wake_function, (void*)cond);
    CMfree(handle);
    CMCondition_wait(cm, cond);
}

typedef struct foreign_handler_struct {
    int header;
    CMNonCMHandler handler;
} *handler_list;

static handler_list foreign_handler_list;
static int foreign_handler_count = 0;

extern void
CMregister_non_CM_message_handler(header, handler)
int header;
CMNonCMHandler handler;
{
    if (foreign_handler_count > 0) {
	foreign_handler_list = CMrealloc(foreign_handler_list, 
					 sizeof(foreign_handler_list[0]) * 
					 (foreign_handler_count + 1));
    } else {
	foreign_handler_list = CMmalloc(sizeof(foreign_handler_list[0]));
    }
    foreign_handler_list[foreign_handler_count].header = header;
    foreign_handler_list[foreign_handler_count].handler = handler;
    foreign_handler_count++;
}

static int
CMdo_non_CM_handler(conn, header)
CMConnection conn;
int header;
{
    int i = 0;
    while (i < foreign_handler_count) {
	if (foreign_handler_list[i].header == header) {
	    foreign_handler_list[i].handler(conn, header);
	    return 1;
	}
	i++;
    }
    return 0;
}

#ifdef MODULE
void wake_up_function(data)
unsigned long data;
{
  wake_up_interruptible((wait_queue_head_t *)data);
}

extern int
CMtask_sleep(time)
int time;
{
  int expire = jiffies + time;

  if (current->qfabric == NULL)
    initialize_qfabric();

  interruptible_sleep_on_timeout(&current->qfabric->kecho_queue, time);
  expire = expire - jiffies;
  return expire < 0 ? 0 : expire;
}

extern void cm_task_wake_up(tsk, fd, pid, ptr, kecho_identifier)
struct task_struct *tsk;
int fd;
int pid;
void *ptr;
int kecho_identifier;
{
  struct timeval tv;
  select_data_ptr sd = (select_data_ptr)ptr;

  tsk->qfabric->sw_pending = 0;

  my_FD_SET(fd, (char *) sd->fdset);
  if (tsk->kthread == 1) {
    if (tsk == current) tsk->state = TASK_RUNNING;
    else if ((tsk->state == TASK_INTERRUPTIBLE) && (tsk != NULL))
      wake_up_interruptible((wait_queue_head_t *)&tsk->qfabric->kecho_queue);
  } else {
    //sd->cm->control_list->tsk = tsk;
    //CMpoll_network(sd->cm);
    //sd->cm->control_list->tsk = NULL;
    wake_up_process(tsk);
    do_gettimeofday(&tv);
    do_upcall(kecho_identifier, tsk->pid, 0, tv.tv_sec, tv.tv_usec, 0);
  }
}

extern int
CMtask_kill(pid)
int pid;
{
  struct task_struct *tsk;

  tsk = find_task_by_pid(pid);
  if (tsk == NULL) return 0;
  if (tsk->qfabric != NULL)
    tsk->qfabric->sw_pending = -1;

  if (tsk->state == TASK_INTERRUPTIBLE)
    wake_up_interruptible((wait_queue_head_t *)&tsk->qfabric->kecho_queue);

  if (current->qfabric == NULL)
    initialize_qfabric();
  sleep_on_timeout(&current->qfabric->kecho_queue, HZ/2);

  return 0;
} 
#endif
