#ifndef __I_O__
#include <io.h>
#endif
#ifndef GEN_THREAD_H
#define thr_mutex_t void*
#define thr_thread_t void *
#endif
#ifdef MODULE
#include <asm/semaphore.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include "../include/utils.h"
#endif

/*
#ifdef MODULE
typedef struct kthread {
	char *name;
	struct task_struct *task;
	struct completion done;
#define KTH_RUNNING	1
#define KTH_SHUTDOWN	2
	long state;
	int (*init)(struct kthread *, void *);
	void (*cleanup)(struct kthread *, void *);
	int (*main)(struct kthread *, void *);
	void *data;
	int used;
} kthread_t;
#endif
*/

typedef struct _transport_item *transport_entry;

typedef struct _CMincoming_format {
    IOFormat format;
    CMHandlerFunc handler;
    void *client_data;
} *CMincoming_format_list;

struct _CMControlList;
typedef struct _CMControlList *CMControlList;

typedef struct _CMbuffer {
    void *buffer;
    int size;
    int in_use_by_cm;
    struct _CMbuffer *next;
} *CMbuffer;

struct _pending_format_requests {
    char *server_id;
    int id_length;
    int condition;
    int top_request;
};

typedef struct _CManager {
    transport_entry *transports;
    int initialized;
    int reference_count;

    CMControlList control_list;	/* the control list for this DE */

    int in_format_count;
    CMincoming_format_list in_formats;
    
    int reg_format_count;
    CMFormat *reg_formats;

    int reg_user_format_count;
    CMFormat *reg_user_formats;
  
    int pending_request_max;
    struct _pending_format_requests *pbio_requests;

    int connection_count;
    CMConnection *connections;

#ifdef MODULE
    struct semaphore exchange_lock;
#else
    thr_mutex_t exchange_lock;
#endif
    int locked;
    int closed;

    IOContext IOcontext;	/* pbio context for data encoding */
#ifdef MODULE
    struct semaphore context_lock;
#else
    thr_mutex_t context_lock;
#endif

    CMbuffer taken_buffer_list;
    CMbuffer cm_buffer_list;

    attr_list *contact_lists;
#ifdef MODULE
    struct timer_list cm_timer;
    int event_delivered;
    void *func;
    void *event;
    int event_len;
    void *client_data;
    attr_list event_attrs;
#endif
} CManager_s;

typedef struct _CMCondition *CMCondition;

typedef void (*CMfree_func) ARGS((void *block));

typedef enum _CMControlStyle {
    CMSingleThreaded, CMDedicatedServerThread, CMOccasionalPolling
} CMControlStyle;

typedef struct free_block_rec {
    int ref_count;
    CManager cm;
    void *block;
    CMfree_func free_func;
    CManager locking_cm;
} *free_block_rec_p;

typedef void (*CMNetworkFunc) ARGS((void *svcs, void *client_data));

typedef struct func_entry {
	CMPollFunc func;
	CManager cm;
	void *client_data;
} func_entry;

typedef void (*select_list_func) ARGS((void *, void*));

#ifdef MODULE
typedef void (*CMAddSelectFunc) ARGS((CManager cm,
			     	      void *svcs, void *select_data, int fd,
				      select_list_func func,
				      void *param1, void *param2));
#else
typedef void (*CMAddSelectFunc) ARGS((void *svcs, void *select_data, int fd,
				      select_list_func func,
				      void *param1, void *param2));
#endif

#ifdef MODULE
typedef void (*CMRemoveSelectFunc) ARGS((CManager cm, void *svcs, void *select_data, int fd));
#else
typedef void (*CMRemoveSelectFunc) ARGS((void *svcs, void *select_data, int fd));
#endif

typedef struct _periodic_task *periodic_task_handle;

#ifdef MODULE
typedef periodic_task_handle (*CMAddPeriodicFunc) 
    ARGS((CManager cm, void *svcs, void *select_data, 
			    int period_sec, int period_usec,
	  select_list_func func, void *param1, void *param2));
#else
typedef periodic_task_handle (*CMAddPeriodicFunc) 
    ARGS((void *svcs, void *select_data, int period_sec, int period_usec,
	  select_list_func func, void *param1, void *param2));
#endif

typedef void (*CMRemovePeriodicFunc) ARGS((void *svcs, void *select_data, 
					   periodic_task_handle handle));

typedef struct _CMControlList {
    func_entry network_blocking_function;
    func_entry network_polling_function;
    func_entry *polling_function_list;
    int cl_consistency_number;

    int select_initialized;
    void *select_data;
    CMAddSelectFunc add_select;
    CMRemoveSelectFunc remove_select;
    CMAddPeriodicFunc add_periodic;
    CMAddPeriodicFunc add_delayed_task;
    CMRemovePeriodicFunc remove_periodic;

    /* 
     * CLs can be used by multiple DEs, close it when ref count reaches
     * zero 
     */
    int reference_count;
    int free_reference_count;

    CMCondition condition_list;
    int next_condition_num;

#ifdef MODULE
    struct task_struct *tsk;
    struct semaphore list_mutex;
#else
    thr_mutex_t list_mutex;
#endif
    int locked;

    int closed;
    int has_thread;
    thr_thread_t server_thread;
} CMControlList_s;

#ifdef MODULE
struct resubmit_timer_data {
	CMConnection conn;
	attr_list attrs;
};
struct resubmit_node {
	struct resubmit_node *next;
	IOEncodeVector buf_ptr;
	int count;
};
#endif

struct _CMConnection {
    CManager cm;
    /* remote contact info */

    transport_entry trans;
    void *transport_data;
    int ref_count;
#ifdef MODULE
    struct semaphore write_lock;
    struct semaphore read_lock;
#else
    thr_mutex_t write_lock;
    thr_mutex_t read_lock;
#endif
    IOBuffer io_out_buffer;
    int closed;

    CMFormat *downloaded_formats;

    CMCloseHandlerFunc close_handler;
    void *close_client_data;

    IOContext IOsubcontext;
    void *foreign_data_handler;

    attr_list attrs;
#ifdef MODULE
    struct resubmit_node *resubmit_buf;
    int resubmit_len;
    //struct timer_list resubmit_timer;
    kthread_t send_thread;
    int thread_used;
#endif
};

struct _CMFormat {
    CManager cm;
    char *format_name;
    IOContext IOsubcontext;
    IOFormat  format;
    void *field_list_addr;
    CMHandlerFunc handler;
    void *client_data;
    IOFieldList field_list;
    CMFormatList subformat_list;
    int registration_pending;
};

#include "cm_transport.h"

typedef void (*DataAvailableCallback) ARGS((transport_entry trans, CMConnection conn));

struct _transport_item {
    char *trans_name;
    CManager cm;
    DataAvailableCallback data_available;
    CMTransport_func  transport_init;
    CMTransport_listen_func  listen;
    CMTransport_conn_func  initiate_conn;
    CMTransport_self_check_func  self_check;
    CMTransport_connection_eq_func  connection_eq;
    CMTransport_shutdown_conn_func  shutdown_conn;
    CMTransport_read_to_buffer_func read_to_buffer_func;
    CMTransport_read_func read_func;
    CMTransport_write_func write_func;
    CMTransport_writev_func writev_func;
    CMTransport_writev_attr_func writev_attr_func;
    void *trans_data;
};

#define CManager_lock(cm) IntCManager_lock(cm, __FILE__, __LINE__)
#define CManager_unlock(cm) IntCManager_unlock(cm, __FILE__, __LINE__)
extern void IntCManager_lock ARGS((CManager cm, char *file, int line));
extern void IntCManager_unlock ARGS((CManager cm, char *file, int line));
extern int CManager_locked ARGS((CManager cm));

extern void CMControlList_lock ARGS((CMControlList cl));
extern void CMControlList_unlock ARGS((CMControlList cl));
extern int CMControlList_locked ARGS((CMControlList cl));
extern void CMConn_write_lock ARGS((CMConnection cl));
extern void CMConn_write_unlock ARGS((CMConnection cl));
extern int CMConn_write_locked ARGS((CMConnection cl));
extern void CMglobal_data_lock();
extern void CMglobal_data_unlock();
extern int CMglobal_data_locked();

typedef enum _CMTraceType {
    CMAlwaysTrace, CMControlVerbose, CMConnectionVerbose, CMLowLevelVerbose, CMDataVerbose, CMTransportVerbose, CMFormatVerbose,
    CMLastTraceType /* add before this one */
} CMTraceType;

extern void 
CMtrace_out ARGS((CManager cm, CMTraceType trace_type, char *format, ...));

extern int
CMtrace_on ARGS((CManager cm, CMTraceType trace_type));

extern void 
CMDataAvailable ARGS((transport_entry trans, CMConnection conn));

extern CMincoming_format_list
CMidentify_CMformat ARGS((CManager cm, IOFormat format));

extern void CMtransport_trace ARGS((CManager cm, char *format, ...));

extern void
CM_fd_add_select ARGS((CManager cm, int fd, select_list_func handler_func,
		       void *param1, void *param2));

extern void CM_fd_remove_select ARGS((CManager cm, int fd));

extern CMConnection
CMConnection_create ARGS((transport_entry trans, void *transport_data,
			  attr_list conn_attrs));

extern void free_CMFormat ARGS((CMFormat format));

extern void CMcomplete_format_registration ARGS((CMFormat format, int lock));
extern int CMcontrol_list_wait ARGS((CMControlList cl));
extern int CMcontrol_list_poll ARGS((CMControlList cl));
extern int load_transport ARGS((CManager cm, const char *trans_name));

extern int CMinternal_listen ARGS((CManager cm, attr_list listen_info));
extern CMConnection CMinternal_get_conn ARGS((CManager cm, attr_list attrs));
extern void CMconn_fail_conditions ARGS((CMConnection conn));
extern int CMpbio_send_format_preload ARGS((IOFormat ioformat, CMConnection conn));
extern void CMformat_preload ARGS((CMConnection conn, CMFormat format));
extern void CMinit_local_formats ARGS((CManager cm));
