#ifndef __CM_TRANSPORT_H__
#define __CM_TRANSPORT_H__

typedef void *(*CMTransport_malloc_func) ARGS((int));
typedef void *(*CMTransport_realloc_func) ARGS((void*, int));
typedef void (*CMTransport_free_func) ARGS((void*));

typedef void (*CMTransport_fd_add_select) ARGS((CManager cm, int fd, select_list_func handler_func,
		       void *param1, void *param2));
typedef void (*CMTransport_fd_remove_select) ARGS((CManager cm, int fd));
typedef void (*CMTransport_trace) ARGS((CManager cm, char *format, ...));
typedef CMConnection (*CMTransport_conn_create) ARGS((transport_entry trans,
						      void *transport_data,
						      attr_list conn_attrs));


typedef struct CMtrans_services {
    CMTransport_malloc_func malloc;
    CMTransport_realloc_func realloc;
    CMTransport_free_func free;
    CMTransport_fd_add_select fd_add_select;
    CMTransport_fd_remove_select fd_remove_select;
    CMTransport_trace trace_out;
    CMTransport_conn_create connection_create;
} *CMtrans_services;


typedef void *(*CMTransport_func) ARGS((CManager cm, CMtrans_services svc));
typedef attr_list (*CMTransport_listen_func) ARGS((CManager cm,
						   CMtrans_services svc,
						   transport_entry trans,
                                                   attr_list listen_info));
typedef void *(*CMTransport_read_func) ARGS((CMtrans_services svc,
					     void *transport_data, int len,
					     int *actual));
#ifdef MODULE
typedef int (*CMTransport_read_to_buffer_func) ARGS((CMtrans_services svc,
						     void *transport_data,
						     void *buffer,
						     int len, 
						     struct task_struct *tsk,
						     int flag));
#else
typedef int (*CMTransport_read_to_buffer_func) ARGS((CMtrans_services svc,
						     void *transport_data,
						     void *buffer,
						     int len));
#endif
typedef int (*CMTransport_write_func) ARGS((CMtrans_services svc,
					    void *transport_data,
					    void *buffer, int len));
typedef int (*CMTransport_writev_func) ARGS((CMtrans_services svc,
					     void *transport_data,
					     void *buffer, int len));
typedef int (*CMTransport_writev_attr_func) ARGS((CMtrans_services svc,
						  void *transport_data,
						  void *buffer, int len,
						  attr_list attrs));
typedef void (*CMTransport_shutdown_conn_func) ARGS((CMtrans_services svc,
						     void *conn_data));

typedef CMConnection (*CMTransport_conn_func) ARGS((CManager cm, 
						    CMtrans_services svc,
						    transport_entry trans, 
						    attr_list attrs));

typedef int (*CMTransport_self_check_func) ARGS((CManager cm,
						 CMtrans_services svc,
						 transport_entry trans,
						 attr_list attrs));

typedef int (*CMTransport_connection_eq_func) ARGS((CManager cm,
						    CMtrans_services svc,
						    transport_entry trans,
						    attr_list attrs,
						    void *conn_data));

#endif

