
#ifdef MODULE
#ifndef __KERNEL__
#define __KERNEL__
#endif
#endif

#ifndef MODULE
#include "config.h"
#include <atl.h>
#include "cm.h"
#include "cm_internal.h"

#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdio.h>
#ifdef STDC_HEADERS
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#include <errno.h>
#else
#include "config.h"
#include "../include/atl.h"
#include "cm.h"
#include "cm_internal.h"
#include "kcm.h"
#include "library.h"
#include <linux/errno.h>

#define printf         printk
#define perror         printk
#define fprintf(fmt, args...) printk (args) 
#define malloc         (void *)DAllocMM
#define free(a)        DFreeMM((addrs_t)a)
#define realloc(a,b)   (void *)DReallocMM((addrs_t)a,b)
#define exit           sys_exit
#define qsort          _quicksort
#endif

extern int vfprintf();


static int trace_val[CMLastTraceType] = {-1, -1, -1, -1, -1, -1};

extern int
CMtrace_on(CManager cm, CMTraceType trace_type)
{
    if (trace_val[0] == -1) {
	trace_val[0] = 0;
	trace_val[CMControlVerbose] = (getenv("CMControlVerbose") != NULL);
	trace_val[CMConnectionVerbose] = (getenv("CMConnectionVerbose") != NULL);
	trace_val[CMDataVerbose] = (getenv("CMDataVerbose") != NULL);
	trace_val[CMTransportVerbose] = (getenv("CMTransportVerbose") != NULL);
	trace_val[CMFormatVerbose] = (getenv("CMFormatVerbose") != NULL);
	if (getenv("CMVerbose") != NULL) {
	    int i;
	    for (i=0; i<CMLastTraceType; i++)
		trace_val[i] = 1;
	}
	/* for low level verbose, value overrides general CMVerbose */
	trace_val[CMLowLevelVerbose] = (getenv("CMLowLevelVerbose") != NULL);
    }

    return trace_val[trace_type];
}

extern void
CMtrace_out(CManager cm, CMTraceType trace_type, char *format, ...)
{
#ifndef MODULE
    va_list ap;

    if (CMtrace_on(cm, trace_type)) {
	if (CMtrace_on(cm, CMLowLevelVerbose)) {
	    printf("%lx - ", (long) getpid());
	}
#ifdef STDC_HEADERS
	va_start(ap, format);
#else
	va_start(ap);
#endif
	vfprintf(stdout, format, ap);
	va_end(ap);
	printf("\n");
    }
#endif
}

extern void
CMtransport_trace(CManager cm, char *format, ...)
{
#ifndef MODULE
    va_list ap;
    if (CMtrace_on(cm, CMTransportVerbose)) {
#ifdef STDC_HEADERS
	va_start(ap, format);
#else
	va_start(ap);
#endif
	vfprintf(stdout, format, ap);
	va_end(ap);
	printf("\n");
    }
#endif
}

extern void*
CMrealloc(ptr, size)
void *ptr;
int size;
{
    void *tmp = realloc(ptr, size);
    if ((tmp == 0) && (size != 0)) {
	perror("realloc");
    }
    return tmp;
}

extern void*
CMmalloc(size)
int size;
{
    return malloc(size);
}

extern void
CMfree(ptr)
void *ptr;
{
    free(ptr);
}

