
#ifdef MODULE
#ifndef __KERNEL__
#define __KERNEL__
#endif
#endif

#ifndef MODULE
/***** Includes *****/
#include "config.h"
#include <sys/types.h>

#ifdef HAVE_WINDOWS_H
#include <windows.h>
#include <winsock.h>
#include <sys/timeb.h>
#define getpid()	_getpid()
#else
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_TIMES_H
#include <sys/times.h>
#endif
#include <sys/socket.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_SYS_UN_H
#include <sys/un.h>
#endif
#ifdef HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif
#ifdef HAVE_HOSTLIB_H
#include "hostLib.h"
#endif
#ifdef HAVE_STREAMS_UN_H
#include <streams/un.h>
#endif
#include <netinet/in.h>
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#endif
#include <stdio.h>
#include <fcntl.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif

#include "gen_thread.h"
#include <atl.h>
#include "cm.h"
#include "cm_internal.h"
#include "cm_transport.h"

#if !defined(SELECT_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int select ARGS((int width, fd_set * readfds, fd_set * writefds,
			fd_set * exceptfds, struct timeval * timeout));
#endif
#if !defined(BIND_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int bind ARGS((int s, struct sockaddr * name, int namelen));
#endif
#if !defined(ACCEPT_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int accept ARGS((int s, struct sockaddr * addr, int *addrlen));
#endif
#if !defined(CONNECT_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int connect ARGS((int s, struct sockaddr * name, int namelen));
#endif
#if !defined(SETSOCKOPT_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int setsockopt ARGS((int s, int level, int optname, const char *optval, int optlen));
#endif
#if !defined(GETSOCKNAME_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int getsockname ARGS((int s, struct sockaddr * name, int *namelen));
#endif
#if defined(HAVE_GETDOMAINNAME) && !defined(GETDOMAINNAME_DEFINED)
extern int getdomainname ARGS((char *name, int namelen));
#endif
#if !defined(LISTEN_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int listen ARGS((int s, int backlog));
#endif
#ifndef HAVE_WINDOWS_H
extern int socket ARGS((int domain, int type, int protocol));
#endif
#if !defined(GETHOSTNAME_DEFINED) & !defined(HAVE_WINDOWS_H)
extern int gethostname ARGS((char *name, int namelen));
#endif
#else
#include "config.h"
#include <linux/types.h>
#include <linux/time.h>
#include <linux/socket.h>
#include <linux/un.h>
#include <linux/uio.h>
#include <linux/fcntl.h>
#include <linux/unistd.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <asm/uaccess.h>
#include "../include/atl.h"
#include "cm.h"
#include "cm_internal.h"
#include "cm_transport.h"
#include "kcm.h"
#include "library.h"
#include "assert.h"

#define MY_ARRAY_SIZE     8

#define printf         printk
#define perror         printk
#define fprintf(fmt, args...) printk (args) 
#define exit           sys_exit
#define qsort          _quicksort
#define getpid()       current->pid
#define signal(a,b)    
#define gethostname    sys_gethostname
#define gethostbyname  lookup_name
#define select         sys_select
#define gettimeofday(a,b)  do_gettimeofday(a)
#define wake_server_thread(a)

extern int register_sock_interest(int fd, void *ptr);
extern int unregister_sock_interest(int fd);
#endif

#ifndef SOCKET_ERROR
#define SOCKET_ERROR -1
#endif

typedef struct func_list_item {
    select_list_func func;
    void *arg1;
    void *arg2;
} FunctionListElement;

typedef struct select_data {
    thr_thread_t server_thread;

#ifndef MODULE
    void *fdset;		/* bitmap of the fds for read select */
#endif
    int 	sel_item_max;
    FunctionListElement *select_items;

    periodic_task_handle periodic_task_list;

    int		closed;
    int 	select_consistency_number;
    int 	wake_read_fd;
    int 	wake_write_fd;
    char 	*hostname;
    int		listen_port;
#ifdef MODULE
    int         loop;
    int         event;
    int		listen_fd;
    char fdset[FD_SETSIZE];
    CManager cm;
#endif
} *select_data_ptr;

#ifndef MODULE
static void wake_server_thread ARGS((select_data_ptr socket_data));
static void setup_wake_mechanism ARGS((CMtrans_services svcs,
				       select_data_ptr *sdp));
#endif
static int remove_periodic_task ARGS((select_data_ptr sd,
				      periodic_task_handle handle));

#ifdef HAVE_WINDOWS_H
/* Winsock init stuff  */
/* ask for ver 1.1 */
static WORD wVersionRequested = MAKEWORD(1, 1);
static WSADATA wsaData;
int nErrorStatus;
static char*WSAerror_str ARGS((int err));
#define FD_SETSIZE 1024
#endif

static void
#ifdef MODULE
init_select_data(svc, cm, sdp)
CManager cm;
#else
init_select_data(svc, sdp)
#endif
CMtrans_services svc;
select_data_ptr *sdp;
{
#ifdef MODULE
    int i;

    select_data_ptr sd = (select_data_ptr)DAllocMM(sizeof(struct select_data));
#else
    select_data_ptr sd = malloc(sizeof(struct select_data));
#endif
    *sdp = sd;
#ifdef MODULE
    for (i = 0; i < FD_SETSIZE; i++)
      sd->fdset[i] = 0;
    sd->cm = cm;
#else
    sd->fdset = svc->malloc(sizeof(fd_set));
    FD_ZERO((fd_set *) sd->fdset);
#endif
    sd->server_thread = NULL;
    sd->closed = 0;
    sd->sel_item_max = 0;
    sd->select_items = (FunctionListElement *) svc->malloc(sizeof(FunctionListElement));
    sd->select_items[0].func = NULL;
    sd->select_items[0].arg1 = NULL;
    sd->select_items[0].arg2 = NULL;
    
    sd->periodic_task_list = NULL;
    sd->select_consistency_number = 0;
    sd->wake_read_fd = -1;
    sd->wake_write_fd = -1;
#ifndef MODULE
    setup_wake_mechanism(svc, sdp);
#endif
}



typedef struct _periodic_task {
    int period_sec;
    int period_usec;
    thr_thread_t executing;
    struct timeval next_time;
    select_list_func func;
    void *arg1;
    void *arg2;
    periodic_task_handle next;
} task_handle_s;

#undef timercmp
#define	timercmp(tvp, uvp, cmp) \
	/* CSTYLED */ \
	(((tvp)->tv_sec cmp (uvp)->tv_sec) || \
	((((tvp)->tv_sec == (uvp)->tv_sec) && \
	/* CSTYLED */ \
	((tvp)->tv_usec cmp (uvp)->tv_usec))))

static void
set_soonest_timeout(timeout, task_list, now)
struct timeval *timeout;
periodic_task_handle task_list;
struct timeval now;
{
    struct timeval this_delay;
    if (task_list == NULL) return;
    this_delay.tv_sec = task_list->next_time.tv_sec - now.tv_sec;
    this_delay.tv_usec = task_list->next_time.tv_usec - now.tv_usec;
    if (this_delay.tv_usec < 0) {
	this_delay.tv_sec--;
	this_delay.tv_usec += 1000000;
    }
    if (this_delay.tv_sec < 0) {
	this_delay.tv_sec = this_delay.tv_usec = 0;
    }
    if ((timeout->tv_sec == -1) || (timercmp(&this_delay, timeout, <))) {
	*timeout = this_delay;
    }
    set_soonest_timeout(timeout, task_list->next, now);
}

static void
increment_time(time, increment_sec, increment_usec)
struct timeval *time;
int increment_sec;
int increment_usec;
{
    time->tv_usec += increment_usec;
    time->tv_sec += increment_sec;
    if (time->tv_usec >= 1000000) {
	time->tv_sec += (time->tv_usec / 1000000);
	time->tv_usec = (time->tv_usec % 1000000);
    }
}

static void
socket_select(sd, timeout_sec, timeout_usec)
select_data_ptr sd;
int timeout_sec;
int timeout_usec;
{
    int i, res;
#ifndef MODULE
    fd_set rd_set;
#endif
    struct timeval timeout;
    int tmp_select_consistency_number = sd->select_consistency_number;

    if (sd->closed) return;

#ifndef MODULE
    if (sd->server_thread == NULL) {
	/* no server thread set, must be this one */
	sd->server_thread = thr_thread_self();
    }
    if (sd->server_thread != thr_thread_self()) {
	/* What?  We're polling, but we're not the server thread? */
	fprintf(stderr, "Warning:  Multiple threads calling CManager_socket_select.\n");
	fprintf(stderr, "          This situation may result in unexpected I/O blocking.\n");
	fprintf(stderr, "          Server thread set to %lx.\n", (long) thr_thread_self());
	sd->server_thread = thr_thread_self();
    }
    rd_set = *(fd_set *) sd->fdset;
#endif
    if ((timeout_sec >= 0) || (sd->periodic_task_list != NULL)) {
	struct timeval now;
#ifndef HAVE_WINDOWS_H
	gettimeofday(&now, NULL);
#else
	/* GSE...  No gettimeofday on windows.  
	 * Must use _ftime, get millisec time, convert to usec.  Bleh.
	 */
	struct _timeb nowb;
	_ftime(&nowb);
	now.tv_sec = nowb.time;
	now.tv_usec = nowb.millitm * 1000;
#endif
	if (timeout_usec >= 1000000) {
	    timeout_sec += (timeout_usec / 1000000);
	    timeout_usec = timeout_usec % 1000000;
	}
	timeout.tv_sec = timeout_sec;
	timeout.tv_usec = timeout_usec;
#ifdef MODULE
	timeout.tv_sec = 0;
	timeout.tv_usec = 0;
	
        res = 1;
#else
	set_soonest_timeout(&timeout, sd->periodic_task_list, now);
	res = select(FD_SETSIZE, &rd_set, (fd_set *) NULL, (fd_set *) NULL,
		     &timeout);
#endif
    } else {
#ifdef MODULE
        res = 1;
#else
	res = select(FD_SETSIZE, &rd_set, (fd_set *) NULL, (fd_set *) NULL,
		     NULL);
#endif
    }
    if (sd->closed) return;
#ifndef HAVE_WINDOWS_H
    if (res == -1) {
	if (errno == EINTR) {
	    return;
	}
	/* 
	 * if upon returning from select the consistency number
	 * has changed, the only safe thing to do is return so 
	 * that a new select can happen with consistent info.
	 */
	if (sd->select_consistency_number != 
	    tmp_select_consistency_number) return;
	if (errno == 0) {
	    /* 
	     * odd, but x86 Solaris seems to return -1 from select but not set
	     * the errno in some circumstances.  Just return when this happens.
	     */
	    return;
	}
	if (errno == EBADF) {
	    int i;
	    int found_one = 0;
	    for (i = 0; i < FD_SETSIZE; i++) {
#ifdef MODULE
		if (my_FD_ISSET(i, (char *)sd->fdset)) {
		    char test_set[my_FD_SETSIZE/MY_ARRAY_SIZE];
#else
		if (FD_ISSET(i, &rd_set)) {
		    fd_set test_set;
#endif
		    timeout.tv_usec = 0;
		    timeout.tv_sec = 0;
#ifdef MODULE
		    my_FD_SET(i, (char *)test_set);
		    res = 1;
#else
		    FD_ZERO(&test_set);
		    FD_SET(i, &test_set);
#endif
		    errno = 0;
#ifndef MODULE
		    select(FD_SETSIZE, &test_set, (fd_set *) NULL,
			   (fd_set *) NULL, &timeout);
#endif
		    if (errno == EBADF) {
			fprintf(stderr, "Select failed, fd %d is bad.  Removing from select list.\n",
				i);
#ifdef MODULE
			my_FD_CLR(i, (char *) sd->fdset);
			found_one++;
#else
			FD_CLR(i, (fd_set *) sd->fdset);
			found_one++;
			FD_CLR(i, &rd_set);
#endif
		    }
		}
	    }
	    if (sd->closed) return;
/* if (found_one == 0) { fprintf(stderr, "Bad file descriptor in select
 * Warning.  Failed to localize.\n"); } */
	} else {
#ifdef HAVE_FDS_BITS
	    fprintf(stderr, "select failed, errno %d, rd_set was %lx, %lx,%lx, %lx\n\n", errno,
		    (long) ((fd_set *) sd->fdset)->fds_bits[0],
		    (long) ((fd_set *) sd->fdset)->fds_bits[1],
		    (long) ((fd_set *) sd->fdset)->fds_bits[2],
		    (long) ((fd_set *) sd->fdset)->fds_bits[3]);
#else
	    fprintf(stderr, "seelct failed, errno %d\n", errno);
#endif
	    exit(1);
	    return;
	}
    }
#else
    /* 
     * Have to ignore the bad invalue for select on NT because it can't
     * do selects on files.  Otherwise you will get a bunch of
     * irritating select errors that aren't really errors 
     */
    if (res == SOCKET_ERROR) {
	int errno_val;
	errno_val = WSAGetLastError();
	if (errno_val == WSAEINTR || errno_val == WSAEINVAL) {
	    return;
	} else {
	    fprintf(stderr, "select failed, errno %d\n", 
		    WSAerror_str(errno_val));
	}
	return;
    }
#endif

    /* 
     * if upon returning from select the consistency number
     * has changed, the only safe thing to do is return so 
     * that a new select can happen with consistent info.
     */
    if (sd->select_consistency_number != 
	tmp_select_consistency_number) return;
    /* 
     * Careful!  We're reading the control list here without locking!
     * Something bad *might* happen, it's just unlikely.
     */
    if (res != 0) {
	for (i = 0; i <= sd->sel_item_max; i++) {
	    if (sd->closed) return;
#ifdef MODULE
	    if (my_FD_ISSET(i, sd->fdset)) {
#else
	    if (FD_ISSET(i, &rd_set)) {
#endif
		if (sd->select_items[i].func != NULL) {
#ifdef MODULE
		    do { 
		    sd->loop = 0;
#endif
		    sd->select_items[i].func(sd->select_items[i].arg1,
					     sd->select_items[i].arg2);
#ifdef MODULE
		    } while (sd->loop == 1); 
		    if ((i != sd->listen_fd) && (sd->event == 1))
		        my_FD_CLR(i, sd->fdset);
#endif
		} 
#ifndef MODULE
                else {
		    assert(!FD_ISSET(i, (fd_set *)sd->fdset));
		}
		if (sd->select_consistency_number != 
		    tmp_select_consistency_number) return;
#endif
	    }
	}
    }
#ifndef MODULE
    if (sd->periodic_task_list != NULL) {
	/* handle periodic tasks */
	periodic_task_handle this_periodic_task = sd->periodic_task_list;
	struct timeval now;

#ifndef HAVE_WINDOWS_H
	gettimeofday(&now, NULL);
#else
	/* GSE...  No gettimeofday on windows.  
	 * Must use _ftime, get millisec time, convert to usec.  Bleh.
	 */
	struct _timeb nowb;
	_ftime(&nowb);
	now.tv_sec = nowb.time;
	now.tv_usec = nowb.millitm * 1000;
#endif
	while (this_periodic_task != NULL ) {
	    
	    if (timercmp(&now, &this_periodic_task->next_time, >)) {
		increment_time(&this_periodic_task->next_time,
			       this_periodic_task->period_sec,
			       this_periodic_task->period_usec);
		this_periodic_task->executing = thr_thread_self();
		this_periodic_task->func(this_periodic_task->arg1,
					 this_periodic_task->arg2);
		this_periodic_task->executing = (thr_thread_t) -1;
		if ((this_periodic_task->period_sec == 0) &&
		    (this_periodic_task->period_usec == 0)) {
		    remove_periodic_task(sd, this_periodic_task);
		}		    
	    }
	    /* 
	     * if upon returning from a handler the consistency number
	     * has changed, the only safe thing to do is return so 
	     * that a new select can happen with consistent info.
	     */
	    if (sd->select_consistency_number != 
		tmp_select_consistency_number) return;
	    this_periodic_task = this_periodic_task->next;
	}
    }
#endif
    sd->select_consistency_number++;
}

extern void
#ifdef MODULE
libcmselect_LTX_add_select(cm, svc, sdp, fd, func, arg1, arg2)
CManager cm;
#else
libcmselect_LTX_add_select(svc, sdp, fd, func, arg1, arg2)
#endif
CMtrans_services svc;
select_data_ptr *sdp;
int fd;
select_list_func func;
void *arg1;
void *arg2;
{
#ifdef MODULE
    int listen_flag = 0;
#endif
    select_data_ptr sd = *((select_data_ptr *)sdp);
    if (sd == NULL) {
#ifdef MODULE
	init_select_data(svc, cm, (select_data_ptr*)sdp);
#else
	init_select_data(svc, (select_data_ptr*)sdp);
#endif
	sd = *((select_data_ptr *)sdp);
    }
    sd->select_consistency_number++;
#ifdef MODULE
    if (fd >= FD_SETSIZE) {
        fd -= FD_SETSIZE;
        listen_flag = 1;
    }
#endif
    if (fd > sd->sel_item_max) {
	int i;
	int size = (fd+1)*sizeof(sd->select_items[0]);
	sd->select_items = 
	    (FunctionListElement *) svc->realloc(sd->select_items, size);
	if (sd->select_items == NULL) {
	    perror("Realloc failed\n");
	    exit(1);
	}

	for (i = sd->sel_item_max + 1; i < fd; i++) {
	    sd->select_items[i].func = NULL;
	    sd->select_items[i].arg1 = NULL;
	    sd->select_items[i].arg2 = NULL;
	}
	sd->sel_item_max = fd;
    }
#ifdef MODULE
    my_FD_SET(fd, (char *) sd->fdset);
    if (listen_flag != 1)
        register_sock_interest(fd, (void *)sd);
    else  
        sd->listen_fd = fd;
    if (fd > my_FD_SETSIZE) {
#else
    FD_SET(fd, (fd_set *) sd->fdset);
    if (fd > FD_SETSIZE) {
#endif
	fprintf(stderr, "Internal Error, stupid WINSOCK large FD bug.\n");
	fprintf(stderr, "Increase FD_SETSIZE.  Item not added to fdset.\n");
    }
    sd->select_items[fd].func = func;
    sd->select_items[fd].arg1 = arg1;
    sd->select_items[fd].arg2 = arg2;
    wake_server_thread(sd);
}

extern periodic_task_handle
#ifdef MODULE
libcmselect_LTX_add_periodic(cm, svc, sdp, interval_sec, interval_usec,
			     func, arg1, arg2)
CManager cm;
#else
libcmselect_LTX_add_periodic(svc, sdp, interval_sec, interval_usec,
			     func, arg1, arg2)
#endif
CMtrans_services svc;
select_data_ptr *sdp;
int interval_sec;
int interval_usec;
select_list_func func;
void *arg1;
void *arg2;
{
    select_data_ptr sd = *((select_data_ptr *)sdp);
#ifdef MODULE
    periodic_task_handle handle = (periodic_task_handle) 
              DAllocMM(sizeof(struct _periodic_task));   
#else
    periodic_task_handle handle = malloc(sizeof(struct _periodic_task));
#endif
    if (sd == NULL) {
#ifdef MODULE
	init_select_data(svc, cm, (select_data_ptr*)sdp);
#else
	init_select_data(svc, (select_data_ptr*)sdp);
#endif
	sd = *((select_data_ptr *)sdp);
    }

    handle->period_sec = interval_sec;
    handle->period_usec = interval_usec;
    handle->executing = (thr_thread_t) -1;
#ifndef HAVE_WINDOWS_H
    gettimeofday(&handle->next_time, NULL);
#else
	/* GSE...  No gettimeofday on windows.  
	 * Must use _ftime, get millisec time, convert to usec.  Bleh.
	 */
    {
	struct _timeb nowb;
	_ftime(&nowb);
	handle->next_time.tv_sec = nowb.time;
	handle->next_time.tv_usec = nowb.millitm * 1000;
    }
#endif
    increment_time(&handle->next_time, interval_sec, interval_usec);
    handle->func = func;
    handle->arg1 = arg1;
    handle->arg2 = arg2;
    handle->next = NULL;

    if (sd->periodic_task_list == NULL) {
	sd->periodic_task_list = handle;
    } else {
	handle->next = sd->periodic_task_list;
	sd->periodic_task_list = handle;
    }
    wake_server_thread(sd);
    return handle;
}

extern periodic_task_handle
#ifdef MODULE
libcmselect_LTX_add_delayed_task(cm, svc, sdp, delay_sec, delay_usec, 
				 func, arg1, arg2)
CManager cm;
#else
libcmselect_LTX_add_delayed_task(svc, sdp, delay_sec, delay_usec, 
				 func, arg1, arg2)
#endif
CMtrans_services svc;
select_data_ptr *sdp;
int delay_sec;
int delay_usec;
select_list_func func;
void *arg1;
void *arg2;
{
    select_data_ptr sd = *((select_data_ptr *)sdp);
#ifdef MODULE
    periodic_task_handle handle = (periodic_task_handle) 
            DAllocMM(sizeof(struct _periodic_task));
#else
    periodic_task_handle handle = malloc(sizeof(struct _periodic_task));
#endif
    if (sd == NULL) {
#ifdef MODULE
	init_select_data(svc, cm, (select_data_ptr*)sdp);
#else
	init_select_data(svc, (select_data_ptr*)sdp);
#endif
	sd = *((select_data_ptr *)sdp);
    }

    handle->period_sec = 0;
    handle->period_usec = 0;
    handle->executing = (thr_thread_t) -1;
#ifndef HAVE_WINDOWS_H
    gettimeofday(&handle->next_time, NULL);
#else
    {
	/* GSE...  No gettimeofday on windows.  
	 * Must use _ftime, get millisec time, convert to usec.  Bleh.
	 */
	struct _timeb nowb;
	_ftime(&nowb);
	handle->next_time.tv_sec = nowb.time;
	handle->next_time.tv_usec = nowb.millitm * 1000;
    }
#endif
    increment_time(&handle->next_time, delay_sec, delay_usec);
    handle->func = func;
    handle->arg1 = arg1;
    handle->arg2 = arg2;
    handle->next = NULL;

    if (sd->periodic_task_list == NULL) {
	sd->periodic_task_list = handle;
    } else {
	handle->next = sd->periodic_task_list;
	sd->periodic_task_list = handle;
    }
    wake_server_thread(sd);
    return handle;
}

static int
remove_periodic_task(sd, handle)
select_data_ptr sd;
periodic_task_handle handle;
{
    periodic_task_handle list, last = NULL;
    list = sd->periodic_task_list;
    
    while(list != handle) {
	last = list;
	list = list->next;
	if (list == NULL) {
	    return 0;
	}
    }
    /* unlink task */
    if (last == NULL) {
	sd->periodic_task_list = list->next;
    } else {
	last->next = list->next;
    }
#ifndef MODULE
    if (handle->executing != thr_thread_self()) {
	/* someone besides us executing this ? */
        int i = 0;
	while (handle->executing != (thr_thread_t)-1) {
	    /* wait until they're done */
	    thr_thread_yield();
	    i++;
	    if (i > 1000) {
	        /* give up */
	        continue;
	    }
	}
    }
    free(handle);
#else
    DFreeMM((addrs_t)handle);
#endif
    sd->select_consistency_number++;
    return 1;
}


extern void
libcmselect_LTX_remove_periodic(svc, sdp, handle)
CMtrans_services svc;
select_data_ptr *sdp;
periodic_task_handle handle;
{
    select_data_ptr sd = *((select_data_ptr *)sdp);
    if (remove_periodic_task(sd, handle) == 0) {
	fprintf(stderr, "Periodic task not found for removal\n");
    }
}

extern void
#ifdef MODULE
libcmselect_LTX_remove_select(cm, svc, sdp, fd)
CManager cm;
#else
libcmselect_LTX_remove_select(svc, sdp, fd)
#endif
CMtrans_services svc;
select_data_ptr *sdp;
int fd;
{
    select_data_ptr sd = *((select_data_ptr *)sdp);
    if (sd == NULL) {
#ifdef MODULE
	init_select_data(svc, cm, (select_data_ptr*)sdp);
#else
	init_select_data(svc, (select_data_ptr*)sdp);
#endif
	sd = *((select_data_ptr *)sdp);
    }
    sd->select_consistency_number++;
#ifdef MODULE
/*
    my_FD_CLR(fd, (char *) sd->fdset);
*/
    unregister_sock_interest(fd);
#else
    FD_CLR(fd, (fd_set *) sd->fdset);
#endif
    sd->select_items[fd].func = NULL;
    sd->select_items[fd].arg1 = NULL;
    sd->select_items[fd].arg2 = NULL;
    wake_server_thread(sd);
}

#ifndef MODULE
/*static void
shutdown_wake_mechanism(sd)
select_data_ptr sd;
{
    if (sd->wake_read_fd == -1) return;
    wake_server_thread(sd);
    close(sd->wake_read_fd);
    close(sd->wake_write_fd);
    sd->wake_read_fd = sd->wake_write_fd = -1;
}*/

static void read_wake_fd(fd_as_ptr, junk)
void *fd_as_ptr;
void *junk;
{
    char buffer;
    int fd = (int) (long)fd_as_ptr;
#ifdef HAVE_WINDOWS_H
    recv(fd, &buffer, 1, 0);
#else
    read(fd, &buffer, 1);
#endif
}

#ifdef HAVE_WINDOWS_H
static char*
WSAerror_str(err)
int err;
{
    switch(err) {
    case WSAEINTR: return "WSAEINTR";
    case WSAEBADF: return "WSAEBADF";
    case WSAEACCES: return "WSAEACCES";
    case WSAEFAULT: return "WSAEFAULT";
    case WSAEINVAL: return "WSAEINVAL";
    case WSAEMFILE: return "WSAEMFILE";
    case WSAEWOULDBLOCK: return "WSAEWOULDBLOCK";
    case WSAEINPROGRESS: return "WSAEINPROGRESS";
    case WSAEALREADY: return "WSAEALREADY";
    case WSAENOTSOCK: return "WSAENOTSOCK";
    case WSAEDESTADDRREQ: return "WSAEDESTADDRREQ";
    case WSAEMSGSIZE: return "WSAEMSGSIZE";
    case WSAEPROTOTYPE: return "WSAEPROTOTYPE";
    case WSAENOPROTOOPT: return "WSAENOPROTOOPT";
    case WSAEPROTONOSUPPORT: return "WSAEPROTONOSUPPORT";
    case WSAESOCKTNOSUPPORT: return "WSAESOCKTNOSUPPORT";
    case WSAEOPNOTSUPP: return "WSAEOPNOTSUPP";
    case WSAEPFNOSUPPORT: return "WSAEPFNOSUPPORT";
    case WSAEAFNOSUPPORT: return "WSAEAFNOSUPPORT";
    case WSAEADDRINUSE: return "WSAEADDRINUSE";
    case WSAEADDRNOTAVAIL: return "WSAEADDRNOTAVAIL";
    case WSAENETDOWN: return "WSAENETDOWN";
    case WSAENETUNREACH: return "WSAENETUNREACH";
    case WSAENETRESET: return "WSAENETRESET";
    case WSAECONNABORTED: return "WSAECONNABORTED";
    case WSAECONNRESET: return "WSAECONNRESET";
    case WSAENOBUFS: return "WSAENOBUFS";
    case WSAEISCONN: return "WSAEISCONN";
    case WSAENOTCONN: return "WSAENOTCONN";
    case WSAESHUTDOWN: return "WSAESHUTDOWN";
    case WSAETOOMANYREFS: return "WSAETOOMANYREFS";
    case WSAETIMEDOUT: return "WSAETIMEDOUT";
    case WSAECONNREFUSED: return "WSAECONNREFUSED";
    case WSAELOOP: return "WSAELOOP";
    case WSAENAMETOOLONG: return "WSAENAMETOOLONG";
    case WSAEHOSTDOWN: return "WSAEHOSTDOWN";
    case WSAEHOSTUNREACH: return "WSAEHOSTUNREACH";
    case WSAENOTEMPTY: return "WSAENOTEMPTY";
    case WSAEPROCLIM: return "WSAEPROCLIM";
    case WSAEUSERS: return "WSAEUSERS";
    case WSAEDQUOT: return "WSAEDQUOT";
    case WSAESTALE: return "WSAESTALE";
    case WSAEREMOTE: return "WSAEREMOTE";
    case WSAEDISCON: return "WSAEDISCON";
    case WSASYSNOTREADY: return "WSASYSNOTREADY";
    case WSAVERNOTSUPPORTED: return "WSAVERNOTSUPPORTED";
    case WSANOTINITIALISED: return "WSANOTINITIALISED";
    default: return "Unknown Winsock error";
    }
}
/*
 *  Note.  Unfortunately, the _pipe() function on WinNT 
 *  produces FDs that you can't use in select().  This ruins what we want
 *  this pipe for, which is to wake up a thread sleeping in select().
 *  So, we need to introduce a pipe function that returns two socket FDs.
 *  NT Sux.
 */

int
pipe(filedes)
int filedes[2];
{
    
    int length;
    struct sockaddr_in sock_addr;
    int sock_opt_val = 1;
    int sock1, sock2, conn_sock;
    unsigned long block = TRUE;
    int delay_value = 1;
   
    conn_sock = socket(AF_INET, SOCK_STREAM, 0);
    if (conn_sock == SOCKET_ERROR) {
	fprintf(stderr, "Cannot open INET socket\n");
	return -1;
    }
    sock_addr.sin_family = PF_INET;
    sock_addr.sin_addr.s_addr = INADDR_ANY;
    sock_addr.sin_port = 0;
    if (bind(conn_sock, (struct sockaddr *) &sock_addr,
	     sizeof sock_addr) == SOCKET_ERROR) {
	fprintf(stderr, "Cannot bind INET socket\n");
	return -1;
    }
    length = sizeof sock_addr;
    if (getsockname(conn_sock, (struct sockaddr *) &sock_addr, &length) < 0) {
	fprintf(stderr, "Cannot get socket name\n");
	return -1;
    }
    /* begin listening for conns */
    if (listen(conn_sock, FD_SETSIZE)) {
	fprintf(stderr, "listen failed\n");
	return -1;
    }

/* send sock */
    if ((sock1 = socket(AF_INET, SOCK_STREAM, 0)) == SOCKET_ERROR) {
	return -1;
    }
    sock_addr.sin_addr.s_addr = 0x0100007f;  /* loopback */
    sock_addr.sin_family = PF_INET;
    if (ioctlsocket(sock1, FIONBIO, &block) != 0) {
	printf("ioctl failed\n");
    }
    if (connect(sock1, (struct sockaddr *) &sock_addr,
		sizeof sock_addr) == SOCKET_ERROR) {
	int err = WSAGetLastError();
	if (err != WSAEWOULDBLOCK) {
	    printf("unexpected error from connect, %s\n", WSAerror_str(err));
	}
    }

    if ((sock2 = accept(conn_sock, (struct sockaddr *) 0, (int *) 0)) == SOCKET_ERROR) {
	    int err = WSAGetLastError();
	    printf("err was %s\n", WSAerror_str(err));
    }
    
    setsockopt(sock2, IPPROTO_TCP, TCP_NODELAY, (char *) &delay_value,
	       sizeof(delay_value));
    {
	fd_set stXcptFDS,stWriteFDS;
	struct timeval stTimeOut;	/* for select() timeout (none) */
	int wRet;

	FD_ZERO((fd_set FAR*)&(stXcptFDS));
	FD_ZERO((fd_set FAR*)&(stWriteFDS));
	FD_SET(sock1, (fd_set FAR*)&(stWriteFDS));
	FD_SET(sock1, (fd_set FAR*)&(stXcptFDS));
	stTimeOut.tv_sec  = 10;
	stTimeOut.tv_usec = 0;
	wRet = select(-1, NULL, 
		      (fd_set FAR*)&(stWriteFDS),
		      (fd_set FAR*)&(stXcptFDS), 
		      NULL);
	if (wRet == SOCKET_ERROR) {
	    int err = WSAGetLastError();
	    printf("err was %s\n", WSAerror_str(err));
	}
    }
    setsockopt(sock1, IPPROTO_TCP, TCP_NODELAY, (char *) &delay_value,
	       sizeof(delay_value));

    filedes[0] = sock1;
    filedes[1] = sock2;
    return 0;
}
#endif

static void
setup_wake_mechanism(svc, sdp)
CMtrans_services svc;
select_data_ptr *sdp;
{
    int filedes[2];

    select_data_ptr sd = *((select_data_ptr *)sdp);
    if (sd->wake_read_fd != -1) return;
    if (pipe(filedes) != 0) {
	perror("Pipe for wake not created.  Wake mechanism inoperative.");
	return;
    }
    sd->wake_read_fd = filedes[0];
    sd->wake_write_fd = filedes[1];
    libcmselect_LTX_add_select(svc, sdp, sd->wake_read_fd, read_wake_fd, 
			       (void*)sd->wake_read_fd, NULL);
}
#endif

#ifndef MODULE
static void
wake_server_thread(sd)
select_data_ptr sd;
{
    static char buffer = 'W';  /* doesn't matter what we write */
    if (sd->wake_write_fd != -1) {
#ifdef HAVE_WINDOWS_H
	send(sd->wake_write_fd, &buffer, 1, 0);
#else
	write(sd->wake_write_fd, &buffer, 1);
#endif
    }
}
#endif

extern void
libcmselect_LTX_blocking_function(svc, client_data)
CMtrans_services svc;
void *client_data;
{
    select_data_ptr sd = *((select_data_ptr *)client_data);
#ifndef MODULE
    if (sd == NULL) {
	init_select_data(svc, (select_data_ptr*)client_data);
	sd = *((select_data_ptr *)client_data);
    }
#endif
    socket_select(sd, -1, 0); /* no timeout */
}

extern void
libcmselect_LTX_polling_function(svc, client_data)
CMtrans_services svc;
void *client_data;
{
    select_data_ptr sd = *((select_data_ptr *)client_data);
#ifndef MODULE
    if (sd == NULL) {
	init_select_data(svc, (select_data_ptr*)client_data);
	sd = *((select_data_ptr *)client_data);
    }
#endif
    socket_select(sd, 0, 0); /* no blocking, immediate timeout */
}

extern void
#ifdef MODULE
libcmselect_LTX_select_initialize(cm, svc, client_data)
CManager cm;
#else
libcmselect_LTX_select_initialize(svc, client_data)
#endif
CMtrans_services svc;
void *client_data;
{
    if (*((select_data_ptr *)client_data) == NULL) {
#ifdef MODULE
	init_select_data(svc, cm, (select_data_ptr*)client_data);
#else
	init_select_data(svc, (select_data_ptr*)client_data);
#endif
    }
}

