#include "kcm.h"

char **__environ = NULL;

char *
__findenv(name, offset)
	register const char *name;
	int *offset;
{
	register int len;
	register char **P, *C;
	register const char *tmp;

	for (tmp = name, len = 0; *tmp && *tmp != '='; ++tmp, ++len);
	for (P = __environ; *P; ++P)
		if (!strncmp(*P, name, len))
			if (*(C = *P + len) == '=') {
				*offset = P - __environ;
				return(++C);
			}
	return(NULL);
}

/* Return the value of the environment variable NAME.  */
char *
getenv (name)
     const char *name;
{
  const size_t len = strlen (name);
  char **ep;

  if (__environ == NULL)
    return NULL;

  for (ep = __environ; *ep != NULL; ++ep)
    if (!strncmp (*ep, name, len) && (*ep)[len] == '=')
      return &(*ep)[len + 1];

  return NULL;
}

/*
 * setenv --
 *	Set the value of the environmental variable "name" to be
 *	"value".  If rewrite is set, replace any current value.
 */
int
setenv(name, value, rewrite)
	register const char *name;
	register const char *value;
	int rewrite;
{
	static int alloced;			/* if allocated space before */
	register char *C;
	int l_value, offset;
	char *__findenv();

	down(&cm_sem);
	if (*value == '=')			/* no `=' in value */
		++value;
	l_value = strlen(value);
	if ((C = __findenv(name, &offset))) {	/* find if already exists */
		if (!rewrite) {
			up(&cm_sem);
			return (0);
		}
		if (strlen(C) >= l_value) {	/* old larger; copy over */
			while ((*C++ = *value++));
			up(&cm_sem);
			return (0);
		}
	} else {					/* create new slot */
		register int	cnt;
		register char	**P;

		for (P = __environ, cnt = 0; *P; ++P, ++cnt);
		if (alloced) {			/* just increase size */
			__environ = (char **)DReallocMM((char *)__environ,
			    (size_t)(sizeof(char *) * (cnt + 2)));
			if (!__environ) {
				up(&cm_sem);
				return (-1);
			}
		}
		else {				/* get new space */
			alloced = 1;		/* copy old entries into it */
			P = (char **)DAllocMM((size_t)(sizeof(char *) *
			    (cnt + 2)));
			if (!P) {
				up(&cm_sem);
				return (-1);
			}
			memcpy(__environ, P, cnt * sizeof(char *));
			__environ = P;
		}
		__environ[cnt + 1] = NULL;
		offset = cnt;
	}
	for (C = (char *)name; *C && *C != '='; ++C);	/* no `=' in name */
	if (!(__environ[offset] =			/* name + `=' + value */
	    DAllocMM((size_t)((int)(C - name) + l_value + 2)))) {
		up(&cm_sem);
		return (-1);
	}
	for (C = __environ[offset]; (*C = *name++) && *C != '='; ++C)
		;
	for (*C++ = '='; (*C++ = *value++); )
		;
	up(&cm_sem);
	return (0);
}

