static int
get_self_ip_addr(svc)
CMtrans_services svc;
{
    struct hostent *host;
    char buf[256];
    char **p;
#ifdef SIOCGIFCONF
    char *ifreqs;
    struct ifreq* ifr;
    struct sockaddr_in *sai;
    struct ifconf ifaces;
    int ifrn;
    int ss;
#ifdef MODULE
    struct socket *socket;
#endif
#endif
    int rv = 0;
    gethostname(buf, sizeof(buf));
#ifdef MODULE
    host = lookup_name(buf);
#else
    host = gethostbyname(buf);
#endif
    if (host == NULL) return 0;
    for (p=host->h_addr_list; *p !=0; p++) {
	struct in_addr *in = *(struct in_addr **)p;
	if (ntohl(in->s_addr) != INADDR_LOOPBACK) {
	    svc->trace_out(NULL, "CMSocket - Get self IP addr %lx, net %d.%d.%d.%d", 
			   ntohl(in->s_addr), 
			   *((unsigned char *)&in->s_addr),
			   *(((unsigned char *)&in->s_addr) + 1),
			   *(((unsigned char *)&in->s_addr) + 2),
			   *(((unsigned char *)&in->s_addr) + 3));
	    return (ntohl(in->s_addr));
	}
    }
#ifdef MODULE
    DFreeMM((addrs_t)host->h_name);
    DFreeMM((addrs_t)host->h_addr_list);
    DFreeMM((addrs_t)host);
#endif
    /*
     *  Since we couldn't find an IP address in some logical way, we'll open
     *  a DGRAM socket and ask it first for the list of its interfaces, and
     *  then checking for an interface we can use, and then finally asking that
     *  interface what its address is.
     */
#ifdef SIOCGIFCONF
#ifdef MODULE
    sock_create(AF_INET, SOCK_DGRAM, 0, &socket);
    ss = get_pseudofd(socket, (struct socket **)current->qfabric->fdqfabric);
#else
    ss = socket (AF_INET, SOCK_DGRAM, 0);
#endif
    ifaces.ifc_len = 64 * sizeof (struct ifreq);
#ifdef MODULE
    ifaces.ifc_buf = ifreqs = DAllocMM (ifaces.ifc_len);
#else
    ifaces.ifc_buf = ifreqs = malloc (ifaces.ifc_len);
#endif
    /*
     *  if we can't SIOCGIFCONF we're kind of dead anyway, bail.
     */
#ifdef MODULE
    if (ioctl (ss, SIOCGIFCONF, (unsigned long)&ifaces) >= 0) {
#else
    if (ioctl (ss, SIOCGIFCONF, &ifaces) >= 0) {
#endif
        ifr = ifaces.ifc_req;
        ifrn = ifaces.ifc_len / sizeof (struct ifreq);
        for (; ifrn--; ifr++) {
            /*
             *  Basically we'll take the first interface satisfying the following:
             *  up, running, not loopback, address family is INET.
             */
#ifdef MODULE
            ioctl (ss, SIOCGIFFLAGS, (unsigned long)ifr);
#else
            ioctl (ss, SIOCGIFFLAGS, ifr);
#endif
            if (ifr->ifr_flags & IFF_LOOPBACK) continue;
            if (!(ifr->ifr_flags & IFF_UP)) continue;
            if (!(ifr->ifr_flags & IFF_RUNNING)) continue;
            sai = (struct sockaddr_in *) &(ifr->ifr_addr);
            /*
               sure would be nice to test for AF_INET here but it doesn't
               cooperate and I've done enough for now ...
               if (sai->sin_addr.s.addr != AF_INET) continue;
            */
            if (sai->sin_addr.s_addr == INADDR_ANY) continue;
            if (sai->sin_addr.s_addr == INADDR_LOOPBACK) continue;
            rv = ntohl(sai->sin_addr.s_addr);
	    svc->trace_out(NULL, "CMSocket - Get self IP addr DHCP %lx, net %d.%d.%d.%d", 
			   ntohl(sai->sin_addr.s_addr), 
			   *((unsigned char *)&sai->sin_addr.s_addr),
			   *(((unsigned char *)&sai->sin_addr.s_addr) + 1),
			   *(((unsigned char *)&sai->sin_addr.s_addr) + 2),
			   *(((unsigned char *)&sai->sin_addr.s_addr) + 3));
            break;
          }
      }
#ifdef MODULE
    sys_close ((int)(long)ss);
    DFreeMM (ifreqs);
#else
    close (ss);
    free (ifreqs);
#endif
#endif
    /*
     *  Absolute last resort.  If we can't figure out anything else, look
     *  for the CM_LAST_RESORT_IP_ADDR environment variable.
     */
#ifndef MODULE
    if (rv == 0) {
        char *c = cercs_getenv ("CM_LAST_RESORT_IP_ADDR");
        if (c != NULL) {
	    rv = inet_addr(c);
/*     struct in_addr inp; inet_aton (c, &inp);   rv = inp.s_addr; */
          }
      }
    /*
     *  hopefully by now we've set rv to something useful.  If not,
     *  GET A BETTER NETWORK CONFIGURATION.
     */
#endif
    return rv;
}

static void
get_qual_hostname(char *buf, int len, CMtrans_services svc)
{
    struct hostent *host = NULL;

#ifdef MODULE
    char *network_string = getenv("CM_NETWORK");
#else
    char *network_string = cercs_getenv("CM_NETWORK");
#endif
    gethostname(buf, len);
    buf[len - 1] = '\0';
    if (memchr(buf, '.', strlen(buf)) == NULL) {
	/* no dots, probably not fully qualified */
#ifdef HAVE_GETDOMAINNAME
	int end = strlen(buf);
	buf[end] = '.';
#ifdef MODULE
	sys_getdomainname((&buf[end]) + 1, len - strlen(buf));
#else
	getdomainname((&buf[end]) + 1, len - strlen(buf));
#endif
	if (buf[end+1] == 0) {
	    char *tmp_name;
#ifdef MODULE
	    struct hostent *ent;
#endif
	    buf[end] = 0;
	    /* getdomainname was useless, hope that gethostbyname helps */
#ifdef MODULE
	    ent = lookup_name(buf);
	    tmp_name = ent->h_name;
	    strncpy(buf, tmp_name, len);
	    DFreeMM((addrs_t)ent->h_name);
	    DFreeMM((addrs_t)ent->h_addr_list);
	    DFreeMM((addrs_t)ent);
#else
	    tmp_name = (gethostbyname(buf))->h_name;
	    strncpy(buf, tmp_name, len);
#endif
	}
#else
	{
	    /* no getdomainname, hope that gethostbyname will help */
#ifdef MODULE
	    ent = lookup_name(buf);
	    tmp_name = ent->h_name;
	    strncpy(buf, tmp_name, len);
	    DFreeMM((addrs_t)ent->h_name);
	    DFreeMM((addrs_t)ent->h_addr_list);
	    DFreeMM((addrs_t)ent);
#else
	    char *tmp_name = (gethostbyname(buf))->h_name;
	    strncpy(buf, tmp_name, len);
#endif
	}
#endif
	buf[len - 1] = '\0';
    }
    svc->trace_out(NULL, "CMSocket - Tentative Qualified hostname %s", buf);
    if (memchr(buf, '.', strlen(buf)) == NULL) {
      /* useless hostname if it's not fully qualified */
      buf[0] = 0;
    }
#ifdef MODULE
    if ((buf[0] != 0) && ((host = lookup_name(buf)) == NULL)) {
#else
    if ((buf[0] != 0) && ((host = gethostbyname(buf)) == NULL)) {
#endif
      /* useless hostname if we can't translate it */
      buf[0] = 0; 
    }
    if (host != NULL) {
      char **p;
      int good_addr = 0;
      for (p=host->h_addr_list; *p !=0; p++) {
	struct in_addr *in = *(struct in_addr **)p;
	if (ntohl(in->s_addr) != INADDR_LOOPBACK) {
	  good_addr++;
	  svc->trace_out(NULL, 
			 "CMSocket - Hostname gets good addr %lx, %d.%d.%d.%d",
			 ntohl(in->s_addr), 
			 *((unsigned char *)&in->s_addr),
			 *(((unsigned char *)&in->s_addr) + 1),
			 *(((unsigned char *)&in->s_addr) + 2),
			 *(((unsigned char *)&in->s_addr) + 3));
	}
      }
      if (good_addr == 0) {
	/* even a fully qualifiedhostname that doesn't get us a valid IP addr is useless */
	buf[0] = 0;
      }
#ifdef MODULE
      DFreeMM((addrs_t)host->h_name);
      DFreeMM((addrs_t)host->h_addr_list);
      DFreeMM((addrs_t)host);
#endif
    }
    if (buf[0] == 0) {
#ifdef MODULE
	return;
#else
	/* bloody hell, what do you have to do? */
	int IP = htonl(get_self_ip_addr(svc));
	svc->trace_out(NULL, "CMSocket - No hostname yet, trying gethostbyaddr on IP %lx", IP);
	host = gethostbyaddr((char*)&IP, sizeof(IP), AF_INET);
	if (host != NULL) {
	    svc->trace_out(NULL, "     result was %s", host->h_name);
	    strncpy(buf, host->h_name, len);
	} else {
	    svc->trace_out(NULL, "     FAILED");
	}
#endif
    }
      
      
    
    if (network_string != NULL) {
	struct hostent *ent;
	int name_len = strlen(buf) + 2 + strlen(network_string);
	char *new_name_str = svc->malloc(name_len);
	char *first_dot = strchr(buf, '.');

	/* stick the CM_NETWORK value in there */
	memset(new_name_str, 0, name_len);
	*first_dot = 0;
	first_dot++;
	sprintf(new_name_str, "%s-%s.%s", buf, network_string, first_dot);
#ifdef MODULE
	if ((ent = lookup_name(new_name_str)) != NULL) {
#else
	if (gethostbyname(new_name_str) != NULL) {
#endif
	    /* host has no NETWORK interface */
	    strcpy(buf, new_name_str);
#ifdef MODULE
       	    DFreeMM((addrs_t)ent->h_name);
            DFreeMM((addrs_t)ent->h_addr_list);
            DFreeMM((addrs_t)ent);
#endif
	}
	svc->free(new_name_str);
    }

#ifdef MODULE
    if (((host = lookup_name(buf)) == NULL) ||
#else
    if (((host = gethostbyname(buf)) == NULL) ||
#endif
	(memchr(buf, '.', strlen(buf)) == NULL)) {
	/* just use the bloody IP address */
	gethostname(buf, len);
#ifdef MODULE
	host  = lookup_name(buf);
#else
	host  = gethostbyname(buf);
#endif
	if (host != NULL) {
	    struct in_addr ip;
	    char *tmp;
	    ip.s_addr = ntohl(get_self_ip_addr(svc));
	    tmp = inet_ntoa(ip);
	    strncpy(buf, tmp, len);
#ifdef MODULE
       	    DFreeMM((addrs_t)host->h_name);
            DFreeMM((addrs_t)host->h_addr_list);
            DFreeMM((addrs_t)host);
#endif
	} else {
	    static int warn_once = 0;
	    if (warn_once == 0) {
		warn_once++;
		printf("Attempts to establish your fully qualified hostname, or indeed any\nuseful network name, have failed horribly.  Sorry.\n");
	    }
	    strncpy(buf, "Unknown", len);
	}
    }
#ifdef MODULE
    if (host != NULL) {
        DFreeMM((addrs_t)host->h_name);
        DFreeMM((addrs_t)host->h_addr_list);
        DFreeMM((addrs_t)host);
    }
#endif
    svc->trace_out(NULL, "CMSocket - GetQualHostname returning %s", buf);
}

