/* -*-C-*-
*******************************************************************************
* 
* File:        group_formats.c
* RCS:         $Id: group_formats.c,v 1.1.1.1 2000/04/30 21:09:48 fabianb Exp $
* Description: Msgs formats decl kept in abc order (Greg's code modified for CM)
* Author:      Fabian E. Bustamante (fabianb@cc.gatech.edu)
*              Systems Research Group
*              College of Computing
*              Georgia Institute of Technology
* Created:
* Modified:
*  
* (C) Copyright 2000, all rights reserved.
*
*******************************************************************************
*/

#ifdef MODULE
#ifndef __KERNEL__
#define __KERNEL__
#endif
#endif

#ifndef MODULE
#include "config.h"

#include "atl.h"
#include "io.h"			/* pbio format decls */
#include "cm.h"
#include "comm_group2.h"
#include "group_formats.h"	/* msgs structs decls */
#else
#include <linux/kernel.h>
#include "config.h"
#include "../include/atl.h"
#include "../include/io.h"			/* pbio format decls */
#include "../include/cm.h"
#include "comm_group2.h"
#include "group_formats.h"	/* msgs structs decls */
#include "kgroup.h"
#endif

IOField assoc_group_info_msg_flds[] =
{
  {"cond", "integer", sizeof(int),
   IOOffset(assoc_group_info_msg_ptr, cond)},
  {"group name", "string", sizeof(char *),
   IOOffset(assoc_group_info_msg_ptr, group_name)},
  {"data_len", "integer", sizeof(int),
   IOOffset(assoc_group_info_msg_ptr, data_len)},
  {"data", "char[data_len]", sizeof(char),
   IOOffset(assoc_group_info_msg_ptr, data)},
  {(char *) 0, (char *) 0, 0, 0,}
};

IOField comm_group_struct_flds[] =
{
  {"user_name", "string", sizeof(char *),
   IOOffset(comm_group_list, user_name)},
  {"application_name", "string", sizeof(char *),
   IOOffset(comm_group_list, application_name)},
  {"group_type", "string", sizeof(char *),
   IOOffset(comm_group_list, group_type)},
  {"group_id", "string", sizeof(char *),
   IOOffset(comm_group_list, group_id)},
  {"host", "string", sizeof(char *),
   IOOffset(comm_group_list, host)},
  {"port", "integer", sizeof(int),
   IOOffset(comm_group_list, port)},
  {"info_len", "integer", sizeof(int),
   IOOffset(comm_group_list, info_len)},
  {"info[info_len]", "char[info_len]", 1,
   IOOffset(comm_group_list, info)},
  {(char *) 0, (char *) 0, 0, 0,}
};

IOField find_matching_msg_flds[] =
{
  {"cond", "integer", sizeof(int),
   IOOffset(find_matching_msg_ptr, cond)},
  {"user_name", "string", sizeof(char *), 
   IOOffset(find_matching_msg_ptr, user_name)},
  {"app_list", "string", sizeof(char *), 
   IOOffset(find_matching_msg_ptr, app_list)},
  {"type_list", "string", sizeof(char *), 
   IOOffset(find_matching_msg_ptr, type_list)},
  {(char *) 0, (char *) 0, 0, 0}
};


IOField init_contact_msg_flds[] =
{
  {"cond", "integer", sizeof(int),
   IOOffset(init_contact_msg_ptr, cond)},
  {"group id", "string", sizeof(char *), 
   IOOffset(init_contact_msg_ptr, group_id)},
  {"user name", "string", sizeof(char *), 
   IOOffset(init_contact_msg_ptr, user_name)},
  {"application name", "string", sizeof(char *),
   IOOffset(init_contact_msg_ptr, application_name)},
  {"group type", "string", sizeof(char *),
   IOOffset(init_contact_msg_ptr, group_type)},
  {(char *) 0, (char *) 0, 0, 0}
};

IOField provisional_use_msg_flds[] =
{
  {"domain", "string", sizeof(char *),
   IOOffset(provisional_use_msg_ptr, domain)},
  {"time", "integer", sizeof(int),
   IOOffset(provisional_use_msg_ptr, time)},
  {(char *) 0, (char *) 0, 0, 0,}
};

IOField remove_group_msg_flds[] =
{
  {"host", "string", sizeof(char *), 
   IOOffset(return_contact_msg_ptr, host)},
  {"port", "integer", sizeof(int), 
   IOOffset(return_contact_msg_ptr, port)},
  {"group id", "string", sizeof(char *), 
   IOOffset(return_contact_msg_ptr, group_id)},
  {(char *) 0, (char *) 0, 0, 0}
};

IOField return_contact_msg_flds[] =
{  
  {"cond", "integer", sizeof(int),
   IOOffset(return_contact_msg_ptr, cond)},
  {"host", "string", sizeof(char *), 
   IOOffset(return_contact_msg_ptr, host)},
  {"port", "integer", sizeof(int), 
   IOOffset(return_contact_msg_ptr, port)},
  {"group id", "string", sizeof(char *), 
   IOOffset(return_contact_msg_ptr, group_id)},
  {(char *) 0, (char *) 0, 0, 0}
};

CMFormatRec return_format_list[] = {
  {"comm_group_struct", comm_group_struct_flds},
  {NULL,NULL}
};

IOField return_matching_msg_flds[] =
{
  {"cond", "integer", sizeof(int),
   IOOffset(comm_group_return, cond)},
  {"count", "integer", sizeof(int), 
   IOOffset(comm_group_return, count)},
  {"list", "comm_group_struct[count]", sizeof(comm_group_struct),
   IOOffset(comm_group_return, list)},
  {(char *) 0, (char *) 0, 0, 0,}
};

IOField setup_group_msg_flds[] =
{
  {"cond", "integer", sizeof(int),
   IOOffset(setup_group_msg_ptr, cond)},
  {"user name", "string", sizeof(char *), 
   IOOffset(setup_group_msg_ptr, user_name)},
  {"application name", "string", sizeof(char *),
   IOOffset(setup_group_msg_ptr, application_name)},
  {"group type", "string", sizeof(char *),
   IOOffset(setup_group_msg_ptr, group_type)},
  {"expiration", "integer", sizeof(int),
   IOOffset(setup_group_msg_ptr, expiration)},
  {"host", "string", sizeof(char *),
   IOOffset(setup_group_msg_ptr, host)},
  {"port", "integer", sizeof(int),
   IOOffset(setup_group_msg_ptr, port)},

  {(char *) 0, (char *) 0, 0, 0}
};
