
/*
 *   kgroup.c
 *
 *   Copyright (C) 2000 Christian Poellabauer
 *
 */

/* 
 *   Kernel-loadable module for event channel support
 */

/* Required definitions. */
#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif

/* Various required header files. */
#include <linux/module.h> 
#include <linux/kernel.h> 
#include <linux/tty.h>
#include <linux/unistd.h>
#include <linux/socket.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <asm/current.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include "kgroup.h"
#include "../include/cm.h"
#include "comm_group2.h"

MODULE_AUTHOR("Christian Poellabauer");
MODULE_DESCRIPTION("CommGroup Module (KECho)");
MODULE_LICENSE("GPL");

DECLARE_MUTEX(group_sem);

/* Prototypes. */
int init_module(void);		       
void cleanup_module(void);
int (*original_syscall_cg)();

/* Global variables. */

/* arch/i386/kernel/entry.S */
extern void *sys_call_table[];
extern void *fast_sys_call_table[];

int int_syscall_cg(data)
void *data;
{
  syscall_attrs_cg *attrs;
  char name1[MAX_STR];
  char name2[MAX_STR];
  char name3[MAX_STR];
  char host[MAX_STR];
  char *retchar = NULL;
  void *kdata = NULL;
  comm_group_return group_return;

  attrs = (syscall_attrs_cg *)data;

  if (attrs == NULL) return -1;

  if (strcmp("setup_comm_group", attrs->name) == 0) {
    name1[0] = '0', name2[0] = '0', name3[0] = '0', host[0] = '0';
    if (attrs->username[0] != '0')
      copy_from_user(name1, attrs->username, strlen(attrs->username)+1);
    if (attrs->appname[0] != '0')
      copy_from_user(name2, attrs->appname, strlen(attrs->appname)+1);
    if (attrs->grouptype[0] != '0')
      copy_from_user(name3, attrs->grouptype, strlen(attrs->grouptype)+1);
    if (attrs->host[0] != '0')
      copy_from_user(host, attrs->host, strlen(attrs->host)+1);
    retchar = setup_comm_group(name1, name2, name3, attrs->exp_time, 
	host, attrs->port);
    if (retchar != NULL) {
      copy_to_user(attrs->groupid, retchar, strlen(retchar)+1);
      attrs->group_id = (char *)attrs->groupid;
    } else
      attrs->group_id = NULL;
    return 0;
  }
  else if (strcmp("setup_specific_comm_group", attrs->name) == 0) {
    copy_from_user(name2, attrs->appname, strlen(attrs->appname)+1);
    copy_from_user(name3, attrs->grouptype, strlen(attrs->grouptype)+1);
    copy_from_user(host, attrs->host, strlen(attrs->host)+1);
    retchar = setup_specific_comm_group(name2, name3, attrs->exp_time, 
	host, attrs->port);
    memcpy(attrs->groupid, retchar, strlen(retchar)+1);
    attrs->group_id = (char *)attrs->groupid;
    return 0;
  }
  else if (strcmp("init_comm_group_contact", attrs->name) == 0) {
    copy_from_user(name1, attrs->username, strlen(attrs->username)+1);
    copy_from_user(name2, attrs->appname, strlen(attrs->appname)+1);
    copy_from_user(name3, attrs->grouptype, strlen(attrs->grouptype)+1);
    copy_from_user(host, attrs->host, strlen(attrs->host)+1);
    attrs->retval = init_comm_group_contact(name1, name2, name3, 
	(char **)&host, &(attrs->port));
    return 0;
  }
  else if (strcmp("init_specific_comm_group_contact", attrs->name) == 0) {
    copy_from_user(name1, attrs->groupid, strlen(attrs->groupid)+1);
    attrs->retval = init_specific_comm_group_contact(name1, 
	(char **)&host, &(attrs->port));
    return 0;
  }
  else if (strcmp("group_server_present", attrs->name) == 0) {
    attrs->retval = group_server_present();
    return 0;
  }
  else if (strcmp("group_server_test", attrs->name) == 0) {
    attrs->retval = group_server_test();
    return 0;
  }
  else if (strcmp("assoc_group_info", attrs->name) == 0) {
    kdata = kmalloc(attrs->data_len, GFP_KERNEL);
    copy_from_user(kdata, attrs->data, attrs->data_len);  
    copy_from_user(name1, attrs->grouptype, strlen(attrs->grouptype)+1);
    assoc_group_info(name1, kdata, attrs->data_len); 
    kfree(kdata);
    return 0;
  }
  else if (strcmp("matching_comm_groups", attrs->name) == 0) {
    copy_from_user(name1, attrs->applist, strlen(attrs->applist)+1);
    copy_from_user(name2, attrs->typelist, strlen(attrs->typelist)+1);
    group_return = matching_comm_groups(name1, name2);
    copy_to_user(attrs->groupinfo, group_return->list[0].info, 
	strlen(group_return->list[0].info)+1);
    return 0;
  }
  else
    return -1;
}

/* Initialize module. */
int init_module(void) 
{
  original_syscall_cg = sys_call_table[__NR_KECho_syscall_cg];

  sys_call_table[__NR_KECho_syscall_cg] = int_syscall_cg;

  return 0;
}

/* Remove module. */
void cleanup_module(void) 
{
  sys_call_table[__NR_KECho_syscall_cg] = original_syscall_cg;
}


