/* Test that save/restore and locals do not overlap in the activation record. */
#include <assert.h>
#include <stdio.h>

#include "drisc.h"


void * mk_test(drisc_ctx c) {
	int i, regs, fregs;
	dr_reg_t al[10];
	int l[100];
	int abortl;

	dr_reg_t r[32];
	dr_reg_t fr[32];

	dr_lambda(c, "foo", "%i%d", al);

	abortl = dr_genlabel(c);

	for(i = 0; i < 5 && dr_getreg(c, &r[i], DR_I, DR_TEMP); i++) {
/*		(void)("allocated register %d\n", i);*/
		dr_seti(c, r[i], i);
		dr_savei(c, r[i]); 
	}
	regs = i;

	for(i = 0; dr_getreg(c, &fr[i], DR_D, DR_TEMP); i++) {
/*		(void)("allocated register %d\n", i);*/
		dr_setd(c, fr[i], (double)i);
		dr_saved(c, fr[i]); 
	}
	fregs = i;

	for(i = 0; i < 10; i++) {
		l[i] = dr_local(c, DR_I);
			
		dr_seti(c, al[0], 100 + i);
		dr_stii(c, al[0], dr_lp(c), l[i]);
	}

	/* call a procedure, then verify that everything is in place. */
	dr_scallv(c, (void*)printf, "%P", "hello world!\n");


	for(i = 0; i < regs; i++) {
		dr_restorei(c, r[i]); 
		dr_bneii(c, r[i], i, abortl);
	}

	for(i = 0; i < fregs; i++) {
/*		(void)("allocated register %d\n", i);*/
		dr_restored(c, fr[i]); 
		dr_setd(c, al[1], (double)i);
		dr_bned(c, fr[i], al[1], abortl);	
	}

	for(i = 0; i < 10; i++) {
		dr_ldii(c, al[0], dr_lp(c), l[i]);
		dr_bneii(c, al[0], 100 + i, abortl);
	}
	dr_retii(c, 1);

	dr_label(c, abortl);
	dr_retii(c, 0);		/* failure. */

	return dr_end(c);
}

int main() { 
	int (*ip)();
	drisc_ctx c = dr_init();

	ip = mk_test(c);

	if(ip()) {
		printf("success!\n");
	} else {
		dr_dump(c);
		printf("failure!\n");
	}
	ip = mk_test(c);
	if(ip())
		printf("success!\n");
	else {
		dr_dump(c);
		printf("failure!\n");
	}

	return 0;
}
