/*
 *   kecl.c
 *
 *   Copyright (C) 2002 Christian Poellabauer
 *
 */

/* 
 *   Kernel-loadable module for event channel support
 */

/* Required definitions. */
#ifdef MODULE
#ifndef __KERNEL__
#define __KERNEL__
#endif
#endif

/* Various required header files. */
#include <linux/config.h> 
#include <linux/module.h> 
#include <linux/kernel.h> 
#include <linux/init.h> 
#include <linux/tty.h>
#include <linux/unistd.h>
#include <linux/socket.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <asm/current.h>
#include <asm/semaphore.h>
#include "kecl.h"

MODULE_AUTHOR("Eisenhauer/Poellabauer");
MODULE_DESCRIPTION("ECL Module (KECho)");
MODULE_LICENSE("GPL");

/* arch/i386/kernel/entry.S */
extern void *sys_call_table[];

/* List of system calls used in the module. */
int     (*sys_exit)       (int error_code);
int     (*sys_fork)       (struct pt_regs regs);
__kernel_ssize_t (*sys_kread) (unsigned int fd, char *buf, size_t count);
__kernel_ssize_t (*sys_kwrite)(unsigned int fd, const char *buf, size_t count);
int     (*sys_kopen)      (const char *filename, int flags, int mode);
int     (*sys_kclose)     (unsigned int fd);
int     (*sys_waitpid)    (__kernel_pid_t pid, unsigned int *stat_addr, 
                           int options);
int     (*sys_unlink)     (const char *pathname);
int     (*sys_time)       (int *tloc);
int     (*sys_ni_syscall) (void);
int     (*sys_getpid)     (void);
int     (*sys_getuid)     (void);
unsigned int (*sys_alarm) (unsigned int seconds);
int     (*sys_kill)       (int pid, int sig);
int     (*sys_pipe)       (unsigned long *fildes);
unsigned long (*sys_signal) (int sig, __sighandler_t handler);
int     (*sys_sigaction)  (int sig, const struct old_sigaction *act,
                           struct old_sigaction *oact);
int     (*sys_setrlimit)  (unsigned int resource, struct rlimit *rlim);
int     (*sys_select)     (int n, fd_set *inp, fd_set *outp,
                          fd_set *exp, struct timeval *tvp);
int     (*sys_truncate)   (const char *path, unsigned long length);
__kernel_ssize_t (*sys_readv) (unsigned long fd, const struct iovec *vector,
                           unsigned long count);
__kernel_ssize_t (*sys_writev)  (unsigned long fd, const struct iovec *vector,
                           unsigned long count);
int     (*sys_rt_sigaction) (int sig, const struct sigaction *act, struct
                             sigaction *oact, size_t sigsetsize);

int     (*sys_socket)      (int family, int type, int protocol);
int     (*sys_connect)     (int fd, struct sockaddr *uservaddr, int addrlen);
long    (*sys_kfcntl)       (unsigned int fd, unsigned int cmd, 
			    unsigned long arg);


/* Initialize module. */
static int __init kecl_init(void) 
{
  /* Get system table entries. */
  sys_exit         =   sys_call_table[__NR_exit];
  sys_fork         =   sys_call_table[__NR_fork];
  sys_kread         =   sys_call_table[__NR_read];
  sys_kwrite        =   sys_call_table[__NR_write];
  sys_kopen        =   sys_call_table[__NR_open];
  sys_kclose       =   sys_call_table[__NR_close];
  sys_waitpid      =   sys_call_table[__NR_waitpid];
  sys_unlink       =   sys_call_table[__NR_unlink];
  sys_time         =   sys_call_table[__NR_time];
  sys_ni_syscall   =   sys_call_table[__NR_break];
  sys_getpid       =   sys_call_table[__NR_getpid];
  sys_getuid       =   sys_call_table[__NR_getuid];
  sys_alarm        =   sys_call_table[__NR_alarm];
  sys_kill         =   sys_call_table[__NR_kill];
  sys_pipe         =   sys_call_table[__NR_pipe];
  sys_signal       =   sys_call_table[__NR_signal];
  sys_sigaction    =   sys_call_table[__NR_sigaction];
  sys_setrlimit    =   sys_call_table[__NR_setrlimit];
  sys_select       =   sys_call_table[__NR_select];
  sys_truncate     =   sys_call_table[__NR_truncate];
  sys_readv        =   sys_call_table[__NR_readv];
  sys_writev       =   sys_call_table[__NR_writev];
  sys_rt_sigaction =   sys_call_table[__NR_rt_sigaction];
  sys_kfcntl       =   sys_call_table[__NR_fcntl];

  return 0;
}

/* Remove module. */
static void __exit kecl_cleanup(void) 
{
}

module_init(kecl_init);
module_exit(kecl_cleanup); 

