
#ifdef MODULE
#ifndef __KERNEL__
#define __KERNEL__
#endif
#endif

#ifndef MODULE
#include "config.h"
#define FUNCPROTO
#include "io.h"
#ifdef HAVE_GEN_THREAD_H
#include "gen_thread.h"
#else
#define gen_thr_initialized() 0
#define thr_mutex_lock(m)
#define thr_mutex_unlock(m)
#endif
#include "atl.h"
#include "cm.h"
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <string.h>
#ifdef STDC_HEADERS
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#ifdef HAVE_WINDOWS_H
#include <windows.h>
#else
#include <netinet/in.h>
#include <arpa/inet.h>
#endif
extern int vfprintf();
#ifndef PRINTF_DEFINED
extern int printf ARGS((const char *,...));
#endif
#ifndef FPRINTF_DEFINED
extern int fprintf ARGS((FILE *, const char *,...));
#endif
#ifndef SPRINTF_DEFINED
extern int sprintf ARGS((char *, const char *,...));
#endif
#ifndef SSCANF_DEFINED
extern int sscanf ARGS((const char *, const char *,...));
#endif

#ifndef NO_DERIVED
#include "ecl.h"
#endif
#include "echo.h"
#include "echo_internal.h"
#else
#include <asm/uaccess.h>
#include <linux/in.h>
#include "config.h"
#include "kecho.h"
#include "library.h"
#include "../include/io.h"
#include "../include/atl.h"
#include "../include/cm.h"
#ifndef NO_DERIVED
#include "../include/ecl.h"
#endif
#include "echo.h"
#include "echo_internal.h"

#define printf         printk
#define perror         printk
#define fprintf(fmt, args...) printk (args) 
#define malloc         (void *)DAllocMM
#define free(a)        DFreeMM((addrs_t)a)
#define realloc(a,b)   (void *)DReallocMM((addrs_t)a,b)
#define exit           sys_exit
#define qsort          _quicksort
#define getpid()       current->pid
#define signal(a,b)    
#define gethostname    sys_gethostname
#define gethostbyname  lookup_name
#define getdomainname  sys_getdomainname
#define select         sys_select

#define my_rdtsc(x)	__asm__ __volatile__ (		\
				"rdtsc \n\t"		\
				"movl %%eax,%0 \n\t"	\
				: "=r" (x)		\
				:			\
				: "%eax","%edx");

static char *
inet_ntoa(struct in_addr ina)
{
    static char buf[4*sizeof "123"];
    unsigned char *ucp = (unsigned char *)&ina;

    sprintf(buf, "%d.%d.%d.%d",
	ucp[0] & 0xff,
	ucp[1] & 0xff,
	ucp[2] & 0xff,
	ucp[3] & 0xff);
    return buf;
}
#endif

IOField channel_id_flds[] =
{
    {"len", "integer", sizeof(int), IOOffset(channel_ID_struct*,len)},
    {"id", "char[len]", 1, IOOffset(channel_ID_struct*,channel)},
    {(char *) 0, (char *) 0, 0, 0}
};

IOField channel_attend_msg_flds[] =
{
    {"channel", "string", sizeof(char *), IOOffset(AttendMsgPtr, chan_str)},
    {"contact_str", "string", sizeof(char *), IOOffset(AttendMsgPtr, 
						       contact_str)},
    {"condition", "integer", sizeof(int), IOOffset(AttendMsgPtr, cond)},
    {"client_channel_ID", "channel_ID", sizeof(channel_ID_struct), 
     IOOffset(AttendMsgPtr, client_channel_id)},
    {(char *) 0, (char *) 0, 0, 0}
};

IOField sink_subscribe_msg_flds[] =
{
#ifdef MODULE
    {"conn_id", "string", sizeof(char *),
     IOOffset(SinkSubscribeMsgPtr, conn_id)},
#endif
    {"global_channel_id", "channel_ID", sizeof(channel_ID_struct), 
     IOOffset(SinkSubscribeMsgPtr, global_channel_id)},
    {"client_channel_id", "channel_ID", sizeof(channel_ID_struct), 
     IOOffset(SinkSubscribeMsgPtr, client_channel_id)},
    {"client_contact_str", "string", sizeof(char *), 
     IOOffset(SinkSubscribeMsgPtr, client_contact_str)},
    {(char *) 0, (char *) 0, 0, 0}
};

IOField sink_unsubscribe_msg_flds[] =
{
    {"global_channel_id", "channel_ID", sizeof(channel_ID_struct), 
     IOOffset(SinkSubscribeMsgPtr, global_channel_id)},
    {"client_channel_id", "channel_ID", sizeof(channel_ID_struct), 
     IOOffset(SinkSubscribeMsgPtr, client_channel_id)},
    {(char *) 0, (char *) 0, 0, 0}
};

/* basically the same but still ... */

IOField source_subscribe_msg_flds[] =
{
    {"global_channel", "channel_ID", sizeof(channel_ID_struct), 
     IOOffset(SourceSubscribeMsgPtr, global_channel_id)},
    {"client_channel", "channel_ID", sizeof(channel_ID_struct), 
     IOOffset(SourceSubscribeMsgPtr, client_channel_id)},
    {"client_contact_str", "string", sizeof(char *), 
     IOOffset(SourceSubscribeMsgPtr, client_contact_str)},
    {(char *) 0, (char *) 0, 0, 0}
};

IOField source_unsubscribe_msg_flds[] =
{
    {"global_channel", "channel_ID", sizeof(channel_ID_struct), 
     IOOffset(SourceSubscribeMsgPtr, global_channel_id)},
    {"client_channel", "channel_ID", sizeof(channel_ID_struct), 
     IOOffset(SourceSubscribeMsgPtr, client_channel_id)},
    {(char *) 0, (char *) 0, 0, 0}
};

IOField channel_exists_attend_response_msg_flds[] =
{
    {"condition", "integer", sizeof(int),
     IOOffset(ExistsAttendResponseMsgPtr, condition)},
    {"existing_channel_id", "channel_ID", sizeof(channel_ID_struct), 
     IOOffset(ExistsAttendResponseMsgPtr, existing_channel_id)},
    {(char *) 0, (char *) 0, 0, 0}
};

IOField AttendRec_flds[] =
{
    {"contact_string", "string", sizeof(char *), 
     IOOffset(AttendRecPtr, contact_string)},
    {"client_channel_id", "channel_ID", sizeof(channel_ID_struct), 
     IOOffset(AttendRecPtr, client_channel_id)},
    {"filter", "string", sizeof(char*), IOOffset(AttendRecPtr, filter)},
    {"filter_has_output_param", "integer", sizeof(int), 
     IOOffset(AttendRecPtr, filter_has_output_param)},
#ifdef MODULE
    {"conn_id", "string", sizeof(char *),
     IOOffset(AttendRecPtr, conn_id)},
#endif
    {(char *) 0, (char *) 0, 0, 0}
};

IOField channel_attend_response_msg_flds[] =
{
    {"condition", "integer", sizeof(int),
     IOOffset(AttendResponseMsgPtr, condition)},
    {"global_channel_id", "channel_ID", sizeof(channel_ID_struct),
     IOOffset(AttendResponseMsgPtr, global_channel_id)},
    {"type_name", "string", sizeof(char *),
     IOOffset(AttendResponseMsgPtr, type_name)},
    {"sink_count", "integer", sizeof(int),
     IOOffset(AttendResponseMsgPtr, sink_count)},
    {"sinks", "AttendRec[sink_count]", sizeof(AttendRec),
     IOOffset(AttendResponseMsgPtr, sinks)},
    {"source_count", "integer", sizeof(int),
     IOOffset(AttendResponseMsgPtr, source_count)},
    {"sources", "AttendRec[source_count]", sizeof(AttendRec),
     IOOffset(AttendResponseMsgPtr, sources)},
    {"member_count", "integer", sizeof(int),
     IOOffset(AttendResponseMsgPtr, member_count)},
    {"members", "AttendRec[member_count]", sizeof(AttendRec),
     IOOffset(AttendResponseMsgPtr, members)},
    {"der_count", "integer", sizeof(int),
     IOOffset(AttendResponseMsgPtr, der_count)},
    {"derivatives", "AttendRec[der_count]", sizeof(AttendRec),
     IOOffset(AttendResponseMsgPtr, derivatives)},
    {(char *) 0, (char *) 0, 0, 0}
};

IOField member_subscribe_msg_flds[] =
{
    {"global_channel_id", "channel_ID", sizeof(channel_ID_struct), 
     IOOffset(MemberSubscribeMsgPtr, global_channel_id)},
    {"client_channel_id", "channel_ID", sizeof(channel_ID_struct), 
     IOOffset(MemberSubscribeMsgPtr, client_channel_id)},
    {"client_contact_str", "string", sizeof(char *), 
     IOOffset(MemberSubscribeMsgPtr, client_contact_str)},
    {(char *) 0, (char *) 0, 0, 0}
};


IOField event_msg_flds[] =
{
    {"global_chanel_id", "channel_ID", sizeof(channel_ID_struct), 
     IOOffset(EventMsgPtr, global_channel_id)},
    {"event_len", "integer", sizeof(int),
     IOOffset(EventMsgPtr, event_len)},
    {"event", "char[event_len]", sizeof(char),
     IOOffset(EventMsgPtr, event)},
    {(char *) 0, (char *) 0, 0, 0}
};

IOField request_event_msg_flds[] =
{
    {"global_channel_id", "channel_ID", sizeof(channel_ID_struct), 
     IOOffset(RequestEventMsgPtr, global_channel_id)},
    {(char *) 0, (char *) 0, 0, 0}
};

IOField event_vec_elem_flds[] =
{
    {"len", "integer", sizeof(((IOEncodeVector)0)[0].iov_len), 
     IOOffset(IOEncodeVector, iov_len)},
    {"elem", "char[len]", sizeof(char), IOOffset(IOEncodeVector, iov_base)},
    {(char *) 0, (char *) 0, 0, 0}
};

IOField eventv_msg_flds[] =
{
    {"global_channel_id", "channel_ID", sizeof(channel_ID_struct), 
     IOOffset(EventVMsgPtr, global_channel_id)},
    {"eventv_len", "integer", sizeof(int),
     IOOffset(EventVMsgPtr, eventv_len)},
    {"eventv", "IOEncodeElem[eventv_len]", sizeof(struct _io_encode_vec),
     IOOffset(EventVMsgPtr, eventv)},
    {(char *) 0, (char *) 0, 0, 0}
};

static event_queue dequeue_event ARGS((EControlContext ec));
static void internal_cancel_source_subscribe ARGS((ECSourceHandle handle));

#ifdef MODULE
int channel_verbose = 0;
#else
int channel_verbose = -1;
#endif
int channel_lock_verbose = 0;
/* trace_out - for debuging and tracing */
/* prints the value of function string, formats the output by */
/* indenting first if direction is a 2, and un-indenting if */
/* direction is a 0, if direction is a 1, indenting is increased */
/* for the traces that follow. */
extern void
ECtrace_out(char *format, int direction, ...)
{
#ifndef MODULE
    va_list ap;
    static int tabstop = 0;
    int i;

    if (channel_verbose == -1) {
	char *verb = getenv("CHANNEL_VERBOSE");
	channel_verbose = 0;
	if (verb) {
	    if (sscanf(verb, "%d", &channel_verbose) != 1) {
		/* enable at lowest level */
		channel_verbose = 1;
	    }
	}
	verb = getenv("CHANNEL_LOCK_VERBOSE");
	if (verb) {
	    channel_lock_verbose = 1;
	}
    }

    if (channel_verbose) {
#ifdef STDC_HEADERS
	va_start(ap, direction);
#else
	va_start(ap);
#endif
	if (direction == 0)
	    tabstop--;
	if (direction == 2)
	    tabstop++;
	for (i = 0; i < tabstop; i++)
	    printf("  ");
	vfprintf(stdout, format, ap);
	va_end(ap);
	switch (direction) {
	case 0:
	    printf(" stop");
	    break;
	case 1:
	    printf(" start");
	    tabstop++;
	    break;
	case 2:
	    tabstop--;
	    break;
	}
	printf("\n");
    }
#endif
}

extern void
dump_EChannel(chan)
EChannel chan;
{
    char *ID = ECglobal_id(chan);
    int i;

    printf("Local Channel %016lx\n", (long) chan);
    printf("    Sink count = %d, Source count = %d\n", chan->sink_count,
	   chan->source_count);
    printf("    full contact string = %s\n", ID);
    CMfree(ID);
    printf("    global ID = ");
    print_channel_id(chan->global_id.global_channel_id);
    printf("\n");
    if (chan->type_format) {
	char *format_name = global_name_of_IOformat(chan->type_format);
	printf("	type = %s\n", format_name);
	free(format_name);
    }
    printf("    %d remote sinks\n", chan->rem_sink_list_len);
    for (i = 0; i < chan->rem_sink_list_len; i++) {
	printf("	%d) - remote_id ", i);
	print_channel_id(chan->rem_sink_list[i].remote_id);
	printf(", %s, %lx\n", chan->rem_sink_list[i].contact_string,
	       (long) chan->rem_sink_list[i].conn);
    }

    printf("    %d remote sources\n", chan->rem_source_list_len);
    for (i = 0; i < chan->rem_source_list_len; i++) {
	printf("	%d) - remote_id ", i);
	print_channel_id(chan->rem_source_list[i].remote_id);
	printf(", %s, %lx\n", chan->rem_source_list[i].contact_string,
	       (long) chan->rem_source_list[i].conn);
    }

    printf("    %d remote members\n", chan->rem_member_list_len);
    for (i = 0; i < chan->rem_member_list_len; i++) {
	printf("	%d) - channel ", i);
	print_channel_id(chan->rem_member_list[i].remote_id);
	printf(", %s, %lx\n", chan->rem_member_list[i].contact_string,
	       (long) chan->rem_member_list[i].conn);
    }
    printf("    %d local sinks\n", chan->local_sink_list_len);
    for (i = 0; i < chan->local_sink_list_len; i++) {
	printf("	%d) - func %lx, data %lx, multi_sink %lx, ec %lx, context %lx, condition %d\n", i,
	       (long) chan->local_sink_list[i].func,
	       (long) chan->local_sink_list[i].client_data,
	       (long) chan->local_sink_list[i].multi_sink_handle,
	       (long) chan->local_sink_list[i].ec,
	       (long) chan->local_sink_list[i].context,
	       (int)  chan->local_sink_list[i].cond);
    }
    printf("    %d local sources\n", chan->local_source_list_len);
    for (i = 0; i < chan->local_source_list_len; i++) {
	printf("	%d) -  func %lx, data %lx, ioformat %lx", i,
	       (long) chan->local_source_list[i].func,
	       (long) chan->local_source_list[i].client_data,
	       (long) chan->local_source_list[i].source_handle);
    }
    printf("    %d derived channels\n", chan->der_list_len);
    for (i = 0; i < chan->der_list_len; i++) {
	printf("    derived event channel %d\n", i);
	printf("    ========================\n");
	printf("	filter = %s\n", chan->der_chan_list[i].filter);
	printf("	has_output_param = %d\n", 
	       chan->der_chan_list[i].has_output_param);
	dump_EChannel(chan->der_chan_list[i].derived_chan);
    }
}


static void
return_free_block(block)
free_block_rec_p block;
{
    block->next = block->locking_context->free_block_list;
    block->locking_context->free_block_list = block;
}

static free_block_rec_p 
alloc_free_block(ec)
EControlContext ec;
{
    if (ec->free_block_list == NULL) {
	return (free_block_rec_p) CMmalloc(sizeof(struct free_block_rec));
    } else {
	free_block_rec_p tmp = ec->free_block_list;
	ec->free_block_list = tmp->next;
	return tmp;
    }
}

static free_block_rec_p
get_free_block_func(locking_context, block, free_func)
EControlContext locking_context;
void *block;
EventFreeFunction free_func;
{
    free_block_rec_p block_rec = alloc_free_block(locking_context);
    block_rec->locking_context = locking_context;
    block_rec->ref_count = 1;
    block_rec->ec = block_rec->free_arg = NULL;
    block_rec->free_func = free_func;
    block_rec->block = block;
    block_rec->iobuffer = NULL;
    return block_rec;
}

static free_block_rec_p
get_free_block_iobuffer(locking_context, iobuffer)
EControlContext locking_context;
IOBuffer iobuffer;
{
    free_block_rec_p block_rec = alloc_free_block(locking_context);
    block_rec->locking_context = locking_context;
    block_rec->ref_count = 1;
    block_rec->ec = block_rec->free_arg = NULL;
    block_rec->free_func = (EventFreeFunction)NULL;
    block_rec->block = NULL;
    block_rec->iobuffer = iobuffer;
    return block_rec;
}

static void
deref_block(block)
free_block_rec_p block;
{
    block->ref_count--;
    EContext_lock(block->locking_context);
    if (block->ref_count == 0) {
	if (block->ec) {
	    CMreturn_buffer(block->ec->cm, block->free_arg);
	} else if (block->iobuffer != NULL) {
	    free_IOBuffer(block->iobuffer);
	} else if (block->free_func != (EventFreeFunction) NULL) {
	    block->free_func(block->block);
	} else if (block->block != NULL) {
	    CMfree(block->block);
	}
	return_free_block(block);
    }
    EContext_unlock(block->locking_context);
}

#ifdef MODULE
static void
set_default_rt_attrs(ECSourceHandle handle)
{
    struct timeval tv_default;

    do_gettimeofday(&tv_default);
    handle->rt_attrs.timestamp = tv_default;
    tv_default.tv_sec = 0;
    tv_default.tv_usec = 0;
    handle->rt_attrs.deadline = tv_default;
    handle->rt_attrs.est = tv_default;
    handle->rt_attrs.lst = tv_default;
    handle->rt_attrs.period = tv_default;
    handle->rt_attrs.type = 0;
    handle->rt_attrs.rt_header[0] = 0x52544500;  /* identifier: RTE\0 */
    handle->rt_attrs.rt_header[1] = 0;
    handle->rt_attrs.priority = -1;
}
#endif

static ECSourceHandle
new_source_handle()
{
    ECSourceHandle handle = CMmalloc(sizeof(*handle));
    handle->chan = NULL;
    handle->ioformat = NULL;
    handle->source_num = 0;
    handle->der_list_len = 0;
    handle->der_source_list = NULL;
#ifdef MODULE
    set_default_rt_attrs(handle);
#endif
    return handle;
}

extern void
IntEChannel_lock(chan, file, line)
EChannel chan;
char *file;
int line;
{
#ifdef MODULE
    down(&chan->lock);
#else
    if (!gen_thr_initialized())
	return;

    if (channel_lock_verbose) {
	printf("Trying to lock channel %lx at %s line %d\n", (long) chan, 
	       file, line);
    }
    if (chan->lock) {
	thr_mutex_lock(chan->lock);
    }
    if (channel_lock_verbose) {
	printf("Locked channel %lx at %s line %d\n", (long) chan, file,
	       line);
    }
#endif
}

extern void
IntEChannel_unlock(chan, file, line)
EChannel chan;
char *file;
int line;
{
#ifdef MODULE
    up(&chan->lock);
#else
    if (!gen_thr_initialized())
	return;

    if (channel_lock_verbose) {
	printf("Unlocking channel %lx at %s line %d\n", (long) chan, file,
	       line);
    }
    if (chan->lock) {
	thr_mutex_unlock(chan->lock);
    }
#endif
}

extern void
IntEContext_lock(cc, file, line)
EControlContext cc;
char *file;
int line;
{
#ifdef MODULE
    down(&cc->lock);
#else
    if (!gen_thr_initialized())
	return;

    if (channel_lock_verbose) {
	printf("Trying to lock context %lx at %s line %d\n", (long) cc, 
	       file, line);
    }
    if (cc->lock) {
	thr_mutex_lock(cc->lock);
    }
    if (channel_lock_verbose) {
	printf("Locked context %lx at %s line %d\n", (long) cc, file,
	       line);
    }
#endif
}

extern void
IntEContext_unlock(cc, file, line)
EControlContext cc;
char *file;
int line;
{
#ifdef MODULE
    up(&cc->lock);
#else
    if (!gen_thr_initialized())
	return;

    if (channel_lock_verbose) {
	printf("Unlocking context %lx at %s line %d\n", (long) cc, file,
	       line);
    }
    if (cc->lock) {
	thr_mutex_unlock(cc->lock);
    }
#endif
}

extern EChannel
string_to_channel_ptr(char *chan_id)
{
    unsigned char chan_char[8] = {0, 0, 0, 0, 0, 0, 0, 0};
    unsigned int chan_int[8] =	{0, 0, 0, 0, 0, 0, 0, 0};
    EChannel chan;
#ifdef MODULE
    char temp[30];
    char temp2[30];
    int i;

    for (i=0; i<8; i++)
    {
      strncpy((char *)temp, chan_id + 2*i, 2);
      chan_int[i] = strtol((char *)temp, (char **)&temp2, 16);
    }
#else
    sscanf(chan_id, "%02x%02x%02x%02x%02x%02x%02x%02x",
	   &chan_int[0], &chan_int[1], &chan_int[2], &chan_int[3],
	   &chan_int[4], &chan_int[5], &chan_int[6], &chan_int[7]);
#endif

    chan_char[0] = (unsigned char) chan_int[0];
    chan_char[1] = (unsigned char) chan_int[1];
    chan_char[2] = (unsigned char) chan_int[2];
    chan_char[3] = (unsigned char) chan_int[3];
    chan_char[4] = (unsigned char) chan_int[4];
    chan_char[5] = (unsigned char) chan_int[5];
    chan_char[6] = (unsigned char) chan_int[6];
    chan_char[7] = (unsigned char) chan_int[7];
    memcpy(&chan, chan_char, sizeof(chan));
    return chan;
}

extern
EChannel
EChannel_typed_create(cc, field_list, format_list)
EControlContext cc;
IOFieldList field_list;
CMFormatList format_list;
{
    EChannel chan = EChannel_create(cc);
    char format_name[] = "anonymous";

    chan->subcontext = create_IOsubcontext(CMget_user_type_context(cc->cm));

    chan->type_format = 
	CMregister_user_format(cc->cm, chan->subcontext, format_name, 
			       field_list, format_list);
    return chan;
}

#ifdef MODULE
extern
 EChannel
EChannel_create_attr_id(ec, attrs, conn_id)
EControlContext ec;
attr_list attrs;
char *conn_id;
{
    EChannel chan;

    chan = EChannel_create_attr(ec, attrs);
    if ((chan != NULL) && (conn_id != NULL))
	chan->conn_id = strdup(conn_id);
    else
	chan->conn_id = NULL;
    return chan;
}
#endif

extern
 EChannel
EChannel_create_attr(ec, attrs)
EControlContext ec;
attr_list attrs;
{
    CManager cm = ec->cm;
    EChannel chan = CMmalloc(sizeof(EChannel_struct));
    attr_list contact_list;

    ECtrace_out("EChannel_create", 1);

    chan->sink_count = 0;
    chan->source_count = 0;

    chan->rem_sink_list_len = 0;
    chan->rem_sink_list = CMmalloc(sizeof(struct _rem_attendr));

    chan->rem_source_list_len = 0;
    chan->rem_source_list = CMmalloc(sizeof(struct _rem_attendr));

    chan->rem_member_list_len = 0;
    chan->rem_member_list = CMmalloc(sizeof(struct _rem_attendr));
    chan->local_sink_list_len = 0;
    chan->local_sink_list = CMmalloc(sizeof(struct _local_sink));
    chan->local_source_list_len = 0;
    chan->local_source_list = CMmalloc(sizeof(struct _local_source));

    chan->der_list_len = 0;
    chan->der_chan_list = CMmalloc(sizeof(struct _derived_chan));
    chan->parent_list_len = 0;
    chan->parent_chan_list = CMmalloc(sizeof(EChannel));

    set_channel_id(&chan->global_id.global_channel_id, chan);
    contact_list = CMget_contact_list(cm);
    if (contact_list == NULL) {
	printf("Warning, no contact list.  Must do CManager_listen() before using channels.\n");
	CMfree(chan);
	return NULL;
    }
    contact_list = attr_join_lists(contact_list, attrs);
    chan->global_id.contact_string = attr_list_to_string(contact_list);
    chan->local_contact_string = attr_list_to_string(contact_list);
    free_attr_list(contact_list);
#ifdef MODULE
    sema_init(&chan->lock, 1);
#else
    chan->lock = thr_mutex_alloc();
#endif
    chan->ready = 1;
    chan->subscribe_func = (ECSubscribeFunction) NULL;

    if (ec->channel_list == NULL) {
	ec->channel_list = CMmalloc(sizeof(EChannel));
    } else {
	ec->channel_list =
	    (EChannel *) CMrealloc(ec->channel_list,
			     sizeof(EChannel) * (ec->channel_count + 1));
    }
    ec->channel_list[ec->channel_count++] = chan;

    ECtrace_out("EChannel_create returning channel %lx", 0, chan);
    chan->subcontext = NULL;
    chan->filter_data = NULL;
    chan->data_field_list = NULL;
    chan->data_format_list = NULL;

    chan->encode_buffer = create_IOBuffer();

    chan->cm = cm;
    /* eventv format included as a frequently used piece of data */
    chan->eventv_format = CMlookup_format(cm, eventv_msg_flds);
    chan->ec = ec;
    chan->type_format = NULL;
    chan->periodic_info = NULL;
#ifdef MODULE
    chan->out_filter = NULL;
    chan->in_filter = NULL;
    chan->conn_id = NULL;

    add_to_channel_pool(chan);	//Sandip
#endif
    return chan;
}

#ifdef MODULE
extern
 EChannel
EChannel_create_id(ec, conn_id)
EControlContext ec;
char *conn_id;
{
    EChannel chan;
    chan = EChannel_create_attr(ec, NULL);
    if ((chan != NULL) && (conn_id !=NULL))
	chan->conn_id = strdup(conn_id);
    else
	chan->conn_id = NULL;
    return chan;
}
#endif

extern
 EChannel
EChannel_create(ec)
EControlContext ec;
{
    return EChannel_create_attr(ec, NULL);
}

extern
void
EChannel_destroy(chan)
EChannel chan;
{
    int i;
    EControlContext ec = chan->ec;

    ECtrace_out("EChannel_destroy of channel %lx", 1, chan);
    if (chan->periodic_info != NULL) {
	DEperiodic_info_free(chan->periodic_info);
    }
    EChannel_lock(chan);
    for(i= 0; i< chan->rem_sink_list_len; i++) {
	free(chan->rem_sink_list[i].contact_string);
    }
    CMfree(chan->rem_sink_list);
    for(i= 0; i< chan->rem_source_list_len; i++) {
	free(chan->rem_source_list[i].contact_string);
    }
    CMfree(chan->rem_source_list);
    for(i= 0; i< chan->rem_member_list_len; i++) {
	free(chan->rem_member_list[i].contact_string);
    }
    CMfree(chan->rem_member_list);
    for(i = 0; i< chan->local_sink_list_len; i++) {
	if (chan->local_sink_list[i].context != NULL) {
	    free_IOcontext(chan->local_sink_list[i].context);
	    chan->local_sink_list[i].context = NULL;
	}
    }
    CMfree(chan->local_sink_list);
    if (chan->subcontext) CMfree_user_type_context(chan->cm, chan->subcontext);

    /* destroy and free the derived event channels */
    for (i = 0; i < chan->der_list_len; i++) {
	EChannel_destroy(chan->der_chan_list[i].derived_chan);
    }
    for (i=0; i< chan->parent_list_len; i++) {
	remove_derived_channel(chan->parent_chan_list[i], chan);
    }
    CMfree(chan->parent_chan_list);

    while(chan->local_source_list_len != 0) {
	int last = chan->local_source_list_len -1;
	internal_cancel_source_subscribe(chan->local_source_list[last].source_handle);
    }
    CMfree(chan->local_source_list);

    CMfree(chan->der_chan_list);
#ifndef MODULE
    thr_mutex_free(chan->lock);
#endif

    chan->rem_member_list_len = 0;
    chan->ready = 0;
    if (chan->encode_buffer != NULL) {
	free_IOBuffer(chan->encode_buffer);
    }

    for (i = 0; i < ec->channel_count; i++) {
	if (ec->channel_list[i] == chan) {
	    ec->channel_count--;
	    for (; i < (ec->channel_count); i++) {
		ec->channel_list[i] = ec->channel_list[i + 1];
	    }
	    break;
	}
    }
    if (chan->global_id.global_channel_id.channel) 
	CMfree(chan->global_id.global_channel_id.channel);

    CMfree(chan);
    ECtrace_out("EChannel_destroy", 0);
}

static void send_request_event ARGS((EChannel chan, int i));

#ifdef MODULE
static void send_remote_event_attr ARGS((EChannel chan, int i, void *event,
				    int event_len, attr_list attrs));
static void sendv_remote_event_attr ARGS((EChannel chan, int i,
				     IOEncodeVector eventV, attr_list attrs));
#else
static void send_remote_event ARGS((EChannel chan, int i, void *event,
				    int event_len));
static void sendv_remote_event ARGS((EChannel chan, int i,
				     IOEncodeVector eventV));
#endif

static void queue_event_for_dispatch ARGS((EChannel chan, int sink_index,
					   void *event, int event_len,
					   free_block_rec_p block_rec,
					   attr_list event_attrs));
static void queue_eventv_for_dispatch ARGS((EChannel chan, int sink_index,
					    IOEncodeVector eventv,
					    free_block_rec_p block_rec,
					    attr_list event_attrs));


static void send_sink_subscribe ARGS((EChannel chan, int i, int subscribe));

static void send_source_subscribe ARGS((EChannel chan, int i, int subscribe));

static void do_event_dispatch ARGS((EControlContext ec));

#ifdef MODULE
static void
internal_submit_general_event_attr ARGS((ECSourceHandle handle, void *event,
				    int event_len,
				    free_block_rec_p block_rec, attr_list attrs));
static void
internal_submit_general_eventV_attr ARGS((ECSourceHandle handle, 
				     IOEncodeVector eventV,
				     free_block_rec_p block_rec, attr_list attrs));
#endif

static void
internal_submit_general_event ARGS((ECSourceHandle handle, void *event,
				    int event_len,
				    free_block_rec_p block_rec));

static void
internal_submit_general_eventV ARGS((ECSourceHandle handle, 
				     IOEncodeVector eventV,
				     free_block_rec_p block_rec));


extern void
ECrequest_event(handle, req_try)
ECSinkHandle handle;
int req_try;
{
    int i = 0;
    EChannel chan = handle->chan;
    int cond = 0;

    EChannel_lock(chan);
    ECtrace_out("In ecrequest_event, handle %lx", 1, handle);
    for (i = 0; i < chan->rem_source_list_len; i++) {
	if (chan->rem_source_list[i].conn != NULL) {
	    EChannel_unlock(chan);
	    send_request_event(chan, i);
	    EChannel_lock(chan);
	}
    }

    for (i = 0; i < chan->local_source_list_len; i++) {
	if (chan->local_source_list[i].func) {
	    chan->local_source_list[i].func(chan->local_source_list[i].source_handle,
					    chan->local_source_list[i].client_data);
	}
    }

    /* get condition for req try */
    if (req_try) {
	cond = CMCondition_get(chan->cm, NULL);
	chan->local_sink_list[handle->sink_num].cond = cond;
    }
  
    EChannel_unlock(chan);

    /* wait on it */
    if (req_try) {
	if (CMCondition_wait(chan->cm, cond)) {
	    ECtrace_out("ECrequest_event sink is ready %lx", 3, chan);
	} else {
	    /* wait failed, conn is dead - is that so ? */
	    EChannel_destroy(chan);
	    ECtrace_out("ECrequest_event, wait failed", 3);
	}
    }

    ECtrace_out("exit ecrequest_event, handle %lx", 0, handle);
}

#ifdef MODULE
extern void *
ECset_rt_attributes(handle, rt_attrs)
ECSourceHandle handle;
ECRealTimeAttrs rt_attrs;
{
    handle->rt_attrs = *rt_attrs;
    handle->rt_attrs.rt_header[0] = 0x52544500;  /* identifier: RTE\0 */
    handle->rt_attrs.rt_header[1] = 0;
    return NULL;
}
#endif

#ifdef MODULE
extern void *
#else
extern void
#endif
ECsubmit_event(handle, event, event_len)
ECSourceHandle handle;
void *event;
int event_len;
{
    internal_submit_general_event(handle, event, event_len, NULL);
#ifdef MODULE
    if (current->kthread != 1)
      return (void *)handle->chan->ec->cm;
    else {
      do_event_dispatch(handle->chan->ec);
      return NULL;
    }
#endif
}

#ifdef MODULE
extern void *
ECsubmit_event_attr(handle, event, event_len, attrs)
ECSourceHandle handle;
void *event;
int event_len;
attr_list attrs;
{
    internal_submit_general_event_attr(handle, event, event_len, NULL, attrs);
    if (current->kthread != 1)
      return (void *)handle->chan->ec->cm;
    else {
      do_event_dispatch(handle->chan->ec);
      return NULL;
    }
}
#endif

#ifdef MODULE
extern void *
#else
extern void
#endif
ECsubmit_typed_event(handle, event)
ECSourceHandle handle;
void *event;
{
    internal_submit_general_event(handle, event, 0, NULL);
#ifdef MODULE
    if (current->kthread != 1)
      return (void *)handle->chan->ec->cm;
    else {
      do_event_dispatch(handle->chan->ec);
      return NULL;
    }
#endif
}

#ifdef MODULE
extern void *
ECsubmit_typed_event_attr(handle, event, attrs)
ECSourceHandle handle;
void *event;
attr_list attrs;
{
    internal_submit_general_event_attr(handle, event, 0, NULL, attrs);
    if (current->kthread != 1)
      return (void *)handle->chan->ec->cm;
    else {
      do_event_dispatch(handle->chan->ec);
      return NULL;
    }
}
#endif

#ifdef MODULE
extern void *
#else
extern void
#endif
ECsubmit_general_typed_event(handle, event, free_func)
ECSourceHandle handle;
void *event;
EventFreeFunction free_func;
{
    free_block_rec_p block_rec = get_free_block_func(handle->chan->ec,
						     event, free_func);
    internal_submit_general_event(handle, event, 0, block_rec);
    deref_block(block_rec);
#ifdef MODULE
    if (current->kthread != 1)
      return (void *)handle->chan->ec->cm;
    else {
      do_event_dispatch(handle->chan->ec);
      return NULL;
    }
#endif
}

#ifdef MODULE
extern void *
ECsubmit_general_typed_event_attr(handle, event, free_func, attrs)
ECSourceHandle handle;
void *event;
EventFreeFunction free_func;
attr_list attrs;
{
    free_block_rec_p block_rec = get_free_block_func(handle->chan->ec,
						     event, free_func);
    internal_submit_general_event_attr(handle, event, 0, block_rec, attrs);
    deref_block(block_rec);
    if (current->kthread != 1)
      return (void *)handle->chan->ec->cm;
    else {
      do_event_dispatch(handle->chan->ec);
      return NULL;
    }
}

#else

static void
int_submit_general_typed_event(handle, event, free_func)
ECSourceHandle handle;
void *event;
EventFreeFunction free_func;
{
    free_block_rec_p block_rec = get_free_block_func(handle->chan->ec,
						     event, free_func);
    internal_submit_general_event(handle, event, 0, block_rec);
    deref_block(block_rec);
}
#endif

#ifdef MODULE
static void
int_submit_general_typed_event_attr(handle, event, free_func, attrs)
ECSourceHandle handle;
void *event;
EventFreeFunction free_func;
attr_list attrs;
{
    free_block_rec_p block_rec = get_free_block_func(handle->chan->ec,
						     event, free_func);
    internal_submit_general_event_attr(handle, event, 0, block_rec, attrs);
    deref_block(block_rec);
}
#endif

#ifdef MODULE
extern void *
#else
extern void
#endif
ECsubmit_general_event(handle, event, event_len, free_func)
ECSourceHandle handle;
void *event;
int event_len;
EventFreeFunction free_func;
{
    free_block_rec_p block_rec = get_free_block_func(handle->chan->ec,
						     event, free_func);
    internal_submit_general_event(handle, event, event_len, block_rec);
    deref_block(block_rec);
#ifdef MODULE
    if (current->kthread != 1)
      return (void *)handle->chan->ec->cm;
    else {
      do_event_dispatch(handle->chan->ec);
      return NULL;
    }
#endif
}

#ifdef MODULE
extern void *
ECsubmit_general_event_attr(handle, event, event_len, free_func, attrs)
ECSourceHandle handle;
void *event;
int event_len;
EventFreeFunction free_func;
attr_list attrs;
{
    free_block_rec_p block_rec = get_free_block_func(handle->chan->ec,
						     event, free_func);
    internal_submit_general_event_attr(handle, event, event_len, block_rec, attrs);
    deref_block(block_rec);
    if (current->kthread != 1)
      return (void *)handle->chan->ec->cm;
    else {
      do_event_dispatch(handle->chan->ec);
      return NULL;
    }
}
#endif

#ifdef MODULE
extern void *
#else
extern void
#endif
ECsubmit_general_eventV(handle, eventV, free_func)
ECSourceHandle handle;
IOEncodeVector eventV;
EventFreeFunction free_func;
{
    free_block_rec_p block_rec = get_free_block_func(handle->chan->ec,
						     (void*)eventV, 
						     free_func);
    internal_submit_general_eventV(handle, eventV, block_rec);
    deref_block(block_rec);
#ifdef MODULE
    if (current->kthread != 1)
      return (void *)handle->chan->ec->cm;
    else {
      do_event_dispatch(handle->chan->ec);
      return NULL;
    }
#endif
}

#ifdef MODULE
extern void *
ECsubmit_general_eventV_attr(handle, eventV, free_func, attrs)
ECSourceHandle handle;
IOEncodeVector eventV;
EventFreeFunction free_func;
attr_list attrs;
{
    free_block_rec_p block_rec = get_free_block_func(handle->chan->ec,
						     (void*)eventV, 
						     free_func);
    internal_submit_general_eventV_attr(handle, eventV, block_rec, attrs);
    deref_block(block_rec);
    if (current->kthread != 1)
      return (void *)handle->chan->ec->cm;
    else {
      do_event_dispatch(handle->chan->ec);
      return NULL;
    }
}
#endif

#ifdef MODULE
extern void *
#else
extern void
#endif
ECsubmit_eventV(handle, eventV)
ECSourceHandle handle;
IOEncodeVector eventV;
{
    internal_submit_general_eventV(handle, eventV, NULL);
#ifdef MODULE
    if (current->kthread != 1)
      return (void *)handle->chan->ec->cm;
    else {
      do_event_dispatch(handle->chan->ec);
      return NULL;
    }
#endif
}

#ifdef MODULE
extern void *
ECsubmit_eventV_attr(handle, eventV, attrs)
ECSourceHandle handle;
IOEncodeVector eventV;
attr_list attrs;
{
    internal_submit_general_eventV_attr(handle, eventV, NULL, attrs);
    if (current->kthread != 1)
      return (void *)handle->chan->ec->cm;
    else {
      do_event_dispatch(handle->chan->ec);
      return NULL;
    }
}
#endif

#ifdef MODULE
extern void *
#else
extern void
#endif
ECsubmit_eventV_block(handle, eventV, free_block)
ECSourceHandle handle;
IOEncodeVector eventV;
void *free_block;
{
    free_block_rec_p block_rec = 
	get_free_block_func(handle->chan->ec, free_block, 
			    (EventFreeFunction)NULL);
    internal_submit_general_eventV(handle, eventV, block_rec);
    deref_block(block_rec);
#ifdef MODULE
    if (current->kthread != 1)
      return (void *)handle->chan->ec->cm;
    else {
      do_event_dispatch(handle->chan->ec);
      return NULL;
    }
#endif
}

#ifdef MODULE
extern void *
ECsubmit_eventV_block_attr(handle, eventV, free_block, attrs)
ECSourceHandle handle;
IOEncodeVector eventV;
void *free_block;
attr_list attrs;
{
    free_block_rec_p block_rec = 
	get_free_block_func(handle->chan->ec, free_block, 
			    (EventFreeFunction)NULL);
    internal_submit_general_eventV_attr(handle, eventV, block_rec, attrs);
    deref_block(block_rec);
    if (current->kthread != 1)
      return (void *)handle->chan->ec->cm;
    else {
      do_event_dispatch(handle->chan->ec);
      return NULL;
    }
}
#endif

static int
sink_valid(local_sink_ptr sink)
{
    return ((sink->func != (ECHandlerFunction)NULL) || 
	    (sink->multi_sink_handle != NULL));
}

#ifdef MODULE
static void
internal_submit_general_event(handle, event, event_len, block_rec)
ECSourceHandle handle;
void *event;
int event_len;
free_block_rec_p block_rec;
{
    internal_submit_general_event_attr(handle, event, event_len, block_rec, NULL);
}

static void
internal_submit_general_event_attr(handle, kevent, kevent_len, block_rec, attrs)
ECSourceHandle handle;
void *kevent;
int kevent_len;
free_block_rec_p block_rec;
attr_list attrs;

#else

static void
internal_submit_general_event(handle, event, event_len, block_rec)
ECSourceHandle handle;
void *event;
int event_len;
free_block_rec_p block_rec;
#endif

{
#ifdef MODULE
    void *event;
    int event_len;
    int real_time = 0;
    struct _ECRealTimeAttrs rt_attrs;
    struct timeval pd;
    struct timeval now;
    long rttime;
#endif
    int i = 0;
    IOEncodeVector encodeV = NULL;
    EChannel_lock(handle->chan);

#ifdef MODULE
    if ((handle->rt_attrs.deadline.tv_sec != 0) || 
		(handle->rt_attrs.deadline.tv_usec != 0) ||
		(handle->rt_attrs.priority != -1)) {
	real_time = 1;
	event = kmalloc(kevent_len + sizeof(struct _ECRealTimeAttrs), GFP_KERNEL);
	memcpy(event, kevent, kevent_len);
	event_len = kevent_len + sizeof(struct _ECRealTimeAttrs);
    } else {
	event = kevent;
	event_len = kevent_len;
    }
#endif

    ECtrace_out("In ecsubmit_general_event, handle %lx, event %lx", 1, handle,
	      event);

#ifdef MODULE
    if (handle->chan->out_filter != NULL) {
      int (*func)(EChannel, attr_list, void *, int);
      func = handle->chan->out_filter;
      event_len = func(handle->chan, attrs, event, event_len);
    }
#endif

    for (i = 0; i < handle->chan->rem_sink_list_len; i++) {
	if (handle->ioformat == NULL) {
	    /* untyped event */
#ifdef MODULE
	    if ((handle->chan->rem_sink_list[i].conn != NULL) && 
		(handle->chan->rem_sink_list[i].block_event != 1)) {
		if (real_time == 1) {
		    rt_attrs = handle->rt_attrs;
		    if (rt_attrs.priority == -1) {
		    pd = handle->chan->rem_sink_list[i].pd;
		    do_gettimeofday(&now);
		    rt_attrs.timestamp = now;
		    if ((rt_attrs.period.tv_sec != 0) || 
			(rt_attrs.period.tv_usec != 0)) {
		        rt_attrs.deadline.tv_sec += rt_attrs.period.tv_sec;
			rt_attrs.deadline.tv_usec += rt_attrs.period.tv_usec;
			while (rt_attrs.deadline.tv_usec >= 1000000) {
			    rt_attrs.deadline.tv_usec -= 1000000;
			    rt_attrs.deadline.tv_sec += 1;
			}
			rt_attrs.est.tv_sec += rt_attrs.period.tv_sec;
			rt_attrs.est.tv_usec += rt_attrs.period.tv_usec;
			while (rt_attrs.est.tv_usec >= 1000000) {
			    rt_attrs.est.tv_usec -= 1000000;
			    rt_attrs.est.tv_sec += 1;
			}
			rt_attrs.lst.tv_sec += rt_attrs.period.tv_sec;
			rt_attrs.lst.tv_usec += rt_attrs.period.tv_usec;
			while (rt_attrs.lst.tv_usec >= 1000000) {
			    rt_attrs.lst.tv_usec -= 1000000;
			    rt_attrs.lst.tv_sec += 1;
			}
			handle->rt_attrs.timestamp = now;
			handle->rt_attrs.deadline = rt_attrs.deadline;
			handle->rt_attrs.est = rt_attrs.est;
			handle->rt_attrs.lst = rt_attrs.lst;
		    } else if (rt_attrs.type == 1) {
		        rt_attrs.deadline.tv_sec += now.tv_sec; 
			rt_attrs.deadline.tv_usec += now.tv_usec; 
			while (rt_attrs.deadline.tv_usec >= 1000000) {
			    rt_attrs.deadline.tv_usec -= 1000000;
			    rt_attrs.deadline.tv_sec += 1;
			}
		        rt_attrs.est.tv_sec = now.tv_sec + 	
				rt_attrs.est.tv_sec;
			rt_attrs.est.tv_usec = now.tv_usec + 
				rt_attrs.est.tv_usec;
			while (rt_attrs.est.tv_usec >= 1000000) {
			    rt_attrs.est.tv_usec -= 1000000;
			    rt_attrs.est.tv_sec += 1;
			}
		        rt_attrs.lst.tv_sec = now.tv_sec + 	
				rt_attrs.lst.tv_sec;
			rt_attrs.lst.tv_usec = now.tv_usec + 
				rt_attrs.lst.tv_usec;
			while (rt_attrs.lst.tv_usec >= 1000000) {
			    rt_attrs.lst.tv_usec -= 1000000;
			    rt_attrs.lst.tv_sec += 1;
			}
			handle->rt_attrs.timestamp = now;
		    }
		    if ((pd.tv_sec != 0) || (pd.tv_usec != 0)) {
			rttime = rt_attrs.deadline.tv_sec * 1000000;
			rttime += rt_attrs.deadline.tv_usec;
			rttime -= ((pd.tv_sec * 1000000) + pd.tv_usec);
		        if (rttime < 0) rttime = 0;
		        rt_attrs.deadline.tv_sec = 0; 	
			rt_attrs.deadline.tv_usec =  0;
			while (rttime >= 1000000) {
			    rt_attrs.deadline.tv_sec++;
			    rttime -= 1000000;
			}
			rt_attrs.deadline.tv_usec = rttime;
		    }
		    }
		    memcpy(event+kevent_len, (void *)&rt_attrs, 
			sizeof(struct _ECRealTimeAttrs));
		}
#else
	    if (handle->chan->rem_sink_list[i].conn != NULL) {
#endif
		EChannel_unlock(handle->chan);
#ifdef MODULE
		send_remote_event_attr(handle->chan, i, event, event_len, attrs);
#else
		send_remote_event(handle->chan, i, event, event_len);
#endif
		EChannel_lock(handle->chan);
	    }
	} else {
	    /* typed event */
	    if (encodeV == NULL) {
		encodeV = 
		    encode_IOcontext_vectorB((IOContext)iofile_of_IOformat(handle->ioformat),
					     handle->chan->encode_buffer,
					     handle->ioformat,
					     event);
		if (encodeV == NULL) {
		    IOperror((IOFile)handle->chan->subcontext, 
			     "encode_IOcontext");
		}
	    }
#ifdef MODULE
	    if ((handle->chan->rem_sink_list[i].conn != NULL) && 
		(handle->chan->rem_sink_list[i].block_event != 1)) {
/* CP: Don't know if that would work below, real-time events for typed events
not tested yet... */
		if (real_time == 1) {
		    rt_attrs = handle->rt_attrs;
		    if (rt_attrs.priority == -1) {
		    pd = handle->chan->rem_sink_list[i].pd;
		    do_gettimeofday(&now);
		    rt_attrs.timestamp = now;
		    if ((rt_attrs.period.tv_sec != 0) || 
			(rt_attrs.period.tv_usec != 0)) {
		        rt_attrs.deadline.tv_sec += rt_attrs.period.tv_sec;
			rt_attrs.deadline.tv_usec += rt_attrs.period.tv_usec;
			while (rt_attrs.deadline.tv_usec >= 1000000) {
			    rt_attrs.deadline.tv_usec -= 1000000;
			    rt_attrs.deadline.tv_sec += 1;
			}
			rt_attrs.est.tv_sec += rt_attrs.period.tv_sec;
			rt_attrs.est.tv_usec += rt_attrs.period.tv_usec;
			while (rt_attrs.est.tv_usec >= 1000000) {
			    rt_attrs.est.tv_usec -= 1000000;
			    rt_attrs.est.tv_sec += 1;
			}
			rt_attrs.lst.tv_sec += rt_attrs.period.tv_sec;
			rt_attrs.lst.tv_usec += rt_attrs.period.tv_usec;
			while (rt_attrs.lst.tv_usec >= 1000000) {
			    rt_attrs.lst.tv_usec -= 1000000;
			    rt_attrs.lst.tv_sec += 1;
			}
			handle->rt_attrs.timestamp = now;
			handle->rt_attrs.deadline = rt_attrs.deadline;
			handle->rt_attrs.est = rt_attrs.est;
			handle->rt_attrs.lst = rt_attrs.lst;
		    } else if (rt_attrs.type == 1) {
		        rt_attrs.deadline.tv_sec += now.tv_sec + 	
				rt_attrs.deadline.tv_sec;
			rt_attrs.deadline.tv_usec += now.tv_usec + 
				rt_attrs.deadline.tv_usec;
			while (rt_attrs.deadline.tv_usec >= 1000000) {
			    rt_attrs.deadline.tv_usec -= 1000000;
			    rt_attrs.deadline.tv_sec += 1;
			}
		        rt_attrs.est.tv_sec += now.tv_sec + 	
				rt_attrs.est.tv_sec;
			rt_attrs.est.tv_usec += now.tv_usec + 
				rt_attrs.est.tv_usec;
			while (rt_attrs.est.tv_usec >= 1000000) {
			    rt_attrs.est.tv_usec -= 1000000;
			    rt_attrs.est.tv_sec += 1;
			}
		        rt_attrs.lst.tv_sec += now.tv_sec + 	
				rt_attrs.lst.tv_sec;
			rt_attrs.lst.tv_usec += now.tv_usec + 
				rt_attrs.lst.tv_usec;
			while (rt_attrs.lst.tv_usec >= 1000000) {
			    rt_attrs.lst.tv_usec -= 1000000;
			    rt_attrs.lst.tv_sec += 1;
			}
			handle->rt_attrs.timestamp = now;
			handle->rt_attrs.deadline = rt_attrs.deadline;
			handle->rt_attrs.est = rt_attrs.est;
			handle->rt_attrs.lst = rt_attrs.lst;
		    }
		    if ((pd.tv_sec != 0) || (pd.tv_usec != 0)) {
			rttime = rt_attrs.deadline.tv_sec * 1000000;
			rttime += rt_attrs.deadline.tv_usec;
			rttime -= ((pd.tv_sec * 1000000) + pd.tv_usec);
		        if (rttime < 0) rttime = 0;
		        rt_attrs.deadline.tv_sec = 0; 	
			rt_attrs.deadline.tv_usec =  0;
			while (rttime >= 1000000) {
			    rt_attrs.deadline.tv_sec++;
			    rttime -= 1000000;
			}
			rt_attrs.deadline.tv_usec = rttime;
		    }
		    }
		    memcpy(event+kevent_len, &rt_attrs, 
			sizeof(struct _ECRealTimeAttrs));
		}
#else
	    if (handle->chan->rem_sink_list[i].conn != NULL) {
#endif
		EChannel_unlock(handle->chan);
#ifdef MODULE
		sendv_remote_event_attr(handle->chan, i, encodeV, attrs);
#else
		sendv_remote_event(handle->chan, i, encodeV);
#endif
		EChannel_lock(handle->chan);
	    }
	}
    }
    EContext_lock(handle->chan->ec);
    for (i = 0; i < handle->chan->local_sink_list_len; i++) {
#ifdef MODULE
	if (sink_valid(&handle->chan->local_sink_list[i]) &&
	   (handle->chan->local_sink_list[i].block_event != 1)) {
	    IOFormat sink_format = handle->chan->local_sink_list[i].format;
	    if (real_time == 1) {
	        rt_attrs = handle->rt_attrs;
		if (rt_attrs.priority == -1) {
	        pd = handle->chan->rem_sink_list[i].pd;
	        do_gettimeofday(&now);
	        rt_attrs.timestamp = now;
	        if ((rt_attrs.period.tv_sec != 0) || 
		    (rt_attrs.period.tv_usec != 0)) {
		    rt_attrs.deadline.tv_sec += rt_attrs.period.tv_sec;
		    rt_attrs.deadline.tv_usec += rt_attrs.period.tv_usec;
		    while (rt_attrs.deadline.tv_usec >= 1000000) {
		    	rt_attrs.deadline.tv_usec -= 1000000;
		    	rt_attrs.deadline.tv_sec += 1;
		    }
		    rt_attrs.est.tv_sec += rt_attrs.period.tv_sec;
		    rt_attrs.est.tv_usec += rt_attrs.period.tv_usec;
		    while (rt_attrs.est.tv_usec >= 1000000) {
		        rt_attrs.est.tv_usec -= 1000000;
		    	rt_attrs.est.tv_sec += 1;
		    }
		    rt_attrs.lst.tv_sec += rt_attrs.period.tv_sec;
		    rt_attrs.lst.tv_usec += rt_attrs.period.tv_usec;
		    while (rt_attrs.lst.tv_usec >= 1000000) {
		    	rt_attrs.lst.tv_usec -= 1000000;
		    	rt_attrs.lst.tv_sec += 1;
		    }
		    handle->rt_attrs.timestamp = now;
		    handle->rt_attrs.deadline = rt_attrs.deadline;
		    handle->rt_attrs.est = rt_attrs.est;
		    handle->rt_attrs.lst = rt_attrs.lst;
	    	} else if (rt_attrs.type == 1) {
	            rt_attrs.deadline.tv_sec += now.tv_sec + 	
	  		rt_attrs.deadline.tv_sec;
		    rt_attrs.deadline.tv_usec += now.tv_usec + 
			rt_attrs.deadline.tv_usec;
		    while (rt_attrs.deadline.tv_usec >= 1000000) {
		    	rt_attrs.deadline.tv_usec -= 1000000;
		    	rt_attrs.deadline.tv_sec += 1;
		    }
		    rt_attrs.est.tv_sec += now.tv_sec + 	
			rt_attrs.est.tv_sec;
		    rt_attrs.est.tv_usec += now.tv_usec + 
			rt_attrs.est.tv_usec;
		    while (rt_attrs.est.tv_usec >= 1000000) {
		    	rt_attrs.est.tv_usec -= 1000000;
		    	rt_attrs.est.tv_sec += 1;
		    }
		    rt_attrs.lst.tv_sec += now.tv_sec + 	
			rt_attrs.lst.tv_sec;
		    rt_attrs.lst.tv_usec += now.tv_usec + 
			rt_attrs.lst.tv_usec;
		    while (rt_attrs.lst.tv_usec >= 1000000) {
		    	rt_attrs.lst.tv_usec -= 1000000;
		    	rt_attrs.lst.tv_sec += 1;
		    }
		    handle->rt_attrs.timestamp = now;
		    handle->rt_attrs.deadline = rt_attrs.deadline;
		    handle->rt_attrs.est = rt_attrs.est;
		    handle->rt_attrs.lst = rt_attrs.lst;
		}
		}
		memcpy(event+kevent_len, &rt_attrs, 
		    sizeof(struct _ECRealTimeAttrs));
	    }
#else
	if (sink_valid(&handle->chan->local_sink_list[i])) {
	    IOFormat sink_format = handle->chan->local_sink_list[i].format;
#endif
	    if (!handle->ioformat || !sink_format ||
		global_name_eq(handle->ioformat, sink_format)) {
#ifdef MODULE
		queue_event_for_dispatch(handle->chan, i, event, event_len,
					 block_rec, attrs);
#else
		queue_event_for_dispatch(handle->chan, i, event, event_len,
					 block_rec, NULL);
#endif
	    } else {
		int len;
		char *buffer;
		IOContext context;
		IOBuffer iobuffer = create_IOBuffer();
		free_block_rec_p block_rec = 
		    get_free_block_iobuffer(handle->chan->ec, iobuffer);

		context = (IOContext)
		    iofile_of_IOformat(handle->ioformat);
		buffer = encode_IOcontext_bufferB(context, handle->ioformat, 
						  iobuffer, event, &len);
#ifdef MODULE
		queue_event_for_dispatch(handle->chan, i, buffer, len, 
					 block_rec, attrs);
#else
		queue_event_for_dispatch(handle->chan, i, buffer, len, 
					 block_rec, NULL);
#endif
	    }
	}
    }
    EContext_unlock(handle->chan->ec);
    for (i = 0; i < handle->der_list_len; i++) {
/* CP: No real time events for derived channels yet. */
	int type_len = handle->der_source_list[i].target_type_len;
	FilterFunc func = handle->der_source_list[i].filter;
	EChannel derived_chan = handle->der_source_list[i].derived_chan;
	ECSourceHandle der_handle = handle->der_source_list[i].source_handle;

	ECtrace_out("working on derived chan %lx", 3, derived_chan);
	ECtrace_out("filter is %lx, has output %d\n", 3, (long) func,
		    handle->der_source_list[i].has_output_param);
	if (!EChas_sinks(der_handle)) {
	    ECtrace_out("Output channel has no sinks", 3);
	    continue;
	}
	if (handle->der_source_list[i].has_output_param) {
	    int tmp;
	    void *event2 = CMmalloc(type_len);
	    memset(event2, 0, type_len);
	    tmp = (func)(event, event2, derived_chan->filter_data);
	    if (tmp != 0) {
		ECtrace_out("Derivation function returned %d, submitting event", 
			    3, tmp);
#ifdef MODULE
		int_submit_general_typed_event_attr(der_handle, event2, DFreeMM, attrs);
#else
		int_submit_general_typed_event(der_handle, event2, free);
#endif
	    } else {
		ECtrace_out("Derivation function returned 0, no event submitted", 3);
	    }
	} else {
	    int tmp = (func)(event, NULL, derived_chan->filter_data);
	    if (tmp != 0) {
		ECtrace_out("Derivation function returned %d, submitting event", 
			    3, tmp);
#ifdef MODULE
		internal_submit_general_event_attr(der_handle, event, 0, 
					      block_rec, attrs);
#else
		internal_submit_general_event(der_handle, event, 0, 
					      block_rec);
#endif
	    } else {
		ECtrace_out("Derivation function returned 0, no event submitted", 3);
	    }
	}
    }

    ECtrace_out("exit ecsubmit_general_event, handle %lx", 0, handle);
    EChannel_unlock(handle->chan);
#ifdef MODULE
    if (real_time == 1) kfree(event);
#endif
}

#ifdef MODULE
static void
internal_submit_general_eventV(handle, eventV, block_rec)
ECSourceHandle handle;
IOEncodeVector eventV;
free_block_rec_p block_rec;
{
    internal_submit_general_eventV_attr(handle, eventV, block_rec, NULL);
}

static void
internal_submit_general_eventV_attr(handle, eventV, block_rec, attrs)
ECSourceHandle handle;
IOEncodeVector eventV;
free_block_rec_p block_rec;
attr_list attrs;

#else

static void
internal_submit_general_eventV(handle, eventV, block_rec)
ECSourceHandle handle;
IOEncodeVector eventV;
free_block_rec_p block_rec;
#endif

{
    int i = 0;
    EChannel_lock(handle->chan);
    ECtrace_out("In ecsubmit_general_eventV, handle %lx, eventV %lx", 1, handle, eventV);
    for (i = 0; i < handle->chan->rem_sink_list_len; i++) {
	if (handle->chan->rem_sink_list[i].conn != NULL) {
	    EChannel_unlock(handle->chan);
#ifdef MODULE
	    sendv_remote_event_attr(handle->chan, i, eventV, attrs);
#else
	    sendv_remote_event(handle->chan, i, eventV);
#endif
	    EChannel_lock(handle->chan);
	}
    }
    EContext_lock(handle->chan->ec);
    for (i = 0; i < handle->chan->local_sink_list_len; i++) {
	if (sink_valid(&handle->chan->local_sink_list[i])) {
#ifdef MODULE
	    queue_eventv_for_dispatch(handle->chan, i, eventV, block_rec,
				      attrs);
#else
	    queue_eventv_for_dispatch(handle->chan, i, eventV, block_rec,
				      NULL);
#endif
	}
    }
    EContext_unlock(handle->chan->ec);
    for (i = 0; i < handle->chan->der_list_len; i++) {
	ECSourceHandle der_handle;
	ECtrace_out("working on derivitive", 3);
	der_handle = handle->der_source_list[i].source_handle;
#ifdef MODULE
	ECsubmit_eventV_attr(der_handle, eventV, attrs);
#else
	ECsubmit_eventV(der_handle, eventV);
#endif
    }
    ECtrace_out("exit ecsubmit_general_eventV, handle %lx", 0, handle);
    EChannel_unlock(handle->chan);
}

extern int
EChas_sinks(handle)
ECSourceHandle handle;
{
    int i;
    int sinks = 0;
    ECSourceHandle der_handle;
    sinks += handle->chan->rem_sink_list_len;
    sinks += handle->chan->local_sink_list_len;

    if (handle->chan->der_list_len == 0) {
	return sinks;
    }
    der_handle = new_source_handle();
    /* check derived event channels */
    for (i = 0; i < handle->chan->der_list_len; i++) {
	der_handle->chan = handle->chan->der_chan_list[i].derived_chan;
	sinks += EChas_sinks(der_handle);
    }
    CMfree(der_handle);
    return sinks;
}

static int
num_local_sinks(chan)
EChannel chan;
{
    int i;
    int count = 0;
    for (i = 0; i< chan->local_sink_list_len; i++) {
	if (sink_valid(&chan->local_sink_list[i])) {
	    count++;
	}
    }
    if (chan->der_list_len != 0) {
	for (i = 0; i < chan->der_list_len; i++) {
	    count += num_local_sinks(chan->der_chan_list[i].derived_chan);
	}
    }
    return count;
}

extern int
ECQCount(ECEventQueueList queues)
{
    return queues->queue_count;
}

extern int
ECQEventCount(ECEventQueueList queues, int queue)
{
    if (queue > queues->queue_count) return -1;
    return queues->event_queues[queue].event_count;
}

extern void *
ECEvent_get_ptr(ECEventRec event_rec)
{
    return event_rec->event;
}

extern ECEventRec
ECQget_event_rec(ECEventQueueList queues, int queue, int event_number)
{
    if (queue > queues->queue_count) return NULL;
    if (event_number > queues->event_queues[queue].event_count) return NULL;
    return &(queues->event_queues[queue].queue[event_number]);
}

static ECSinkHandle
internal_sink_subscribe ARGS((EChannel chan, EControlContext ec,
			      multi_sink_info multi_sink_handle,
			      ECHandlerFunction func, 
			      void *client_data));

static ECSinkHandle
internal_typed_sink_subscribe(chan, ec, multi_sink_handle, 
			      field_list, format_list, func, client_data)
EChannel chan;
EControlContext ec;
multi_sink_info multi_sink_handle;
IOFieldList field_list;
CMFormatList format_list;
ECTypedHandlerFunction func;
void *client_data;
{
    IOFieldList type_field_list;
    ECSinkHandle handle;
    IOContext context;
    char format_name[] = "anonymous";

    if (chan->type_format == NULL) {
	fprintf(stderr, "Typed subscribe to an untyped channel not allowed\n");
	return NULL;
    }

    if (has_conversion_IOformat(chan->type_format)) {
	type_field_list = field_list_of_IOformat(chan->type_format);

	if (!compare_field_lists(type_field_list, field_list)) {
	    fprintf(stderr, "Incompatible field lists for the same incoming format in different ECtyped_sink_subscribe() calls.\n");
	    return NULL;
	}
    }

    handle = internal_sink_subscribe(chan, ec, NULL, 
				     (ECHandlerFunction) func,
				     client_data);
    chan->local_sink_list[handle->sink_num].context = context =
	CMget_user_type_context(ec->cm);
    chan->local_sink_list[handle->sink_num].format = 
      CMregister_user_format(ec->cm, context, format_name, field_list, 
                             format_list);

    chan->local_sink_list[handle->sink_num].multi_sink_handle =
	multi_sink_handle;
    chan->local_sink_list[handle->sink_num].field_list = field_list;
    chan->local_sink_list[handle->sink_num].format_list = format_list;
    return handle;
    
}

static
ECSinkHandle
internal_sink_multi_subscribe(sink_specs, ec, func, client_data)
ECMultiSubscribePtr sink_specs;
EControlContext ec;
ECMultiHandlerFunction func;
void *client_data;
{
    multi_sink_info multi_sink;
    ECEventQueueList queue_list;
    
    int channel_count = 0, i;
    multi_sink = malloc(sizeof(*multi_sink));
    queue_list = malloc(sizeof(*queue_list));
    while (sink_specs[channel_count].channel != NULL) channel_count++;

    multi_sink->func = func;
    multi_sink->client_data = client_data;
    multi_sink->queue_list = queue_list;
    queue_list->queue_count = channel_count;
    queue_list->event_queues = malloc(sizeof(queue_list->event_queues[0])
				      * channel_count);
    for (i=0; i< channel_count; i++) {
	queue_list->event_queues[i].event_count = 0;
	queue_list->event_queues[i].queue_size = 1;
	queue_list->event_queues[i].queue = 
	    malloc(sizeof(queue_list->event_queues[i].queue[0]));
	queue_list->event_queues[i].sink_handle = 
	    internal_typed_sink_subscribe(sink_specs[i].channel, ec, 
					  multi_sink,
					  sink_specs[i].field_list,
					  sink_specs[i].subformat_list,
					  (ECTypedHandlerFunction)NULL, 
					  NULL);
    }
    return queue_list->event_queues[0].sink_handle;
}

#ifdef MODULE
extern ECSinkHandle
ECsink_typed_subscribe_filter(chan, field_list, format_list, func, client_data,
	filter)
EChannel chan;
IOFieldList field_list;
CMFormatList format_list;
ECTypedHandlerFunction func;
void *client_data;
char *filter;
{
    int i = 0;

    while ((i < MAX_FILTERS) && (strcmp(filters[i].module, filter) != 0))
      i++;
    if (i < MAX_FILTERS) {
	chan->out_filter = filters[i].outfunc;
	chan->in_filter = filters[i].infunc;
    }

    return internal_typed_sink_subscribe(chan, chan->ec, 
					 (multi_sink_info) NULL, field_list,
					 format_list, func, client_data);
}
#endif

extern ECSinkHandle
ECsink_typed_subscribe(chan, field_list, format_list, func, client_data)
EChannel chan;
IOFieldList field_list;
CMFormatList format_list;
ECTypedHandlerFunction func;
void *client_data;
{
    return internal_typed_sink_subscribe(chan, chan->ec, 
					 (multi_sink_info) NULL, field_list,
					 format_list, func, client_data);
}

extern ECSinkHandle
ECsink_typed_subscribe_context(chan, field_list, format_list, func, 
			       client_data, ec)
EChannel chan;
IOFieldList field_list;
CMFormatList format_list;
ECTypedHandlerFunction func;
void *client_data;
EControlContext ec;
{
    return internal_typed_sink_subscribe(chan, ec, (multi_sink_info)NULL,
					 field_list, format_list, 
					 func, client_data);
}

extern ECSinkHandle
ECsink_multi_subscribe(sink_specs, ec, func, client_data)
ECMultiSubscribePtr sink_specs;
EControlContext ec;
ECMultiHandlerFunction func;
void *client_data;
{
    return internal_sink_multi_subscribe(sink_specs, ec, func, client_data);
}

#ifdef MODULE
extern ECSinkHandle
ECsink_subscribe_filter(chan, func, client_data, filter)
EChannel chan;
ECHandlerFunction func;
void *client_data;
char *filter;
{
    int i = 0;

    while ((i < MAX_FILTERS) && (strcmp(filters[i].module, filter) != 0))
      i++;
    if (i < MAX_FILTERS) {
	chan->out_filter = filters[i].outfunc;
	chan->in_filter = filters[i].infunc;
    }

    return internal_sink_subscribe(chan, chan->ec, NULL, func, client_data);
}
#endif

extern ECSinkHandle
ECsink_subscribe(chan, func, client_data)
EChannel chan;
ECHandlerFunction func;
void *client_data;
{
    return internal_sink_subscribe(chan, chan->ec, NULL, func, client_data);
}

extern ECSinkHandle
ECsink_subscribe_context(chan, func, client_data, ec)
EChannel chan;
ECHandlerFunction func;
void *client_data;
EControlContext ec;
{
    return internal_sink_subscribe(chan, ec, NULL, func, client_data);
}

extern void
ECcancel_sink_subscribe(handle)
ECSinkHandle handle;
{
    EChannel chan = handle->chan;
    EChannel_lock(handle->chan);

    ECtrace_out("ECcancel_sink_subscribe, chan %lx", 1, chan);

    if (num_local_sinks(chan) == 1) {
	int i;
	ECtrace_out("working on remote members", 3);
	for (i = 0; i < chan->rem_member_list_len; i++) {
	    if (chan->rem_member_list[i].conn != NULL) {
		EChannel_unlock(chan);
		send_sink_subscribe(chan, i, 0); /* false means unsubscribe */
		EChannel_lock(chan);
	    }
	}
    }
    if (chan->local_sink_list[handle->sink_num].context != NULL) {
        CMfree_user_type_context (chan->cm,
                                  chan->local_sink_list[handle->sink_num].context);
	chan->local_sink_list[handle->sink_num].context = NULL;
    }
    chan->local_sink_list[handle->sink_num].func = (ECHandlerFunction)NULL;
    chan->local_sink_list[handle->sink_num].multi_sink_handle = NULL;
    ECtrace_out("ECcancel_sink_subscribe, chan %lx", 0, chan);
    EChannel_unlock(chan);
    CMfree(handle);
}
    
static ECSinkHandle
internal_sink_subscribe(chan, ec, multi_sink_handle, func, client_data)
EChannel chan;
EControlContext ec;
multi_sink_info multi_sink_handle;
ECHandlerFunction func;
void *client_data;
{
    int i;
    ECSinkHandle handle = CMmalloc(sizeof(*handle));

    ECtrace_out("ECsink_subscribe, chan %lx", 1, chan);
    handle->chan = chan;
    EChannel_lock(handle->chan);
    ECtrace_out("working on remote members", 3);
    for (i = 0; i < chan->rem_member_list_len; i++) {
	if (chan->rem_member_list[i].conn != NULL) {
	    EChannel_unlock(chan);
	    send_sink_subscribe(chan, i, 1); /* true means "subscribe" */
	    EChannel_lock(chan);
	}
    }
    chan->local_sink_list =
	CMrealloc(chan->local_sink_list,
	   sizeof(struct _local_sink) * (chan->local_sink_list_len + 1));
    chan->local_sink_list[chan->local_sink_list_len].func = func;
    chan->local_sink_list[chan->local_sink_list_len].client_data = client_data;
    chan->local_sink_list[chan->local_sink_list_len].ec = ec;
    chan->local_sink_list[chan->local_sink_list_len].context = NULL;
    chan->local_sink_list[chan->local_sink_list_len].multi_sink_handle =
	multi_sink_handle;
#ifdef MODULE
    chan->local_sink_list[chan->local_sink_list_len].block_event = 0;
    if (chan->conn_id != NULL) 
        chan->local_sink_list[chan->local_sink_list_len].conn_id = 
	    strdup(chan->conn_id);
    else
        chan->local_sink_list[chan->local_sink_list_len].conn_id = NULL;
#endif
    chan->local_sink_list[chan->local_sink_list_len].cond = 0;
    handle->sink_num = chan->local_sink_list_len;

    (chan->local_sink_list_len)++;
    EChannel_unlock(chan);
    if (chan->subscribe_func) {
	chan->subscribe_func(1, chan->rem_sink_list_len,
			     chan->subscribe_client_data);
    }
    if (channel_verbose > 0) {
	printf("After Subscribe\n");
	dump_EChannel(chan);
    }
    ECtrace_out("ECsink_subscribe chan %lx, return handle %lx", 0, chan, handle);
    return handle;
}

/*
 *   Extended with (func, client_data) to support a pull model. The idea is 
 *   that the handler would be executed when the source get a request and 
 *   it would submit an event as part of its execution - fabianb
 *
 */

extern ECSourceHandle
ECsource_subscribe(chan)
EChannel chan;
{
    return ECPsource_subscribe(chan, (ECRequestHandlerFunction)NULL, NULL);
}

#ifdef MODULE
extern ECSourceHandle
ECsource_subscribe_filter(chan, filter)
EChannel chan;
char *filter;
{
    int i = 0;

    while ((i < MAX_FILTERS) && (strcmp(filters[i].module, filter) != 0))
      i++;
    if (i < MAX_FILTERS) {
	chan->out_filter = filters[i].outfunc;
	chan->in_filter = filters[i].infunc;
    }

    return ECPsource_subscribe(chan, (ECRequestHandlerFunction)NULL, NULL);
}
#endif

extern ECSourceHandle
ECPsource_subscribe(chan, func, client_data)
EChannel chan;
ECRequestHandlerFunction func;
void *client_data;
{
    int i;
  
    ECSourceHandle handle = new_source_handle();

    if (chan->type_format != NULL) {
	fprintf(stderr, 
		"ECsource_subscribe() failed.  No untyped sources on typed channels\n");
	return NULL;
    }

    ECtrace_out("ECsource_subscribe, chan %lx", 1, chan);
    handle->chan = chan;
    EChannel_lock(chan);
    chan->local_source_list =
	CMrealloc(chan->local_source_list,
		  sizeof(struct _local_source) * (chan->local_source_list_len + 1));
    chan->local_source_list[chan->local_source_list_len].func = func;
    chan->local_source_list[chan->local_source_list_len].client_data = client_data;
    chan->local_source_list[chan->local_source_list_len].source_handle = handle;
    handle->source_num = chan->local_source_list_len;

    (chan->local_source_list_len)++;

    ECtrace_out("working on remote members", 3);
    for (i = 0; i < chan->rem_member_list_len; i++) {
	if (chan->rem_member_list[i].conn != NULL) {
	    EChannel_unlock(chan);
	    send_source_subscribe(chan, i, 1); /* true means "subscribe" */
	    EChannel_lock(chan);
	}
    }

    EChannel_unlock(chan);

    ECtrace_out("Member counts: \n Local sources %d Local sinks %d \n Remote Sources %d Remote Sinks %d Remote Members %d", 3, 
		chan->local_source_list_len,
		chan->local_sink_list_len,
		chan->rem_source_list_len, 
		chan->rem_sink_list_len, 
		chan->rem_member_list_len);

    ECtrace_out("ECsource_subscribe, chan %lx, returning handle %lx", 0, chan,
		handle);
    return handle;
}

extern void
ECcancel_source_subscribe(handle)
ECSourceHandle handle;
{
    EChannel chan = handle->chan;
    EChannel_lock(chan);
    internal_cancel_source_subscribe(handle);
    EChannel_unlock(chan);
}

static void
internal_cancel_source_subscribe(handle)
ECSourceHandle handle;
{
    int i;  
    EChannel chan = handle->chan;

    if (handle->ioformat != NULL) 
      CMfree_user_type_context (chan->cm, (IOContext)iofile_of_IOformat(handle->ioformat));
    
    /* remove source from channel's list */
    for (i=0; i< handle->chan->local_source_list_len; i++) {
	if (handle->chan->local_source_list[i].source_handle == handle) {
	    for ( ; i < handle->chan->local_source_list_len -1 ; i++) {
		handle->chan->local_source_list[i] =
		    handle->chan->local_source_list[i+1];
	    }
	    handle->chan->local_source_list_len--;
	    break;
	}
    }

    /* 
     *  only if it is the last one in this side it will send a msg 
     *  we need this for remote sources now 
     *
     */
  
    if (chan->local_source_list_len == 1) {
	int i;
	ECtrace_out("working on remote members", 3);
	for (i = 0; i < chan->rem_member_list_len; i++) {
	    if (chan->rem_member_list[i].conn != NULL) {
		EChannel_unlock(chan);
		send_source_subscribe(chan, i, 0); /* false means unsubscribe*/
		EChannel_lock(chan);
	    }
	}
    }
#ifndef NO_DERIVED
    /* deal with derived sources */
    for (i=0; i< handle->der_list_len; i++) {
	ecl_code_free((void*)(long)handle->der_source_list[i].code_struct);
	ECcancel_source_subscribe(handle->der_source_list[i].source_handle);
    }
    
#endif
    CMfree(handle);
}

/*
 *   Extended with (func, client_data) to support a pull model. The idea is 
 *   that the handler would be executed when the source get a request and 
 *   it would submit an event as part of its execution - fabianb
 */ 


#ifdef MODULE
extern ECSourceHandle 
ECsource_typed_subscribe_filter (chan, field_list, format_list, filter)
EChannel chan;
IOFieldList field_list;
CMFormatList format_list;
char *filter;
{
    int i = 0;

    while ((i < MAX_FILTERS) && (strcmp(filters[i].module, filter) != 0))
      i++;
    if (i < MAX_FILTERS) {
	chan->out_filter = filters[i].outfunc;
	chan->in_filter = filters[i].infunc;
    }

    return ECPsource_typed_subscribe(chan, field_list, format_list, 
				     (ECRequestHandlerFunction) NULL, 
				     NULL);
}
#endif

extern ECSourceHandle 
ECsource_typed_subscribe (chan, field_list, format_list)
EChannel chan;
IOFieldList field_list;
CMFormatList format_list;
{
    return ECPsource_typed_subscribe(chan, field_list, format_list, 
				     (ECRequestHandlerFunction) NULL, 
				     NULL);
}


extern ECSourceHandle
ECPsource_typed_subscribe(chan, field_list, format_list, func, client_data)
EChannel chan;
IOFieldList field_list;
CMFormatList format_list;
ECRequestHandlerFunction func;
void *client_data;
{
    ECSourceHandle handle = new_source_handle();
    IOContext	source_context = CMget_user_type_context(chan->cm);
    char format_name[] = "anonymous";
    int i = 0;

    if (chan->type_format == NULL) {
	fprintf(stderr, "ECsource_subscribe() failed.  No typed sources on untyped channels");
	return NULL;
    }
    ECtrace_out("ECsource_subscribe, chan %lx", 1, chan);
    handle->chan = chan;
    EChannel_lock(chan);

    /* assert that the formats are compatible -GSE- */
    
    handle->ioformat = CMregister_user_format(chan->cm, source_context, 
					      format_name, field_list, 
					      format_list);

    /* add to our channels list of sources */
    chan->local_source_list =
	CMrealloc(chan->local_source_list,
		  sizeof(struct _local_source) * (chan->local_source_list_len + 1));
    chan->local_source_list[chan->local_source_list_len].func = func;
    chan->local_source_list[chan->local_source_list_len].client_data = client_data;
    chan->local_source_list[chan->local_source_list_len].source_handle = handle;
    handle->source_num = chan->local_source_list_len;
    (chan->local_source_list_len)++;

    ECtrace_out("working on remote members", 3);
    for (i = 0; i < chan->rem_member_list_len; i++) {
	if (chan->rem_member_list[i].conn != NULL) {
	    EChannel_unlock(chan);
	    send_source_subscribe(chan, i, 1); /* true means "subscribe" */
	    EChannel_lock(chan);
	}
    }

    /* for each channel derived for ours, derive a source */
    for (i=0; i< chan->der_list_len; i++) {
	EChannel der_chan = chan->der_chan_list[i].derived_chan;
	IOFieldList output_field_list = NULL;
	CMFormatList output_format_list = NULL;
	int j, ret;
	
	if (chan->der_chan_list[i].has_output_param) {
	    output_field_list = EChannel_get_field_list(der_chan);
	    output_format_list = EChannel_get_format_list(der_chan);
	}
	ret = do_internal_source_derive(handle, der_chan,
					chan->der_chan_list[i].filter, 
					field_list, format_list,
					output_field_list, output_format_list,
					der_chan->data_field_list, 
					der_chan->data_format_list);
	if (ret != 1) {
	    EChannel_destroy(der_chan);
	    return NULL;
	}
	j=0;
	while (output_format_list && 
	       (output_format_list[j].format_name != NULL)) { 
	    free_field_list(output_format_list[j].field_list);
	    free(output_format_list[j].format_name);
	    j++;
	}
	if (output_format_list)	free(output_format_list);
	if (output_field_list) free_field_list(output_field_list);
    }

    EChannel_unlock(chan);

    ECtrace_out("Member counts: \n Local sources %d Local sinks %d \n Remote Sources %d Remote Sinks %d Remote Members %d", 3,
		chan->local_source_list_len,
		chan->local_sink_list_len, 
		chan->rem_source_list_len,  
		chan->rem_sink_list_len,
		chan->rem_member_list_len);

    ECtrace_out("ECsource_subscribe, chan %lx, returning handle %lx", 0, chan,
		handle);

    return handle;
}

extern char *
ECglobal_id(chan)
EChannel chan;
{
    int len = 0;
    char chan_str[17] =
    {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    unsigned char chan_char[8] =
    {0, 0, 0, 0, 0, 0, 0, 0};
    char *global_id_str;

    memset(chan_char, 0, sizeof(chan_char));
    memcpy(chan_char, &chan, sizeof(chan));
    sprintf(chan_str, "%02x%02x%02x%02x%02x%02x%02x%02x",
	    (unsigned int) chan_char[0],
	    (unsigned int) chan_char[1],
	    (unsigned int) chan_char[2],
	    (unsigned int) chan_char[3],
	    (unsigned int) chan_char[4],
	    (unsigned int) chan_char[5],
	    (unsigned int) chan_char[6],
	    (unsigned int) chan_char[7]);
    len = strlen(chan->global_id.contact_string);
    len += 16 /* chan_str */  + 1 /* separators */ ;
    global_id_str = CMmalloc(len + 1);
    memset(global_id_str, 0, len + 1);
    sprintf(global_id_str, "%s@%s", chan->global_id.contact_string,
	    chan_str);
    return global_id_str;
}

extern void
CMsetup_conn_or_local(ec, chan_id, conn_ptr, local_ptr)
EControlContext ec;
char *chan_id;
CMConnection *conn_ptr;
void **local_ptr;
{
    char *chan_id_copy = CMmalloc(strlen(chan_id) + 1);
    char *contact_str;
    char *chan_str;
    attr_list attrs;
    char *tmp;
    
    *conn_ptr = NULL;
    *local_ptr = NULL;
    strcpy(chan_id_copy, chan_id);
    contact_str = chan_id_copy;
    tmp = strchr(chan_id_copy, '@');
    if (tmp == NULL) {
	ECtrace_out("ECho, bad id %s", 2, chan_id);
	return;
    }
    *tmp = 0;			/* kill first @ */
    chan_str = ++tmp;

    attrs = attr_list_from_string(contact_str);
    if (CMcontact_self_check(ec->cm, attrs)) {
	free_attr_list(attrs);
	/* this channel is really local */
	*local_ptr = string_to_channel_ptr(chan_str);
	return;
    }
    ECtrace_out("getting a connection to %s, registering close handler", 3, 
	      contact_str);
    if (attrs == NULL) {
	printf("malformed contactlist\n");
	return;
    }
    *conn_ptr = CMget_conn(ec->cm, attrs);
    free_attr_list(attrs);
    if (*conn_ptr == NULL) {
	return;
    }
    CMconn_register_close_handler(*conn_ptr, EChannel_close_handler, ec);
}

#ifdef MODULE
extern EChannel
EChannel_open_id(ec, chan_id, conn_id)
EControlContext ec;
char *chan_id;
char *conn_id;
{
    EChannel chan;

    chan = EChannel_open(ec, chan_id);
    if ((chan != NULL) && (conn_id != NULL))
	chan->conn_id = strdup(conn_id);
    else
	chan->conn_id = NULL;
    return chan;
}
#endif

extern EChannel
EChannel_open(ec, chan_id)
EControlContext ec;
char *chan_id;
{
    EChannel chan;
    CMConnection conn;
    AttendMsg attend_msg;
    attr_list self_contact_list;
    int cond;

    ECtrace_out("ECchannel_open, ID=%s", 1, chan_id);
    CMsetup_conn_or_local(ec, chan_id, &conn, (void**)&chan);
    if (conn == NULL) {
	ECtrace_out("EChannel_open, returning %lx", 0, chan);
	return chan;
    }
    chan = EChannel_create(ec);
    cond = CMCondition_get(ec->cm, conn);
    CMCondition_set_client_data(ec->cm, cond, &chan);
    memset(&attend_msg, 0, sizeof(attend_msg));
    attend_msg.chan_str = strchr(chan_id, '@') + 1;
    attend_msg.cond = cond;
    set_channel_id(&attend_msg.client_channel_id, chan);
    self_contact_list = CMget_contact_list(ec->cm);
    if (self_contact_list == NULL) {
	printf("Warning, no contact list.  Must do CManager_listen() before using channels.\n");
	CMfree(chan);
	return NULL;
    }
    attend_msg.contact_str = attr_list_to_string(self_contact_list);
    ECtrace_out("creating attend_msg and writing to connection", 3);
    if (!CMwrite(conn, CMlookup_format(ec->cm, channel_attend_msg_flds),
		 &attend_msg)) {
	EChannel_destroy(chan);
	ECtrace_out("ECchannel_open, write failed, channel stillborn", 0);
	return NULL;
    }
    free(attend_msg.contact_str);
    if (channel_verbose > 0) {
	printf("Sending attend msg, channel %s cond %d\n",
	       attend_msg.chan_str, attend_msg.cond);
    }
    if (CMCondition_wait(ec->cm, cond)) {
	ECtrace_out("ECchannel_open channel is ready %lx", 0, chan);
	if (channel_verbose) {
	    printf("\tOpened channel ID is ");
	    print_channel_id(chan->global_id.global_channel_id);
	    printf("\n");
	}
#ifdef MODULE
	add_to_channel_pool(chan);	//Sandip
#endif
	return chan;
    } else {
	/* wait failed, conn is dead */
	EChannel_destroy(chan);
	ECtrace_out("ECchannel_open, wait failed, channel stillborn", 0);
	return NULL;
    }
}

extern void
ECcancel_sink_subscribe ARGS((ECSinkHandle handle));

static void
send_request_event(chan, i)
EChannel chan;
int i;
{
    RequestEventMsg msg;

    ECtrace_out("send_request_event", 1);
    ECtrace_out("create msg & write to rem_source_list port", 3);

    memset(&msg, 0, sizeof(msg));
    msg.global_channel_id = chan->global_id.global_channel_id;

    CMwrite(chan->rem_source_list[i].conn,
		CMlookup_format(chan->cm, request_event_msg_flds), &msg);
    ECtrace_out("send_request_event", 0);
}

#ifdef MODULE
static void
send_remote_event_attr(chan, i, event, event_len, attrs)
EChannel chan;
int i;
void *event;
int event_len;
attr_list attrs;
{
    EventMsg msg;
    memset(&msg, 0, sizeof(msg));
    msg.event = event;
    msg.event_len = event_len;

    msg.global_channel_id = chan->global_id.global_channel_id;

    ECtrace_out("send_remote_event, chan %lx, event %lx", 1, chan, event);
    ECtrace_out("create event msg and write to remote member list port", 3);
    if (!CMwrite_attr(chan->rem_sink_list[i].conn, 
		 CMlookup_format(chan->cm, event_msg_flds), &msg, attrs)) {
	ECtrace_out("Write failed", 0);
    }
    ECtrace_out("send_remote_event", 0);
}

static void
sendv_remote_event_attr(chan, i, eventV, attrs)
EChannel chan;
int i;
IOEncodeVector eventV;
attr_list attrs;
{
    int vec_count = 0;
    EventVMsg msg;
    memset(&msg, 0, sizeof(msg));
    while (eventV[vec_count].iov_base != NULL) {
	vec_count++;
    }
    ECtrace_out("sendv_remote_event, chan %lx, eventv %lx", 1, chan, eventV);
    ECtrace_out("create event msg and write to remote member list port", 3);
    msg.eventv = eventV;
    msg.eventv_len = vec_count;

    msg.global_channel_id = chan->global_id.global_channel_id;

    if (!CMwrite_attr(chan->rem_sink_list[i].conn, 
		 chan->eventv_format, &msg, attrs)) {
	ECtrace_out("send remote event write failed", 0);
    }
    ECtrace_out("sendv_remote_event", 0);
}
#else
static void
send_remote_event(chan, i, event, event_len)
EChannel chan;
int i;
void *event;
int event_len;
{
    EventMsg msg;
    memset(&msg, 0, sizeof(msg));
    msg.event = event;
    msg.event_len = event_len;

    msg.global_channel_id = chan->global_id.global_channel_id;

    ECtrace_out("send_remote_event, chan %lx, event %lx", 1, chan, event);
    ECtrace_out("create event msg and write to remote member list port", 3);
    if (!CMwrite(chan->rem_sink_list[i].conn, 
		 CMlookup_format(chan->cm, event_msg_flds), &msg)) {
	ECtrace_out("Write failed", 0);
    }
    ECtrace_out("send_remote_event", 0);
}

static void
sendv_remote_event(chan, i, eventV)
EChannel chan;
int i;
IOEncodeVector eventV;
{
    int vec_count = 0;
    EventVMsg msg;
    memset(&msg, 0, sizeof(msg));
    while (eventV[vec_count].iov_base != NULL) {
	vec_count++;
    }
    ECtrace_out("sendv_remote_event, chan %lx, eventv %lx", 1, chan, eventV);
    ECtrace_out("create event msg and write to remote member list port", 3);
    msg.eventv = eventV;
    msg.eventv_len = vec_count;

    msg.global_channel_id = chan->global_id.global_channel_id;

    if (!CMwrite(chan->rem_sink_list[i].conn, 
		 chan->eventv_format, &msg)) {
	ECtrace_out("send remote event write failed", 0);
    }
    ECtrace_out("sendv_remote_event", 0);
}
#endif

static void
send_sink_subscribe(chan, i, subscribe)
EChannel chan;
int i;
int subscribe;
{
    SinkSubscribeMsg msg;

    memset(&msg, 0, sizeof(msg));
    msg.global_channel_id = chan->global_id.global_channel_id;
#ifdef MODULE
    msg.conn_id = chan->conn_id;
#endif
    set_channel_id(&msg.client_channel_id, chan);
    ECtrace_out("send_sink_subscription", 1);
    if (channel_verbose) {
	printf("\tsending to channel with global id ");
	print_channel_id(msg.global_channel_id);
	printf("\n");
    }
    ECtrace_out("create msg and write to rem_member_list port", 3);
    if (subscribe) {
	msg.client_contact_str = chan->local_contact_string;
	if (!CMwrite(chan->rem_member_list[i].conn, 
		     CMlookup_format(chan->cm, sink_subscribe_msg_flds),
		     &msg)) {
	    ECtrace_out("send sink subscribe Write failed", 0);
	}
    } else {
	msg.client_contact_str = NULL;
	if (!CMwrite(chan->rem_member_list[i].conn, 
		     CMlookup_format(chan->cm, sink_unsubscribe_msg_flds),
		     &msg)) {
	    ECtrace_out("send sink unsubscribe write failed", 0);
	}
    }
    ECtrace_out("send_sink_subscription, subscribe = %d", 0, subscribe);
}

static void
send_source_subscribe(chan, i, subscribe)
EChannel chan;
int i;
int subscribe;
{
    SourceSubscribeMsg msg;

    memset(&msg, 0, sizeof(msg));
    msg.global_channel_id = chan->global_id.global_channel_id;
    set_channel_id(&msg.client_channel_id, chan);
    ECtrace_out("send_source_subscription", 1);
    ECtrace_out("create msg and write to rem_member_list port", 3);
    if (subscribe) {
	msg.client_contact_str = chan->local_contact_string;
	if (!CMwrite(chan->rem_member_list[i].conn, 
		     CMlookup_format(chan->cm, source_subscribe_msg_flds), 
		     &msg)) {
	    ECtrace_out("send source subscribe write failed", 0);
	}
    } else {
	msg.client_contact_str = NULL;
	if (!CMwrite(chan->rem_member_list[i].conn, 
		     CMlookup_format(chan->cm, source_unsubscribe_msg_flds),
		     &msg)) {
	    ECtrace_out("send source unsubscribe write failed", 0);
	}
    }
    ECtrace_out("send_source_subscription, subscribe = %d", 0, subscribe);
}

static void
send_member_subscribe(chan, i)
EChannel chan;
int i;
{
    MemberSubscribeMsg msg;

    ECtrace_out("send_member_subscription", 1);
    ECtrace_out("create msg & write to rem_member_list port", 3);
    memset(&msg, 0, sizeof(msg));
    msg.global_channel_id = chan->global_id.global_channel_id;
    msg.client_contact_str = chan->local_contact_string;
    set_channel_id(&msg.client_channel_id, chan);
    if (!CMwrite(chan->rem_member_list[i].conn,
		 CMlookup_format(chan->cm, member_subscribe_msg_flds), &msg)) {
	ECtrace_out("send member subscribe write failed\n", 0);
    }
    ECtrace_out("send_member_subscription", 0);
}


static
event_queue
get_event_item(EControlContext ec)
{
    if (ec->free_event_item_list == NULL) {
	return (event_queue)malloc(sizeof(event_item));
    } else {
	event_queue tmp = ec->free_event_item_list;
	ec->free_event_item_list = tmp->next;
	return tmp;
    }
}

static void
release_event_queue_item(EControlContext ec, event_queue item)
{
    if (item->attrs != NULL) {
	free_attr_list(item->attrs);
	item->attrs = NULL;
    }
    item->next = ec->free_event_item_list;
    ec->free_event_item_list = item;
}

#ifdef MODULE
int edf_order(ECRealTimeAttrs rt_attrs1, ECRealTimeAttrs rt_attrs2) {
    if (rt_attrs1->priority != -1) {
	if (rt_attrs1->priority <= rt_attrs2->priority)
		return 1;
	else 
		return 0;
    }
    if (rt_attrs1->deadline.tv_sec > rt_attrs2->deadline.tv_sec)
        return 1;
    else if ((rt_attrs1->deadline.tv_sec == rt_attrs2->deadline.tv_sec) && 
	(rt_attrs1->deadline.tv_usec >= rt_attrs2->deadline.tv_usec)) 
	return 1;
    else 
	return 0;
}
#endif

static void
queue_event_for_dispatch(chan, sink_index, event, event_len, block_rec,
			 event_attrs)
EChannel chan;
int sink_index;
void *event;
int event_len;
free_block_rec_p block_rec;
attr_list event_attrs;
{
    event_item *item;
    EControlContext ec = chan->ec;
    IOContext context = chan->local_sink_list[sink_index].context;
    CMFormatList format_list = chan->local_sink_list[sink_index].format_list;
    IOFieldList field_list = chan->local_sink_list[sink_index].field_list;
    int native_struct_size;
#ifdef MODULE
    event_item *hptr;
#endif
//struct timeval now;

    if ((context != NULL) && (event_len != 0)) {
	IOFormat ioformat;
	void *decode_buffer;
	/* decode typed event */
	ioformat = CMget_format_IOcontext(chan->cm, context, event);
	if (ioformat == NULL) {
	    fprintf(stderr, "Unknown IOFormat in encoded block\n");
	}
	ECtrace_out("Queue event got typed event, format \"%s\"", 
		  3, name_of_IOformat(ioformat));
	if (!has_conversion_IOformat(ioformat)) {
	    IOFormat *ioformats;
	    IOFormat *format;
	    if ((ioformats = CMget_subformats_IOcontext(chan->cm, context,
							event)) == NULL) {
		fprintf(stderr, "IOFormat list is null\n");
		return;
	    }
	    format = ioformats;
	    while (format_list && (*format != NULL)) {
		int i = 0;
		if (!has_conversion_IOformat(*format)) {
		    while(format_list[i].format_name != NULL) {
			if (strcmp(format_list[i].format_name,
				   name_of_IOformat(*format)) == 0) {
			    break;
			}
			i++;
		    }
		    if (format_list[i].field_list != NULL) {
			native_struct_size = 
			    struct_size_IOfield((IOFile)context, 
						format_list[i].field_list);
			CMset_conversion_IOcontext(chan->cm, context, 
						   *format,
						   format_list[i].field_list,
						   native_struct_size);
		    }
		}
		format++;
	    }
	    free(ioformats);
	    native_struct_size = struct_size_IOfield((IOFile)context, 
						     field_list);
	    CMset_conversion_IOcontext(chan->cm, context, ioformat,
				     field_list, native_struct_size);
	}
	if (!decode_in_place_possible(ioformat)) {
	    decode_buffer = CMmalloc(this_IOrecord_length(context, event,event_len));
	    decode_to_buffer_IOcontext(context, event, decode_buffer);
	    block_rec = get_free_block_func(chan->ec, decode_buffer, 
					    (EventFreeFunction)NULL);
	    block_rec->ref_count = 0;
	    event = decode_buffer;
	    event_len = 0;
	} else {
	    void *decoded_event;
	    decode_in_place_IOcontext(context, event, &decoded_event);
	    event = decoded_event;
	}
    }
	
    item = get_event_item(chan->ec);
    if (chan->local_sink_list[sink_index].multi_sink_handle != NULL) {
	multi_sink_info msink_handle = 
	    chan->local_sink_list[sink_index].multi_sink_handle;
	struct _ECEventQueue *eq;
	int i = 0;
	while (msink_handle->queue_list->event_queues[i].sink_handle->chan
	       != chan) {
	    i++;
	    if (i >= msink_handle->queue_list->queue_count) {
		printf("Channel not found in  queue \n");
		return;
	    }
	}
	eq = &msink_handle->queue_list->event_queues[i];
	if ((eq->event_count+1) > eq->queue_size) {
	    eq->queue_size = eq->event_count+1;
	    eq->queue = realloc(eq->queue,
				sizeof(eq->queue[0])*eq->queue_size);
	}
	eq->queue[eq->event_count].event = event;
	if (block_rec != NULL) {
	    block_rec->ref_count++;
	} else {
	    static int block_copy_warning_done = 0;
	    if (!block_copy_warning_done) {
		fprintf(stderr, "Warning:  An event submitted with ECsubmit_event(), ECsubmit_typed_event() or\n    ECsubmit_eventV() has been passed to a multi-sink handler.\n    This violates ECho event data memory handling requirements.  See\n   http://www.cc.gatech.edu/systems/projects/ECho/event_memory.html\n");
		block_copy_warning_done++;
	    }
	}
	eq->queue[eq->event_count].free_block = block_rec;
	ECtrace_out("Added event %d to queue %d\n", eq->event_count, i);
	eq->event_count++;
	event = NULL;
	event_len = 0;
	block_rec = NULL;
    } else {
	if ((block_rec == NULL) && (chan->ec != ec)) {
	    static int block_copy_warning_done = 0;
	    if (!block_copy_warning_done) {
		fprintf(stderr, "Warning:  An event submitted with ECsubmit_event(), ECsubmit_typed_event() or\n   ECsubmit_eventV() has been passed to handler in another EControlContext.\n   This violates ECho event data memory handling requirements.  See \nhttp://www.cc.gatech.edu/systems/projects/ECho/event_memory.html\n");
		block_copy_warning_done++;
	    }
	}
    }

    item->event = event;
    item->event_len = event_len;
#ifdef MODULE
    item->rt_attrs = NULL;
    if (event_len >= sizeof(struct _ECRealTimeAttrs)) {
	item->rt_attrs = (ECRealTimeAttrs) (event + 
		(event_len - sizeof(struct _ECRealTimeAttrs)));
	if ((item->rt_attrs->rt_header[0] == 0x52544500) || 
		(item->rt_attrs->rt_header[0] == 0x00455452)) {
	    event_len = event_len - sizeof(struct _ECRealTimeAttrs);	  
	} else {
	  item->rt_attrs = NULL;
	}
    }
#endif
    item->eventv = NULL;
    item->sink_index = sink_index;
    add_ref_attr_list(event_attrs);
    item->attrs = event_attrs;
    item->chan = chan;
    if (block_rec != NULL)
	block_rec->ref_count++;
    item->block_rec = block_rec;
    ECtrace_out("Queue event for local dispatch, chan %lx, event %lx", 
		3, chan, event);

#ifdef MODULE
    if (item->rt_attrs == NULL) {
        item->next = NULL;
        item->prev = ec->event_queue_tail;
        if (item->prev) {
	    item->prev->next = item;
        }
        ec->event_queue_tail = item;
        if (ec->event_queue_head == NULL) {
	    ec->event_queue_head = item;
        }
    } else {
        if (ec->dispatch_func == NULL) { /* dflt dispatching algorithm (EDF) */
            hptr = ec->event_queue_head; 
  	    while (hptr != NULL) {
		if ((hptr->rt_attrs == NULL) || 
		  (edf_order(hptr->rt_attrs, item->rt_attrs) == 0)) {
    		    item->next = hptr;
    		    item->prev = hptr->prev;
    		    if (item->prev == NULL) 
            		ec->event_queue_head = item; 
		    else
			item->prev->next = item;
		    hptr->prev = item;
		    break;
		}
		hptr = hptr->next;
	    }
	    if (hptr == NULL) {
    	        item->next = hptr;
    		item->prev = ec->event_queue_tail;
    		if (item->prev == NULL) 
            	    ec->event_queue_head = item; 
		else
		    item->prev->next = item;
    		ec->event_queue_tail = item;
	    }
	} else { /* customized dispatching algorithm */
//	    DispatchFunc func = (DispatchFunc) ec->dispatch_func(item,ec);
	    //(func)(item, ec);
	}
    }
#else 
    item->next = NULL;
    item->prev = ec->event_queue_tail;
    if (item->prev) {
	item->prev->next = item;
    }
    ec->event_queue_tail = item;
    if (ec->event_queue_head == NULL) {
	ec->event_queue_head = item;
    }
#endif
    if (chan->ec != ec) {
	/* if we're putting this in some other DE's queue, wake any server */
/*	wake_CM_server(ec->cm);*/
	printf("Should be waking server\n");
    }
}


static void
queue_eventv_for_dispatch(chan, sink_index, eventv, block_rec, event_attrs)
EChannel chan;
int sink_index;
IOEncodeVector eventv;
free_block_rec_p block_rec;
attr_list event_attrs;
{
    event_item *item = get_event_item(chan->ec);
    item->event = NULL;
    item->event_len = 0;
    item->eventv = eventv;
    item->sink_index = sink_index;
    item->chan = chan;
    add_ref_attr_list(event_attrs);
    item->attrs = event_attrs;
    block_rec->ref_count++;
    item->block_rec = block_rec;

    item->next = NULL;
    item->prev = chan->ec->event_queue_tail;
    chan->ec->event_queue_tail = item;
    if (chan->ec->event_queue_head == NULL) {
	chan->ec->event_queue_head = item;
    }
    if (channel_verbose > 0)
	printf("queuev for dispatch\n");
}

static event_queue
dequeue_event(ec)
EControlContext ec;
{
    event_queue item;

    if (ec->event_queue_head == NULL) {
	return NULL;
    } else {
	item = ec->event_queue_head;
	ec->event_queue_head = item->next;
    }
    if (item == ec->event_queue_tail) {
	ec->event_queue_tail = NULL;
    }
    item->prev = NULL;
    item->next = NULL;
    return item;
}

#ifdef MODULE
static void
dispatch_event2(ec, func, event, event_len, client_data, event_attrs)
EControlContext ec;
ECHandlerFunction func;
void *event;
int event_len;
void *client_data;
attr_list event_attrs;
{
    if (current->kthread != 1) {
	CMdeliver_event(ec->cm, (void *)func, event, event_len, client_data,
		event_attrs);
   	return;
    }
    func(event, event_len, client_data, event_attrs);
}

static void
dispatch_typed_event2(ec, func, event, event_len, client_data, event_attrs)
EControlContext ec;
ECHandlerFunction func;
void *event;
int event_len;
void *client_data;
attr_list event_attrs;
{
    if (current->kthread != 1) {
	CMdeliver_event(ec->cm, (void *)func, event, event_len, client_data,
		event_attrs);
   	return;
    }
    ((ECTypedHandlerFunction)func)(event, client_data, event_attrs);
}

static void
dispatch_multi_sink_event2(ec, func, queue_list, len, client_data)
EControlContext ec;
ECMultiHandlerFunction func;
ECEventQueueList queue_list;
int len;
void *client_data;
{
    if (current->kthread != 1) {
	CMdeliver_event(ec->cm, (void *)func, queue_list, len, client_data,
		NULL);
   	return;
    }
    ((ECMultiHandlerFunction)func)(queue_list, client_data);
}
#else

static void
dispatch_event(func, event, event_len, client_data, event_attrs)
ECHandlerFunction func;
void *event;
int event_len;
void *client_data;
attr_list event_attrs;
{
    func(event, event_len, client_data, event_attrs);
}

static void
dispatch_typed_event(func, event, client_data, event_attrs)
ECHandlerFunction func;
void *event;
void *client_data;
attr_list event_attrs;
{
    ((ECTypedHandlerFunction)func)(event, client_data, event_attrs);
}

static void
dispatch_multi_sink_event(func, queue_list, client_data)
ECMultiHandlerFunction func;
ECEventQueueList queue_list;
void *client_data;
{
    ((ECMultiHandlerFunction)func)(queue_list, client_data);
}
#endif

static void
do_event_dispatch(ec)
EControlContext ec;
{
    event_queue item;
#ifdef MODULE
    int mqueue_len;
#endif
    EContext_lock(ec);
    while ((item = dequeue_event(ec)) != NULL) {
	EChannel chan = item->chan;
	ECtrace_out("dequeued event", 3);
	ec->current_event_item = item;
	EContext_unlock(ec);
	if (chan->local_sink_list[item->sink_index].func !=
	    (ECHandlerFunction) NULL) {
	    int sink = item->sink_index;
	    if (item->eventv != NULL) {
		int i = 0;
		char *tmp;
		char *event;
		int event_len = 0;

		IOEncodeVector eventV = item->eventv;
		while (eventV[i].iov_base != NULL) {
		    event_len += eventV[i++].iov_len;
		}
		tmp = event = CMmalloc(event_len);
		i = 0;
		while (eventV[i].iov_base != NULL) {
		    memcpy(tmp, eventV[i].iov_base, eventV[i].iov_len);
		    tmp += eventV[i++].iov_len;
		}
#ifdef MODULE
		dispatch_event2(ec, chan->local_sink_list[sink].func,
			       event, event_len,
			       chan->local_sink_list[sink].client_data,
			       item->attrs);
#else
		dispatch_event(chan->local_sink_list[sink].func,
			       event, event_len,
			       chan->local_sink_list[sink].client_data,
			       item->attrs);
#endif

		EChannel_lock(chan);
		if (chan->local_sink_list[sink].cond) {
		    CMCondition_signal(chan->cm,
				       chan->local_sink_list[sink].cond);
		}
		chan->local_sink_list[sink].cond = 0;
		EChannel_unlock(chan);
		CMfree(event);
	    } else {
		if (chan->local_sink_list[item->sink_index].context == NULL) {
		    /* untyped event */
#ifdef MODULE
		    dispatch_event2(ec, chan->local_sink_list[sink].func,
				   item->event, item->event_len,
				   chan->local_sink_list[sink].client_data,
				   item->attrs);
#else
		    dispatch_event(chan->local_sink_list[sink].func,
				   item->event, item->event_len,
				   chan->local_sink_list[sink].client_data,
				   item->attrs);
#endif
		} else {
		    /* typed event */
#ifdef MODULE
		    dispatch_typed_event2(ec, chan->local_sink_list[sink].func,
					 item->event, item->event_len,
					 chan->local_sink_list[sink].client_data,
					 item->attrs);
#else
		    dispatch_typed_event(chan->local_sink_list[sink].func,
					 item->event, 
					 chan->local_sink_list[sink].client_data,
					 item->attrs);
#endif
		}

		EChannel_lock(chan);
		if (chan->local_sink_list[sink].cond) {
		    CMCondition_signal(chan->cm, 
				       chan->local_sink_list[sink].cond);
		}
		chan->local_sink_list[sink].cond = 0;
		EChannel_unlock(chan);
	    }
	} else if (chan->local_sink_list[item->sink_index].multi_sink_handle) {
	    multi_sink_info msink_info = 
		chan->local_sink_list[item->sink_index].multi_sink_handle;
	    ec->current_event_item = NULL;
#ifdef MODULE
	    mqueue_len = msink_info->queue_list->event_queues->queue_size;
	    dispatch_multi_sink_event2(ec, msink_info->func,
				      msink_info->queue_list,
				      mqueue_len,
				      msink_info->client_data);
#else
	    dispatch_multi_sink_event(msink_info->func,
				      msink_info->queue_list,
				      msink_info->client_data);
#endif
	}
	if (item->block_rec) {
	    deref_block(item->block_rec);
	}
	EContext_lock(ec);
	ec->current_event_item = NULL;
	release_event_queue_item(ec, item);
#ifdef MODULE
	if (current->kthread != 1) break;
#endif
    }
    EContext_unlock(ec);
}

int
ECtake_event_buffer(ec, event)
EControlContext ec;
void *event;
{
    event_queue item;

    if (ec->current_event_item == NULL) {
	fprintf(stderr, "No event handler with takeable buffer executing on this EControlContext.\n");
	return 0;
    }
    if (ec->current_event_item->event != event) {
	fprintf(stderr, "Event address (%lx) in ECtake_event_buffer does not match currently executing event on this EControlContext.\n", (long)event);
	return 0;
    }
    if (ec->current_event_item->block_rec == NULL) {
	static int take_event_warning = 0;
	if (take_event_warning == 0) {
	    fprintf(stderr, "Warning:  ECtake_event_buffer called on an event submitted with \n    ECsubmit_event(), ECsubmit_typed_event() or ECsubmit_eventV() .\n    This violates ECho event data memory handling requirements.  See \n    http://www.cc.gatech.edu/systems/projects/ECho/event_memory.html\n");
	    take_event_warning++;
	}
	return 0;
    }
    EContext_lock(ec);
    item = get_event_item(ec);
    item->event = ec->current_event_item->event;
    item->event_len = ec->current_event_item->event_len;
    item->eventv = ec->current_event_item->eventv;
    item->sink_index = ec->current_event_item->sink_index;
    item->chan = ec->current_event_item->chan;
    item->attrs = NULL;
    ec->current_event_item->block_rec->ref_count++;
    item->block_rec = ec->current_event_item->block_rec;

    item->next = ec->taken_events_list;
    item->prev = NULL;

    ec->taken_events_list = item;
    EContext_unlock(ec);
    return 1;
}

void
ECreturn_event_buffer(ec, event)
EControlContext ec;
void *event;
{
    event_queue tmp, last = NULL;
    /* search through list for event and then dereference it */
    EContext_lock(ec);
    tmp = ec->taken_events_list;
    while(tmp != NULL) {
	if (tmp->event == event) {
	    if (last == NULL) {
		ec->taken_events_list = tmp->next;
	    } else {
		last->next = tmp->next;
	    }
	    release_event_queue_item(ec, tmp);
	    EContext_unlock(ec);
	    deref_block(tmp->block_rec);
	    return;
	}
	last = tmp;
	tmp = tmp->next;
    }
    fprintf(stderr, "Event %lx not found in taken events list\n", 
	    (long)event);
    EContext_unlock(ec);
}

static void
do_event_flush(ec)
EControlContext ec;
{
    event_queue item;
    EContext_lock(ec);
    while ((item = dequeue_event(ec)) != NULL) {
	if (channel_verbose > 0)
	    printf("flushing event\n");
	if (item->block_rec) {
	    deref_block(item->block_rec);
	}
	release_event_queue_item(ec, item);
    }
    EContext_unlock(ec);
}

extern void
EChannel_subscribe_handler(chan, func, client_data)
EChannel chan;
ECSubscribeFunction func;
void *client_data;
{
    ECtrace_out("EChannel_subscribe_handler", 1);
    chan->subscribe_func = func;
    chan->subscribe_client_data = client_data;
    ECtrace_out("EChannel_subscribe_handler", 0);
}

extern void
DEChannel_RequestEvent_handler(cm, conn, data, client_data)
CManager cm;
CMConnection conn;
void *data;
void *client_data;
{
    int i = 0;
    EChannel chan;
    RequestEventMsg *msg = (RequestEventMsg *) data;
    EControlContext ec = (EControlContext) client_data;

    if (channel_verbose > 0)
	printf("In request event handler\n");
    chan = get_local_channel_from_id(msg->global_channel_id, ec);

    if (chan == NULL) return;

    EChannel_lock(chan);

    for (i = 0; i < chan->local_source_list_len; i++) {
	if (chan->local_source_list[i].func != 
	    (ECRequestHandlerFunction) NULL) {
	    chan->local_source_list[i].func(chan->local_source_list[i].source_handle,
					    chan->local_source_list[i].client_data);
	}
    }

    EChannel_unlock(chan);
}

extern void
DEChannel_Event_handler(cm, conn, data, client_data)
CManager cm;
CMConnection conn;
void *data;
void *client_data;
{
    int i = 0;
    EChannel chan;
    EControlContext ec = (EControlContext) client_data;
    EventMsg *msg = (EventMsg *) data;
    free_block_rec_p block_rec = get_free_block_func(ec, NULL, 
						     (EventFreeFunction)NULL);
    block_rec->ec = ec;
    block_rec->free_arg = CMtake_buffer(ec->cm, data);

    if (channel_verbose > 0)
	printf("In event handler\n");
    chan =  get_local_channel_from_id(msg->global_channel_id, ec);

    if (chan == NULL) return;

    EChannel_lock(chan);
#ifdef MODULE
    if (chan->in_filter != NULL) {
      void (*func)(EChannel, attr_list, void *);
      func = chan->in_filter;
      func(chan, NULL, msg->event);
    }
#endif
    for (i = 0; i < chan->local_sink_list_len; i++) {
#ifdef MODULE
	if ((sink_valid(&chan->local_sink_list[i])) && 
		(chan->local_sink_list[i].block_event != 1)) {
#else
	if (sink_valid(&chan->local_sink_list[i])) {
#endif
	    queue_event_for_dispatch(chan, i, msg->event, msg->event_len,
				     block_rec, CMConnection_get_attrs(conn));
	}
    }
    deref_block(block_rec);
    EChannel_unlock(chan);
}

extern void
DEChannel_EventV_handler(cm, conn, data, client_data)
CManager cm;
CMConnection conn;
void *data;
void *client_data;
{
    EventVMsg *msg = (EventVMsg *) data;
    int i, event_len = 0;
    EChannel chan;
    EControlContext ec = (EControlContext) client_data;
    free_block_rec_p block_rec = 
	get_free_block_func(ec, NULL, (EventFreeFunction)NULL);
    block_rec->ec = ec;
    block_rec->free_arg = CMtake_buffer(ec->cm, data);

    if (channel_verbose > 0)
	printf("In eventv handler\n");
    for (i = 0; i < msg->eventv_len - 1; i++) {
	event_len += msg->eventv[i].iov_len;
    }
    event_len += msg->eventv[msg->eventv_len - 1].iov_len;
    chan = get_local_channel_from_id(msg->global_channel_id, ec);
    if (chan == NULL) return;
    EChannel_lock(chan);
    for (i = 0; i < chan->local_sink_list_len; i++) {
	if (sink_valid(&chan->local_sink_list[i])) {
	    queue_event_for_dispatch(chan, i, msg->eventv[0].iov_base,
				     event_len, block_rec,
				     CMConnection_get_attrs(conn));
	}
    }
    deref_block(block_rec);
    EChannel_unlock(chan);
}

extern void
DEChannel_Attend_handler(cm, conn, data, client_data)
CManager cm;
CMConnection conn;
void *data;
void *client_data;
{
    AttendMsgPtr msg = (AttendMsgPtr) data;
    AttendResponseMsg ret_msg;
    ExistsAttendResponseMsg exists_msg;
    EChannel chan, der_chan;
    int i;

#ifdef MODULE
    initiate_resubmit_buf(conn);
#endif

    ECtrace_out("DEChannel_Attend_handler, connection %lx", 1, conn);
    CMconn_register_close_handler(conn, EChannel_close_handler, (EControlContext) client_data);

    chan = string_to_channel_ptr(msg->chan_str);

    if (channel_verbose > 0) {
	printf("In attend handler, channel %s (%lx) cond %d, global id",
	       msg->chan_str, (long) chan, msg->cond);
	print_channel_id(chan->global_id.global_channel_id);
	printf("\n");
    }
    /* check to see that channel is valid */

    ECtrace_out("Checking for already existing remote channel chan %lx", 3, 
	      chan);
    for (i = 0; i < chan->rem_member_list_len; i++) {
	if (chan->rem_member_list[i].conn != NULL) {
	    if (chan->rem_member_list[i].conn == conn) {
		ECtrace_out("Found existing channel, notify attender", 3);
		exists_msg.condition = msg->cond;
		exists_msg.existing_channel_id =
		    chan->rem_member_list[i].remote_id;

		ECtrace_out("writing exists response name", 3);
		if (!CMwrite(conn, CMlookup_format(cm, channel_exists_attend_response_msg_flds), 
			     &exists_msg)) {
		    ECtrace_out("send channel exists Write failed", 0);
		}
		ECtrace_out("DEChannel_Attend_handler", 0);
		return;
	    }
	}
    }

    /* send him the sink list (add local) */
    ECtrace_out("send him sink list, local and remote", 3);
    memset(&ret_msg, 0, sizeof(ret_msg));
    ret_msg.condition = msg->cond;
    ret_msg.type_name = NULL;
    ret_msg.global_channel_id = chan->global_id.global_channel_id;
    if (chan->type_format) {
	ret_msg.type_name = global_name_of_IOformat(chan->type_format);
	ECtrace_out("type of format is %s", 3, ret_msg.type_name);
    }
    ret_msg.sink_count = 0;
    ret_msg.sinks = CMmalloc(sizeof(AttendRec));
    if (chan->local_sink_list_len != 0) {
	ret_msg.sinks[ret_msg.sink_count].contact_string = 
	    chan->global_id.contact_string;
	ret_msg.sinks[ret_msg.sink_count].filter = NULL;
#ifdef MODULE
	ret_msg.sinks[ret_msg.sink_count].conn_id = chan->conn_id;
#endif
	set_channel_id(&ret_msg.sinks[ret_msg.sink_count].client_channel_id,
		       chan);
	ret_msg.sink_count++;
    }



    for (i = 0; i < chan->rem_sink_list_len; i++) {
	if (chan->rem_sink_list[i].conn != NULL) {
	    ret_msg.sinks = CMrealloc(ret_msg.sinks,
			   (ret_msg.sink_count + 1) * sizeof(AttendRec));

	    ret_msg.sinks[ret_msg.sink_count].contact_string = 
		chan->rem_sink_list[i].contact_string;
	    ret_msg.sinks[ret_msg.sink_count].filter = NULL;
#ifdef MODULE
	    ret_msg.sinks[ret_msg.sink_count].conn_id = 
		chan->rem_sink_list[i].conn_id;
#endif
	    ret_msg.sinks[ret_msg.sink_count].client_channel_id = 
		chan->rem_sink_list[i].remote_id;
	    ret_msg.sink_count++;
	}
    }

    ECtrace_out("send him source list, local and remote", 3);

    ret_msg.source_count = 0;
    ret_msg.sources = CMmalloc(sizeof(AttendRec));
    if (chan->local_source_list_len != 0) {
	ret_msg.sources[ret_msg.source_count].contact_string = 
	    chan->global_id.contact_string;
	ret_msg.sources[ret_msg.source_count].filter = NULL;
#ifdef MODULE
	ret_msg.sources[ret_msg.source_count].conn_id = NULL;
#endif
	set_channel_id(&ret_msg.sources[ret_msg.source_count].client_channel_id,
		       chan);
	ret_msg.source_count++;
    }
    for (i = 0; i < chan->rem_source_list_len; i++) {
	if (chan->rem_source_list[i].conn != NULL) {
	    ret_msg.sources = CMrealloc(ret_msg.sources,
					(ret_msg.source_count + 1) * sizeof(AttendRec));

	    ret_msg.sources[ret_msg.source_count].contact_string = 
		chan->rem_source_list[i].contact_string;
	    ret_msg.sources[ret_msg.source_count].filter = NULL;
#ifdef MODULE
	    ret_msg.sources[ret_msg.source_count].conn_id = NULL;
#endif
	    ret_msg.sources[ret_msg.source_count].client_channel_id =
		chan->rem_source_list[i].remote_id;
	    ret_msg.source_count++;
	}
    }

    ECtrace_out("send him the remote list", 3);

    ret_msg.member_count = 1;
    ret_msg.members = CMmalloc(sizeof(AttendRec));
    ret_msg.members[0].contact_string = chan->global_id.contact_string;
    ret_msg.members[0].filter = NULL;
#ifdef MODULE
    ret_msg.members[0].conn_id = NULL;
#endif
    set_channel_id(&ret_msg.members[0].client_channel_id, chan);
    for (i = 0; i < chan->rem_member_list_len; i++) {
	if (chan->rem_member_list[i].conn != NULL) {
	    ret_msg.members = CMrealloc(ret_msg.members,
			 (ret_msg.member_count + 1) * sizeof(AttendRec));
	    ret_msg.members[ret_msg.member_count].contact_string =
		chan->rem_member_list[i].contact_string;
	    ret_msg.members[ret_msg.member_count].filter = NULL;
#ifdef MODULE
	    ret_msg.members[ret_msg.member_count].conn_id = NULL;
#endif
	    ret_msg.members[ret_msg.member_count].client_channel_id =
		chan->rem_member_list[i].remote_id;
	    ret_msg.member_count++;
	}
    }

    ECtrace_out("Sending list: \n Remote sink count %d Remote source count %d Remote member count %d",
		3, ret_msg.sink_count, ret_msg.source_count,
		ret_msg.member_count);
    
    ECtrace_out("sending derivitive channel IDs", 3);
    ret_msg.der_count = 0;
    for (i = 0; i < chan->der_list_len; i++) {
	der_chan = chan->der_chan_list[i].derived_chan;
	if (ret_msg.der_count == 0)
	    ret_msg.derivatives = CMmalloc(sizeof(AttendRec));
	else
	    ret_msg.derivatives = CMrealloc(ret_msg.derivatives,
			    (ret_msg.der_count + 1) * sizeof(AttendRec));
	ret_msg.derivatives[ret_msg.der_count].contact_string =
	    ECglobal_id(der_chan);
	ret_msg.derivatives[ret_msg.der_count].client_channel_id =
	    der_chan->global_id.global_channel_id;
	ret_msg.derivatives[ret_msg.der_count].filter = 
	    chan->der_chan_list[i].filter;
#ifdef MODULE
	ret_msg.derivatives[ret_msg.der_count].conn_id = NULL;
#endif
	ret_msg.derivatives[ret_msg.der_count].filter_has_output_param =
	    chan->der_chan_list[i].has_output_param;
	ret_msg.der_count++;
    }

    if (!CMwrite(conn, CMlookup_format(cm, channel_attend_response_msg_flds),
		&ret_msg)) {
	ECtrace_out("sent channel attend response write failed", 0);
    }
    for (i = 0; i < ret_msg.der_count; i++) {
	CMfree(ret_msg.derivatives[i].contact_string);
    }
    if (ret_msg.derivatives) CMfree(ret_msg.derivatives);
    if (ret_msg.sinks) CMfree(ret_msg.sinks);
    if (ret_msg.sources) CMfree(ret_msg.sources);
    if (ret_msg.members) CMfree(ret_msg.members);
    CMfree(ret_msg.type_name);
    ECtrace_out("adding new member to remote member list", 3);
    if (chan->rem_member_list == NULL) {
	chan->rem_member_list = CMmalloc(sizeof(struct _rem_attendr));
    }
    chan->rem_member_list =
	CMrealloc(chan->rem_member_list,
	  sizeof(struct _rem_attendr) * (chan->rem_member_list_len + 1));
    chan->rem_member_list[chan->rem_member_list_len].conn = conn;
    chan->rem_member_list[chan->rem_member_list_len].remote_id =
	copy_channel_id(msg->client_channel_id);

    chan->rem_member_list[chan->rem_member_list_len].contact_string = 
	strdup(msg->contact_str);
    chan->rem_member_list_len++;
    if (channel_verbose > 0) {
	printf("\n\nDEChannel_Attend_handler\n");
	dump_EChannel(chan);
    }
    ECtrace_out("DEChannel_Attend_handler", 0);
}

extern void
DEChannel_SinkSubscribe_handler(cm, conn, data, client_data)
CManager cm;
CMConnection conn;
void *data;
void *client_data;
{
    SinkSubscribeMsgPtr msg = (SinkSubscribeMsgPtr) data;
    EControlContext ec = (EControlContext) client_data;
    EChannel chan;

    if (channel_verbose) {
	printf("DEChannel_SinkSubscribe_handler, global chan_id = ");
	print_channel_id(msg->global_channel_id);
	printf("\n");
    }
    chan = get_local_channel_from_id(msg->global_channel_id, ec);

    if (chan != NULL) {
	CMconn_register_close_handler(conn, EChannel_close_handler, (EControlContext) client_data);
	ECtrace_out("add him to the sink list", 3);
	if (channel_verbose > 0)
	    printf("adding sink to chan %lx\n", (long) chan);

	chan->rem_sink_list =
	    CMrealloc(chan->rem_sink_list,
		      sizeof(struct _rem_attendr) * (chan->rem_sink_list_len + 1));
#ifdef MODULE
	initiate_resubmit_buf(conn);
#endif
	chan->rem_sink_list[chan->rem_sink_list_len].conn = conn;
	chan->rem_sink_list[chan->rem_sink_list_len].remote_id =
	    copy_channel_id(msg->client_channel_id);
	chan->rem_sink_list[chan->rem_sink_list_len].contact_string = 
	    strdup(msg->client_contact_str);
#ifdef MODULE
	chan->rem_sink_list[chan->rem_sink_list_len].block_event = 0;
        if (msg->conn_id != NULL) 
	    chan->rem_sink_list[chan->rem_sink_list_len].conn_id = 
	        strdup(msg->conn_id);
	else
	    chan->rem_sink_list[chan->rem_sink_list_len].conn_id = NULL;
	chan->rem_sink_list[chan->rem_sink_list_len].pd.tv_sec = 0;
	chan->rem_sink_list[chan->rem_sink_list_len].pd.tv_usec = 0;
#endif
printk("2Add sinks\n");
	chan->rem_sink_list_len++;
	if (chan->subscribe_func) {
	    chan->subscribe_func(1, chan->rem_sink_list_len,
				 chan->subscribe_client_data);
	}
	if (channel_verbose) {
	    printf("\n\nDEChannel_SinkSubscribe_handler\n");
	    dump_EChannel(chan);
	}
    }
    ECtrace_out("DEChannel_SinkSubscribe_handler", 0);
}

extern void
DEChannel_SinkUnsubscribe_handler(cm, conn, data, client_data)
CManager cm;
CMConnection conn;
void *data;
void *client_data;
{
    SinkSubscribeMsgPtr msg = (SinkSubscribeMsgPtr) data;
    EControlContext ec = (EControlContext) client_data;
    EChannel chan;
    int i;

    chan = get_local_channel_from_id(msg->global_channel_id, ec);
    ECtrace_out("DEChannel_SinkUnsubscribe_handler, chan %lx", 1, chan);

    if (chan != NULL) {
	ECtrace_out("remove him from the sink list", 3);
	if (channel_verbose > 0)
	    printf("removing sink from chan %lx\n", (long) chan);

	for (i=0; i< chan->rem_sink_list_len; i++) {
	    int local_len = chan->rem_sink_list[i].remote_id.len;
	    if ((local_len == msg->client_channel_id.len) &&
		(memcmp(chan->rem_sink_list[i].remote_id.channel,
			msg->client_channel_id.channel, local_len) == 0)) {
		chan->rem_sink_list[i].conn = NULL;
		if (i == (chan->rem_sink_list_len -1)) {
		    chan->rem_sink_list_len--;
		}
	    }
	}
	if (channel_verbose) {
	    printf("\n\nDEChannel_SinkUnsubscribe_handler\n");
	    dump_EChannel(chan);
	}
    }
    ECtrace_out("DEChannel_SinkUnsubscribe_handler", 0);
}

extern void
DEChannel_SourceSubscribe_handler(cm, conn, data, client_data)
CManager cm;
CMConnection conn;
void *data;
void *client_data;
{
    SourceSubscribeMsgPtr msg = (SourceSubscribeMsgPtr) data;
    EControlContext ec = (EControlContext) client_data;
    EChannel chan;

    chan = get_local_channel_from_id(msg->global_channel_id, ec);
    ECtrace_out("DEChannel_SourceSubscribe_handler, chan %lx", 1, chan);

    if (chan != NULL) {

	ECtrace_out("add source to chan %lx source list", 3, (long) chan);

	chan->rem_source_list =
	    CMrealloc(chan->rem_source_list,
		      sizeof(struct _rem_attendr) * (chan->rem_source_list_len + 1));
	chan->rem_source_list[chan->rem_source_list_len].conn = conn;
	chan->rem_source_list[chan->rem_source_list_len].remote_id =
	    copy_channel_id(msg->client_channel_id);
	chan->rem_source_list[chan->rem_source_list_len].contact_string = 
	    strdup(msg->client_contact_str);
	chan->rem_source_list_len++;

	if (channel_verbose) {
	    printf("\n\nDEChannel_SourceSubscribe_handler\n");
	    dump_EChannel(chan);
	}
    }
    ECtrace_out("DEChannel_SourceSubscribe_handler", 0);
}

extern void
DEChannel_SourceUnsubscribe_handler(cm, conn, data, client_data)
CManager cm;
CMConnection conn;
void *data;
void *client_data;
{
    SourceSubscribeMsgPtr msg = (SourceSubscribeMsgPtr) data;
    EControlContext ec = (EControlContext) client_data;
    EChannel chan;
    int i;

    chan = get_local_channel_from_id(msg->global_channel_id, ec);
    ECtrace_out("DEChannel_SourceUnsubscribe_handler, chan %lx", 1, chan);

    if (chan != NULL) {
	/* check to see that channel is valid */

	ECtrace_out("remove him from the source list", 3);
	if (channel_verbose > 0)
	    printf("removing source from chan %lx\n", (long) chan);

	for (i=0; i< chan->rem_source_list_len; i++) {
	    int local_len = chan->rem_source_list[i].remote_id.len;
	    if ((local_len == msg->client_channel_id.len) &&
		(memcmp(chan->rem_source_list[i].remote_id.channel,
			msg->client_channel_id.channel, local_len) == 0)) {
		chan->rem_source_list[i].conn = NULL;
		if (i == (chan->rem_source_list_len -1)) {
		    chan->rem_source_list_len--;
		}
	    }
	}
	if (channel_verbose) {
	    printf("\n\nDEChannel_SourceUnsubscribe_handler\n");
	    dump_EChannel(chan);
	}
    }
    ECtrace_out("DEChannel_SourceUnsubscribe_handler", 0);
}

extern void
DEChannel_MemberSubscribe_handler(cm, conn, data, client_data)
CManager cm;
CMConnection conn;
void *data;
void *client_data;
{
    MemberSubscribeMsgPtr msg = (MemberSubscribeMsgPtr) data;
    EControlContext ec = (EControlContext) client_data;
    EChannel chan;


    ECtrace_out("DEChannel_MemberSubscribe_handler", 1);
    CMconn_register_close_handler(conn, EChannel_close_handler, ec);
    /* check to see that channel is valid */
    chan = get_local_channel_from_id(msg->global_channel_id, ec);

    if (chan != NULL) {
	/* add him to the the member list */
	ECtrace_out("add him to the the member list", 3);
	chan->rem_member_list =
	    CMrealloc(chan->rem_member_list,
		      sizeof(struct _rem_attendr) * (chan->rem_member_list_len + 1));
#ifdef MODULE
printk("DEChannel_MemberSubscribe_handler\n");
	initiate_resubmit_buf(conn);
#endif
	chan->rem_member_list[chan->rem_member_list_len].conn = conn;
	chan->rem_member_list[chan->rem_member_list_len].contact_string =
	    strdup(msg->client_contact_str);
	chan->rem_member_list[chan->rem_member_list_len].remote_id =
	    copy_channel_id(msg->client_channel_id);
	chan->rem_member_list_len++;

	if (channel_verbose) {
	    printf("\n\nDEChannel_MemberSubscribe_handler\n");
	    dump_EChannel(chan);
	}
    }
    ECtrace_out("DEChannel_MemberSubscribe_handler", 0);
}

extern
void
EChannel_close_handler(cm, conn, client_data)
CManager cm;
CMConnection conn;
void *client_data;
{
    int i;

    EControlContext ec = (EControlContext) client_data;
    ECtrace_out("DEChannel_close_handler, ec = %lx, conn = %lx", 1, cm, conn);
    for (i = 0; i < ec->channel_count; i++) {
	EChannel chan = ec->channel_list[i];
	int j;
	int deleted = 0;
	ECtrace_out("removing him from remote member list", 3);
	for (j = 0; j < chan->rem_member_list_len; j++) {
	    if (chan->rem_member_list[j].conn == conn) {
		chan->rem_member_list[j].conn = NULL;
	    }
	}
	for (j = 0; (j + deleted) < chan->rem_member_list_len; j++) {
	    if (chan->rem_member_list[j].conn == NULL) {
		deleted++;
	    }
	    if ((deleted != 0) &&
		(j + deleted != chan->rem_member_list_len)) {
		chan->rem_member_list[j] =
		    chan->rem_member_list[j + deleted];
	    }
	}
	chan->rem_member_list_len -= deleted;

	deleted = 0;
	ECtrace_out("removing him from remote sink list", 3);
	for (j = 0; j < chan->rem_sink_list_len; j++) {
	    if (chan->rem_sink_list[j].conn == conn) {
		chan->rem_sink_list[j].conn = NULL;
	    }
	}
	for (j = 0; (j + deleted) < chan->rem_sink_list_len; j++) {
	    if (chan->rem_sink_list[j].conn == NULL) {
		deleted++;
	    }
	    if ((deleted != 0) &&
		(j + deleted != chan->rem_sink_list_len)) {
		chan->rem_sink_list[j] = chan->rem_sink_list[j + deleted];
	    }
	}
	chan->rem_sink_list_len -= deleted;
	ECtrace_out("Remote sinks deleted %d", 3, deleted);

	deleted = 0;
	ECtrace_out("removing him from remote sources list", 3);
	for (j = 0; j < chan->rem_source_list_len; j++) {
	    if (chan->rem_source_list[j].conn == conn) {
		chan->rem_source_list[j].conn = NULL;
	    }
	}
	for (j = 0; (j + deleted) < chan->rem_source_list_len; j++) {
	    if (chan->rem_source_list[j].conn == NULL) {
		deleted++;
	    }
	    if ((deleted != 0) &&
		(j + deleted != chan->rem_source_list_len)) {
		chan->rem_source_list[j] =
		    chan->rem_source_list[j + deleted];
	    }
	}
	chan->rem_source_list_len -= deleted;

	ECtrace_out("Remote sources deleted %d", 3, deleted);

    }
    ECtrace_out("DEChannel_close_handler", 0);
}


extern void
DEChannel_Exists_Attend_Response_handler(cm, conn, data, client_data)
CManager cm;
CMConnection conn;
void *data;
void *client_data;
{
    ExistsAttendResponseMsgPtr msg = (ExistsAttendResponseMsgPtr) data;
    EChannel chan, old_chan, *chan_ptr;
    EControlContext ec = (EControlContext) client_data;

    ECtrace_out("DEChannel_Exists_Attend_Response_handler", 1);
    if (channel_verbose > 0) {
	printf("In exists attend handler response, cond %d\n", msg->condition);
    }

    CMconn_register_close_handler(conn, EChannel_close_handler, ec);
    chan_ptr = CMCondition_get_client_data(cm, msg->condition);
    old_chan = *chan_ptr;

    ECtrace_out("DEChannel_exists_Attend_Response_handler, destroying old chan %lx", 
	      3, old_chan);
    EChannel_destroy(old_chan);

    /* 
     * change the channel variable in the waiting EChannel_open() to return
     * the already-existing channel instead.
     */
    chan = get_local_channel_from_id(msg->existing_channel_id, ec);
    *chan_ptr = chan;
    CMCondition_signal(cm, msg->condition);
    ECtrace_out("DEChannel_Exists_Attend_Response_handler", 0);
}


extern void
DEChannel_Attend_Response_handler(cm, conn, data, client_data)
CManager cm;
CMConnection conn;
void *data;
void *client_data;
{
    AttendResponseMsgPtr msg = (AttendResponseMsgPtr) data;
    EControlContext ec = (EControlContext) client_data;
    EChannel chan, der_chan, *chan_ptr;
    int i;

    ECtrace_out("DEChannel_Attend_Response_handler", 1);
    if (channel_verbose > 0) {
	printf("In attend handler response, cond %d\n", msg->condition);
    }
    /* preserve the message... */

    CMconn_register_close_handler(conn, EChannel_close_handler, ec);
    chan_ptr = CMCondition_get_client_data(cm, msg->condition);
    chan = *chan_ptr;

    if (chan->global_id.global_channel_id.channel) 
	CMfree(chan->global_id.global_channel_id.channel);
    chan->global_id.global_channel_id = 
	copy_channel_id(msg->global_channel_id);
    ECtrace_out("DEChannel_Attend_Response_handler, chan %lx, type %x", 3, chan,
	      msg->type_name);
    if (msg->type_name != NULL) {
	chan->type_format = 
	    CMget_IOformat_by_name(chan->cm, 
				   CMget_user_type_context(chan->cm),
				   msg->type_name);
    }

    ECtrace_out("Building lists: \n Remote sink count %d Remote source count %d Remote member count %d\n",
		3, msg->sink_count, msg->source_count, msg->member_count);
    
    /* build remote sink list */

    chan->rem_sink_list =
	CMrealloc(chan->rem_sink_list,
		  sizeof(struct _rem_attendr) * msg->sink_count);
    chan->rem_sink_list_len = 0;

    ECtrace_out("connect and register close handler for remote sinks", 3);
    for (i = 0; i < msg->sink_count; i++) {
	int j = chan->rem_sink_list_len;
	attr_list attrs = attr_list_from_string(msg->sinks[i].contact_string);
	if (CMcontact_self_check(ec->cm, attrs)) {
	    /* this channel is really local */
	    free_attr_list(attrs);
	    continue;
	}
	chan->rem_sink_list[j].conn = CMget_conn(cm, attrs);
#ifdef MODULE
printk("DEChannel_AttendResponse_handler\n");
	initiate_resubmit_buf(chan->rem_sink_list[j].conn);
#endif
	free_attr_list(attrs);
	if (chan->rem_sink_list[j].conn == NULL) {
	    printf("Failed to contact %s\n", msg->sinks[i].contact_string);
	    continue;
	}
	CMconn_register_close_handler(chan->rem_sink_list[j].conn, 
				      EChannel_close_handler, ec);
	chan->rem_sink_list[j].remote_id =
	    copy_channel_id(msg->sinks[i].client_channel_id);
	chan->rem_sink_list[j].contact_string = 
	    strdup(msg->sinks[i].contact_string);
#ifdef MODULE
	chan->rem_sink_list[j].block_event = 0;
   	if (msg->sinks[i].conn_id != NULL) 
	    chan->rem_sink_list[j].conn_id = strdup(msg->sinks[i].conn_id);
	else
	    chan->rem_sink_list[j].conn_id = NULL;
	chan->rem_sink_list[j].pd.tv_sec = 0;
	chan->rem_sink_list[j].pd.tv_usec = 0;
#endif
printk("1Add sinks\n");
	chan->rem_sink_list_len++;
    }

    /* build remote source list */

    chan->rem_source_list =
	CMrealloc(chan->rem_source_list,
		  sizeof(struct _rem_attendr) * msg->source_count);
    chan->rem_source_list_len = 0;

    ECtrace_out("connect and register close handler for remote sources", 3);
    for (i = 0; i < msg->source_count; i++) {
	int j = chan->rem_source_list_len;
	attr_list attrs = attr_list_from_string(msg->sources[i].contact_string);
	if (CMcontact_self_check(ec->cm, attrs)) {
	    free_attr_list(attrs);
	    /* this channel is really local */
	    continue;
	}
	chan->rem_source_list[j].conn = CMget_conn(cm, attrs);
	free_attr_list(attrs);
	if (chan->rem_source_list[j].conn == NULL) {
	    printf("Failed to contact %s\n", 
		   msg->sources[i].contact_string);
	    continue;
	}
	CMconn_register_close_handler(chan->rem_source_list[j].conn, 
				      EChannel_close_handler, ec);
	chan->rem_source_list[j].remote_id =
	    copy_channel_id(msg->sources[i].client_channel_id);
	chan->rem_source_list[j].contact_string = 
	    strdup(msg->sources[i].contact_string);
	chan->rem_source_list_len++;
    }

    /* build remote member list */

    chan->rem_member_list =
	CMrealloc(chan->rem_member_list,
		  sizeof(struct _rem_attendr) * msg->member_count);
    chan->rem_member_list_len = 0;
    for (i = 0; i < msg->member_count; i++) {
	int j = chan->rem_member_list_len;
	attr_list attrs = attr_list_from_string(msg->members[i].contact_string);
	if (CMcontact_self_check(ec->cm, attrs)) {
	    /* this channel is really local */
	    free_attr_list(attrs);
	    continue;
	}
	ECtrace_out("connect and register close handler for remote members", 3);
	chan->rem_member_list[j].conn =
	    CMget_conn(cm, attrs);
	free_attr_list(attrs);
	if (chan->rem_member_list[j].conn == NULL) {
	    printf("Failed to contact %s\n", 
		   msg->members[i].contact_string);
	    continue;
	}
	CMconn_register_close_handler(chan->rem_member_list[j].conn, 
				      EChannel_close_handler, ec);
	chan->rem_member_list[i].remote_id =
	    copy_channel_id(msg->members[i].client_channel_id);
	chan->rem_member_list[j].contact_string = 
	    strdup(msg->members[i].contact_string);
	if (i != 0) {
	    /* first entry is channel we got this from */
	    send_member_subscribe(chan, j);
	}
	chan->rem_member_list_len++;
    }
    for (i = 0; i < msg->der_count; i++) {
	char *filter;
	ECtrace_out("open, subscribe, and add derived channel to der_list", 3);
	der_chan = EChannel_open(ec, msg->derivatives[i].contact_string);

	chan->der_chan_list = CMrealloc(chan->der_chan_list,
			    (chan->der_list_len + 1) * sizeof(struct _derived_chan));
	chan->der_chan_list[chan->der_list_len].derived_chan = der_chan;
	filter = CMmalloc(strlen(msg->derivatives[i].filter) + 2);
	strcpy(filter, msg->derivatives[i].filter);
	filter[strlen(filter) + 1] = 0; /* add extra NULL termination */
	chan->der_chan_list[chan->der_list_len].filter = filter;
	chan->der_chan_list[chan->der_list_len].has_output_param =
	    msg->derivatives[i].filter_has_output_param;
	chan->der_list_len++;
    }

    if (channel_verbose > 0) {
	printf("\n\nDEChannel_Attend_Response_handler\n");
	dump_EChannel(chan);
    }
    ECtrace_out("DEChannel_Attend_Response_handler", 0);
    chan->ready++;
    CMCondition_signal(cm, msg->condition);
}

extern
void
EControlContext_event_dispatch(ec)
EControlContext ec;
{
    if (ec->event_queue_head != NULL) {
#ifdef MODULE
	CMevent_delivered(ec->cm);
#endif
	do_event_dispatch(ec);
    }
}

extern
void
EControlContext_event_flush(ec)
EControlContext ec;
{
    if (ec->event_queue_head != NULL) {
	do_event_flush(ec);
    }
}


CMFormatRec attend_response_format_list[] = {
    {"AttendRec", AttendRec_flds},
    {"channel_ID", channel_id_flds},
    {NULL, NULL}
};

CMFormatRec eventv_format_list[] = {
    {"IOEncodeElem", event_vec_elem_flds},
    {"channel_ID", channel_id_flds},
    {NULL, NULL}
};

CMFormatRec channel_id_format_list[] = {
    {"channel_ID", channel_id_flds},
    {NULL, NULL}
};

extern IOField field_list_flds[];
extern IOField format_list_flds[];
extern IOField channel_derive_msg_flds[];
extern IOField channel_remote_derive_msg_flds[];
extern IOField src_derive_msg_flds[];
extern IOField src_derive_resp_msg_flds[];
extern IOField proto_derive_msg_flds[];
extern IOField data_update_msg_flds[];
extern IOField channel_info_request_msg_flds[];
extern IOField channel_info_response_msg_flds[];

CMFormatRec channel_derive_format_list[] = {
    {"IOfield_list", field_list_flds},
    {"DEFormatList", format_list_flds},
    {"channel_ID", channel_id_flds},
    {NULL, NULL}
};

static
void
ECho_dispatch_poll_func(cm, ecv)
CManager cm;
void *ecv;
{
    EControlContext_event_dispatch((EControlContext)ecv);
}

extern EControlContext ECho_CM_init(cm)
CManager cm;
{
    EControlContext ec = malloc(sizeof(struct _EControlContext));
    CMFormat tmp_format;
    attr_list contact_list;

#ifdef MODULE
    if (current->qfabric == NULL)
      initialize_qfabric();
    ec->dispatch_func = NULL;
#endif 
    ec->event_queue_head = NULL;
    ec->event_queue_tail = NULL;
    ec->free_event_item_list = NULL;
    ec->current_event_item = NULL;
    ec->taken_events_list = NULL;

    ec->free_block_list = NULL;
    ec->channel_count = 0;
    ec->channel_list = NULL;
    ec->cm = cm;
#ifdef MODULE
    sema_init(&ec->lock, 1);
#else
    ec->lock = thr_mutex_alloc();
#endif

    if ((contact_list = CMget_contact_list(cm)) == NULL) {
	CMlisten(cm);
    } else {
	free_attr_list(contact_list);
    }
    CMadd_poll(cm, ECho_dispatch_poll_func, (void*) ec);

    tmp_format = CMregister_format(cm, "Channel Attend", 
				   channel_attend_msg_flds,
				   channel_id_format_list);
    CMregister_handler(tmp_format, DEChannel_Attend_handler, ec);

    tmp_format = CMregister_format(cm, "Sink Subscribe", 
				   sink_subscribe_msg_flds, 
				   channel_id_format_list);
    CMregister_handler(tmp_format, DEChannel_SinkSubscribe_handler, ec);

    tmp_format = CMregister_format(cm, "Sink Unsubscribe", 
				   sink_unsubscribe_msg_flds,
				   channel_id_format_list);
    CMregister_handler(tmp_format, DEChannel_SinkUnsubscribe_handler, ec);

    tmp_format = CMregister_format(cm, "Source Subscribe",
				   source_subscribe_msg_flds, 
				   channel_id_format_list);
    CMregister_handler(tmp_format, DEChannel_SourceSubscribe_handler, ec);

    tmp_format = CMregister_format(cm, "Source Unsubscribe", 
				   source_unsubscribe_msg_flds, 
				   channel_id_format_list);
    CMregister_handler(tmp_format, DEChannel_SourceUnsubscribe_handler, ec);

    tmp_format = CMregister_format(cm, "Request Event", 
				   request_event_msg_flds, 
				   channel_id_format_list);
    CMregister_handler(tmp_format, DEChannel_RequestEvent_handler, ec);

    tmp_format = CMregister_format(cm, "Member Subscribe", 
				   member_subscribe_msg_flds,
				   channel_id_format_list);
    CMregister_handler(tmp_format, DEChannel_MemberSubscribe_handler, ec);

    tmp_format = CMregister_format(cm, "Channel Attend Response", 
				   channel_attend_response_msg_flds, 
				   attend_response_format_list);
    CMregister_handler(tmp_format, DEChannel_Attend_Response_handler, ec);

    tmp_format = CMregister_format(cm, "Channel Exists Attend Response", 
				   channel_exists_attend_response_msg_flds, 
				   channel_id_format_list);
    CMregister_handler(tmp_format, DEChannel_Exists_Attend_Response_handler,
		       ec);

    tmp_format = CMregister_format(cm, "Event Message", event_msg_flds, 
				   channel_id_format_list);
    CMregister_handler(tmp_format, DEChannel_Event_handler, ec);

    tmp_format = CMregister_format(cm, "EventV Message", 
				   eventv_msg_flds, eventv_format_list);
    CMregister_handler(tmp_format, DEChannel_EventV_handler, ec);

    tmp_format = CMregister_format(cm, "Channel Derive", 
				   channel_derive_msg_flds,
				   channel_derive_format_list);
    CMregister_handler(tmp_format, Channel_Derive_handler, ec);

    tmp_format = CMregister_format(cm, "Channel Remote Derive", 
				   channel_remote_derive_msg_flds,
				   channel_derive_format_list);
    CMregister_handler(tmp_format, Channel_Remote_Derive_handler, ec);

    tmp_format = CMregister_format(cm, "Channel Source Derive", 
				   src_derive_msg_flds,
				   channel_id_format_list);
    CMregister_handler(tmp_format, Channel_Source_Derive_handler, ec);

    tmp_format = CMregister_format(cm, "Channel Source Derive Resp", 
				   src_derive_resp_msg_flds,
				   channel_id_format_list);
    CMregister_handler(tmp_format, Channel_Source_Derive_Resp_handler, ec);

    tmp_format = CMregister_format(cm, "Proto Derive", 
				   proto_derive_msg_flds, 
				   channel_derive_format_list);
    CMregister_handler(tmp_format, Proto_Derive_handler, ec);

    tmp_format = CMregister_format(cm, "Data Update Message", 
				   data_update_msg_flds, 
				   channel_id_format_list);
    CMregister_handler(tmp_format, Data_Update_handler, ec);

    tmp_format = CMregister_format(cm, "Channel Info Request Message", 
				   channel_info_request_msg_flds, 
				   NULL);
    CMregister_handler(tmp_format, Channel_Info_Request_handler, ec);

    tmp_format = CMregister_format(cm, "Channel Info Response Message", 
				   channel_info_response_msg_flds, 
				   NULL);
    CMregister_handler(tmp_format, Channel_Info_Response_handler, ec);

    return ec;
}


extern void 
ECdump_typed_event_as_XML (EChannel chan, void* event)
{
  dump_unencoded_IOrecord_as_XML ((IOFile)chan->subcontext, 
                                  chan->type_format,
                                  event);
}

extern void 
ECdump_event_as_XML (EChannel chan, void* event)
{
  dump_raw_IOrecord_as_XML ((IOFile)chan->subcontext, 
                            chan->type_format,
                            event);
}

extern
channel_ID_struct
copy_channel_id(channel_ID_struct chan_id)
{
    channel_ID_struct ret;
    if (chan_id.len == 0) {
	ret.len = 0;
	ret.channel = NULL;
    } else {
	ret.len = chan_id.len;
	ret.channel = malloc(ret.len);
	memcpy(ret.channel, chan_id.channel, ret.len);
    }
    return ret;
}

static int host_IP = 0;
static int pid = 0;

extern
void
set_channel_id(channel_ID_struct *chan_id, EChannel chan)
{
    /* 
     * change this if we ever allow more than one control context per process
     */
    chan_id->len = 4 /* IP addr */ + 4 /* pid */ + 0 /* echo context */ + 
	sizeof(chan) /* channel id */;
    if (host_IP == 0) {
	struct hostent *host_addr;
	char my_host_name[256];
	char my_domain_name[256];
	gethostname(my_host_name, sizeof(my_host_name));
	if (memchr((void *)my_host_name, '.', strlen(my_host_name)) == NULL) {
	    strcat(my_host_name, ".");
	    getdomainname(my_domain_name, sizeof(my_domain_name));
	    strcat(my_host_name, my_domain_name);
 	}
	host_addr = gethostbyname(my_host_name);
	if (sizeof(host_IP) != host_addr->h_length) {
	    printf("Host addr size mismatch in ECho channel.c\n");
	}
	memcpy(&host_IP, host_addr->h_addr, host_addr->h_length);
#ifdef MODULE
	DFreeMM((addrs_t)host_addr->h_name);
	DFreeMM((addrs_t)host_addr->h_addr_list);
	DFreeMM((addrs_t)host_addr);
#endif
	pid = (int) getpid();
    }
	
    chan_id->channel = malloc(chan_id->len);
    *((int*)(chan_id->channel + 0)) = host_IP;
    *((int*)(chan_id->channel + 4)) = pid;
    *((void**)(chan_id->channel + 8)) = (void*)chan;
}

extern EChannel
get_local_channel_from_id(channel_ID_struct chan_id, EControlContext cc)
{
    int id_host_IP;
    int id_pid;
#ifdef MODULE
    int host_IP_local;
    int pid_local;
    mm_segment_t fs;
    if (1) {
	struct hostent *host_addr;
	char my_host_name[256];
	fs = get_fs();
	set_fs(get_ds());
	sys_gethostname(my_host_name, sizeof(my_host_name));
	set_fs(fs);
	host_addr = lookup_name(my_host_name);
	if (sizeof(host_IP_local) != host_addr->h_length) {
	    printk("Host addr size mismatch in ECho channel.c\n");
	}
	memcpy(&host_IP_local, host_addr->h_addr, host_addr->h_length);
	DFreeMM((addrs_t)host_addr->h_name);
	DFreeMM((addrs_t)host_addr->h_addr_list);
	DFreeMM((addrs_t)host_addr);
	pid_local = current->pid;
    }
#else
    if (host_IP == 0) {
	struct hostent *host_addr;
	char my_host_name[256];
	gethostname(my_host_name, sizeof(my_host_name));
	host_addr = gethostbyname(my_host_name);
	if (sizeof(host_IP) != host_addr->h_length) {
	    printf("Host addr size mismatch in ECho channel.c\n");
	}
	memcpy(&host_IP, host_addr->h_addr, host_addr->h_length);
	pid = (int) getpid();
    }
#endif
	
    memcpy(&id_host_IP, chan_id.channel + 0, 4);
    memcpy(&id_pid, chan_id.channel + 4, 4);
#ifdef MODULE
    if ((id_host_IP == host_IP_local) && (id_pid == pid_local)) {
#else
    if ((id_host_IP == host_IP) && (id_pid == pid)) {
#endif
	/* channel ID is local */
	int i;
	EChannel chan;
	memcpy(&chan, chan_id.channel + 8, sizeof(chan));
	for(i=0; i< cc->channel_count; i++) {
	    if (chan == cc->channel_list[i]) {
		return chan;
	    }
	}
	printf("Channel %lx not found in channel list \n", (long)chan);
	return NULL;
    } else {
	/* channel ID is remote, search for local instance */
	int i;
	for(i=0; i< cc->channel_count; i++) {
	    channel_ID_struct this_id = 
		cc->channel_list[i]->global_id.global_channel_id;
	    if ((chan_id.len == this_id.len) &&
		(memcmp(chan_id.channel, this_id.channel, chan_id.len) ==0)){
		return cc->channel_list[i];
	    }
	}
	printf("Global channel not found in channel list \n");
    }
    return NULL;
}

extern void
print_channel_id(channel_ID_struct chan_id)
{
    struct in_addr id_host_IP;
    int id_pid;
    int i;

    if (chan_id.len == 0) {
	printf("NULL");
	return;
    }
    if (chan_id.len < 12) {
	printf("Invalid Length");
	return;
    }
    memcpy(&id_host_IP, chan_id.channel + 0, 4);
    memcpy(&id_pid, chan_id.channel + 4, 4);
    printf("host_IP = %s, pid = %d, chan = ", inet_ntoa(id_host_IP), id_pid);
    for (i=8 ; i < chan_id.len; i++) {
	printf("%02x", (unsigned char) chan_id.channel[i]);
    }
}

#ifdef MODULE
extern int 
ECregister_channel_filter(module_name, out_filter, in_filter)
char *module_name;
void *out_filter;
void *in_filter;
{
    int i = 0;

    while ((filters[i].module != NULL) && (i < MAX_FILTERS))
      i++;

    if (i == MAX_FILTERS) return -1;
   
    filters[i].module = strdup(module_name);
    filters[i].outfunc = out_filter;
    filters[i].infunc = in_filter;

    return i;
}

extern void 
ECunregister_channel_filter(id)
int id;
{
    kfree(filters[id].module);
    filters[id].module = NULL;  
    filters[id].outfunc = NULL;  
    filters[id].infunc = NULL;  
}

extern int
ECget_num_members(chan, type)
EChannel chan;
int type; /* 1=rem sink, 2=rem source, 3=local sink, 4=local source */
{
	if (type == 1) 
		return chan->rem_sink_list_len;
	if (type == 2)
		return chan->rem_source_list_len;
	if (type == 3)
		return chan->local_sink_list_len;
	if (type == 4)
		return chan->local_source_list_len;
	return -1;
}

extern void
ECset_link_attrs(chan, ptr, index, type)
EChannel chan;
char *ptr;
int index;
int type; /* 1=rem sink, 2=rem source, 3=local sink, 4=local source */
{
	if (type == 1) 
		chan->rem_sink_list[index].ptr = ptr;
	if (type == 2) 
		chan->rem_source_list[index].ptr = ptr;
	if (type == 3) 
		chan->local_sink_list[index].ptr = ptr;
	if (type == 4) 
		chan->local_source_list[index].ptr = ptr;
}

extern char *
ECget_link_attrs(chan, index, type)
EChannel chan;
int index;
int type; /* 1=rem sink, 2=rem source, 3=local sink, 4=local source */
{
	if (type == 1) 
		return chan->rem_sink_list[index].ptr;
	if (type == 2) 
		return chan->rem_source_list[index].ptr;
	if (type == 3) 
		return chan->local_sink_list[index].ptr;
	if (type == 4) 
		return chan->local_source_list[index].ptr;
	return NULL;
}

extern char *
ECget_link_id(chan, index, type)
EChannel chan;
int index;
int type; /* 1=rem sink, 2=rem source, 3=local sink, 4=local source */
{
	if (type == 1) 
		return chan->rem_sink_list[index].conn_id;
	if (type == 2) 
		return chan->rem_source_list[index].conn_id;
	if (type == 3) 
		return chan->local_sink_list[index].conn_id;
	if (type == 4) 
		return chan->local_source_list[index].conn_id;
	return NULL;
}

extern void
ECset_link_block(chan, index, type, block)
EChannel chan;
int index;
int type; /* 1=rem sink, 2=rem source, 3=local sink, 4=local source */
int block;
{
	if (type == 1) 
		chan->rem_sink_list[index].block_event = block;
	if (type == 3) 
		chan->local_sink_list[index].block_event = block;
}

#endif

