
#ifdef MODULE
//#define NO_DERIVED
#endif

typedef int (*FilterFunc) ARGS((void *, void *, void*));

typedef struct {
    int  len;
    char *channel;
} channel_ID_struct;

typedef struct _ECID {
    char *contact_string;
    channel_ID_struct global_channel_id;
} ECID;

typedef struct _rem_attendr {
    CMConnection conn;
    char *contact_string;
    channel_ID_struct remote_id;
#ifdef MODULE
    char *conn_id;
    int block_event;
    struct timeval pd;
    char *ptr;
#endif
} *rem_attendr_list;

typedef struct _multi_sink_rec *multi_sink_info;

typedef struct _local_sink {
    void *client_data;
    ECHandlerFunction func;
    EControlContext ec;
    IOContext context;
    multi_sink_info multi_sink_handle;
    IOFieldList field_list;
    CMFormatList format_list;
    IOFormat format;  
    int cond;
#ifdef MODULE
    char *conn_id;
    int block_event;
    char *ptr;
#endif
} *local_sink_ptr;

typedef struct _derived_source {
    FilterFunc filter;
    int has_output_param;
    int target_type_len;
#ifndef NO_DERIVED
    ecl_code code_struct;
#endif
    EChannel derived_chan;
    ECSourceHandle source_handle;
} *derived_source;

typedef struct _local_source {
    void *client_data;
    ECRequestHandlerFunction func;
    ECSourceHandle source_handle;
#ifdef MODULE
    char *conn_id;
    char *ptr;
#endif
} *local_source_ptr;

typedef struct _derived_chan {
    char *filter;
    int has_output_param;
    EChannel derived_chan;
} *derived_chan;

typedef struct _EChannel {
    int sink_count;
    int source_count;
    char *local_contact_string;
    ECID global_id;

    int rem_sink_list_len;
    rem_attendr_list rem_sink_list;

    int rem_source_list_len;
    rem_attendr_list rem_source_list;

    int rem_member_list_len;
    rem_attendr_list rem_member_list;
    int local_sink_list_len;
    local_sink_ptr local_sink_list;
    int local_source_list_len;
    local_source_ptr local_source_list;

    int der_list_len;
    derived_chan der_chan_list;

    int parent_list_len;
    EChannel *parent_chan_list;

    /* type info */
    IOContext subcontext;
    IOFormat type_format;
    void *filter_data;
    IOFieldList data_field_list;
    CMFormatList data_format_list;
    IOBuffer encode_buffer;

#ifdef MODULE
    struct semaphore lock;
#else
    thr_mutex_t lock;
#endif
    ECSubscribeFunction subscribe_func;
    void *subscribe_client_data;
    
    void *periodic_info;
    CManager cm;
    CMFormat eventv_format;
    EControlContext ec;
    int ready;
#ifdef MODULE
    void *out_filter;
    void *in_filter;
    char *conn_id;
#endif
} EChannel_struct;

typedef struct free_block_rec {
    int ref_count;
    EControlContext ec;
    void *free_arg;
    void *block;
    EventFreeFunction free_func;
    EControlContext locking_context;
    IOBuffer iobuffer;
    struct free_block_rec *next;
} *free_block_rec_p;

typedef struct _event_item {
    struct _event_item *next;
    struct _event_item *prev;
    EChannel chan;
    int sink_index;
    void *event;
    int event_len;
    ECEventQueueList event_queue_list;
    IOEncodeVector eventv;
    free_block_rec_p block_rec;
    attr_list attrs;
#ifdef MODULE
    ECRealTimeAttrs rt_attrs;
#endif
} event_item, *event_queue;

#ifdef MODULE
typedef int (*DispatchFunc) ARGS((void *, EControlContext));
#endif

struct _EControlContext {
    event_queue event_queue_head;
    event_queue event_queue_tail;
    event_queue free_event_item_list;
    event_queue current_event_item;
    event_queue taken_events_list;

    free_block_rec_p free_block_list;
    int channel_count;
    EChannel *channel_list;
    CManager cm;
#ifdef MODULE
    DispatchFunc dispatch_func;
    struct semaphore lock;
#else
    thr_mutex_t lock;
#endif
};
    
typedef struct _ECSinkHandle {
    EChannel chan;
    int sink_num;
} ECSinkStruct;

typedef struct _ECSourceHandle {
    EChannel chan;
    IOFormat ioformat;
    int source_num;
    int der_list_len;
    derived_source der_source_list;
#ifdef MODULE
    struct _ECRealTimeAttrs rt_attrs;    
#endif
} ECSourceStruct;

struct _ECEventRec {
    void *event;
    free_block_rec_p free_block;
};

struct _ECEventQueue {
    int event_count;
    int queue_size;
    struct _ECEventRec *queue;
    ECSinkHandle sink_handle;
};

struct _ECMultiEventQueue {
    int queue_count;
    struct _ECEventQueue *event_queues;
};

struct _multi_sink_rec {
    ECMultiHandlerFunction func;
    void *client_data;
    ECEventQueueList queue_list;
};

typedef struct _AttendMsg {
    char *chan_str;
    char *contact_str;
    int cond;
    channel_ID_struct client_channel_id;
} AttendMsg, *AttendMsgPtr;

typedef struct _SinkSubscribeMsg {
#ifdef MODULE
    char *conn_id;
#endif
    channel_ID_struct global_channel_id;
    channel_ID_struct client_channel_id;
    char *client_contact_str;
} SinkSubscribeMsg, *SinkSubscribeMsgPtr;

/* basically the same but still ... */
typedef struct _SourceSubscribeMsg {
    channel_ID_struct global_channel_id;
    channel_ID_struct client_channel_id;
    char *client_contact_str;
} SourceSubscribeMsg, *SourceSubscribeMsgPtr;

typedef struct _ExistsAttendResponseMsg {
    int condition;
    channel_ID_struct existing_channel_id;
} ExistsAttendResponseMsg, *ExistsAttendResponseMsgPtr;

typedef struct _AttendRec {
    char *contact_string;
    channel_ID_struct client_channel_id;
    char *filter;
    int filter_has_output_param;
#ifdef MODULE
    char *conn_id;
#endif
} AttendRec, *AttendRecPtr;

typedef struct _AttendResponseMsg {
    int condition;
    channel_ID_struct global_channel_id;
    char *type_name;
    int sink_count;
    AttendRec *sinks;
    int source_count;
    AttendRec *sources;
    int member_count;
    AttendRec *members;
    int der_count;
    AttendRec *derivatives;
} AttendResponseMsg, *AttendResponseMsgPtr;

typedef struct _MemberSubscribeMsg {
    channel_ID_struct global_channel_id;
    channel_ID_struct client_channel_id;
    char *client_contact_str;
} MemberSubscribeMsg, *MemberSubscribeMsgPtr;

typedef struct _RequestEventMsg {
    channel_ID_struct global_channel_id;
} RequestEventMsg, *RequestEventMsgPtr;

typedef struct _EventMsg {
    channel_ID_struct global_channel_id;
    int event_len;
    char *event;
} EventMsg, *EventMsgPtr;

typedef struct _EventVMsg {
    channel_ID_struct global_channel_id;
    int eventv_len;
    IOEncodeVector eventv;
} EventVMsg, *EventVMsgPtr;

typedef struct _FormatListElement {
    char *format_name;
    int field_list_len;
    IOFieldList field_list;
} format_list_element;

#ifdef MODULE
#define MAX_FILTERS 64

typedef struct _filter_struct {
    char *module;
    void *outfunc;
    void *infunc;
} filter_struct;

struct _filter_struct filters[MAX_FILTERS];

#endif

extern void ECtrace_out ARGS((char *format, int direction, ...));
extern void dump_EChannel ARGS((EChannel chan));
extern void DEChannel_close_handler ARGS((CManager cm, CMConnection conn));
extern EChannel string_to_channel_ptr ARGS((char *chan_id));
extern int channel_verbose;
extern int do_internal_source_derive ARGS((ECSourceHandle src,
					   EChannel der_chan, 
					   char *filter, 
					   IOFieldList input_field_list,
					   CMFormatList input_format_list,
					   IOFieldList output_field_list,
					   CMFormatList output_format_list,
					   IOFieldList data_field_list,
					   CMFormatList data_format_list));
extern void
remove_derived_channel ARGS((EChannel parent_chan, EChannel der_chan));

extern void DEperiodic_info_free ARGS((void *vperiodic));
extern void EChannel_register_subformats ARGS((CManager cm, 
					       IOContext context,
					       IOFieldList field_list,
					       CMFormatList format_list));
extern void EChannel_lock ARGS((EChannel chan));
extern void EChannel_unlock ARGS((EChannel chan));
extern void EChannel_close_handler ARGS((CManager cm, CMConnection conn, void *ec));
extern void 
Channel_Derive_handler ARGS((CManager cm, CMConnection conn,
			     void *data, void *client_data));
extern void 
Channel_Remote_Derive_handler ARGS((CManager cm, CMConnection conn,
				    void *data, void *client_data));
extern void 
Data_Update_handler ARGS((CManager cm, CMConnection conn,
			  void *data, void *client_data));
extern void 
Proto_Derive_handler ARGS((CManager cm, CMConnection conn,
			     void *data, void *client_data));
extern void 
Channel_Source_Derive_Resp_handler ARGS((CManager cm, CMConnection conn,
					 void *data, void *client_data));
extern void 
Channel_Source_Derive_handler ARGS((CManager cm, CMConnection conn,
				    void *data, void *client_data));
extern void 
Channel_Info_Request_handler ARGS((CManager cm, CMConnection conn,
				   void *data, void *client_data));
extern void 
Channel_Info_Response_handler ARGS((CManager cm, CMConnection conn,
				    void *data, void *client_data));

extern void
set_channel_id ARGS((channel_ID_struct * id_struct, EChannel chan));

extern channel_ID_struct
copy_channel_id ARGS((channel_ID_struct id_struct));

extern EChannel
get_local_channel_from_id ARGS((channel_ID_struct id_struct, 
				EControlContext cc));

extern void
print_channel_id ARGS((channel_ID_struct id_struct));

#define EChannel_lock(chan) IntEChannel_lock(chan, __FILE__, __LINE__)
#define EChannel_unlock(chan) IntEChannel_unlock(chan, __FILE__, __LINE__)
#define EContext_lock(context) IntEContext_lock(context, __FILE__, __LINE__)
#define EContext_unlock(context) IntEContext_unlock(context, __FILE__, __LINE__)
extern void IntEChannel_lock ARGS((EChannel chan, char *file, int line));
extern void IntEChannel_unlock ARGS((EChannel chan, char *file, int line));
extern void IntEContext_lock ARGS((EControlContext cc, char *file, int line));
extern void IntEContext_unlock ARGS((EControlContext cc, char *file, int line));
extern void CMsetup_conn_or_local ARGS((EControlContext ec, char *chan_id, 
					CMConnection *conn_ptr,
					void **local_ptr));

