
/*
 *   kecho.c
 *
 *   Copyright (C) 2000 Christian Poellabauer
 *
 */

/* 
 *   Kernel-loadable module for event channel support
 */

/* Required definitions. */
#ifndef MODULE
#define MODULE
#endif
#ifndef __KERNEL__
#define __KERNEL__
#endif

/* Various required header files. */
#include <linux/module.h> 
#include <linux/kernel.h> 
#include <linux/tty.h>
#include <linux/unistd.h>
#include <linux/socket.h>
#include <asm/current.h>
#include <asm/uaccess.h>
#include "kecho.h"
#include "../include/io.h"
#include "echo.h"
#include "../include/ecalls_module.h"
#ifndef NO_DERIVED
#include "../include/ecl.h"
#endif
#include "echo_internal.h"
#include "../include/cm.h"

MODULE_AUTHOR("Christian Poellabauer");
MODULE_DESCRIPTION("Kecho Module (KECho)");
MODULE_LICENSE("GPL");

/* Prototypes. */
int init_module(void);		       
void cleanup_module(void);
int (*original_syscall_echo)(); 
int (*original_fast_syscall_echo)(); 

/* Global variables. */
int channel_pool_count = 0;	//Sandip
EChannel channel_pool[MAX_CHANNEL];	//Sandip
struct proc_dir_entry *kecho_dir;	//Sandip
struct proc_dir_entry *kecho_entries[MAX_CHANNEL];	//Sandip
extern int kecho_identifier;

/* arch/i386/kernel/entry.S */
extern void *sys_call_table[];
extern void *fast_sys_call_table[];

int int_syscall_echo(data)
void *data;
{
  syscall_attrs *attrs;
  char *string;
  char name[MAX_STR];
  char conn_id[MAX_STR];
  char filter[MAX_STR];
  void *kevent[20]={NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};
  int index = 0;

  attrs = (syscall_attrs *)data;

  if (current->qfabric == NULL)
    initialize_qfabric();

  if (strcmp("ECho_CM_init", attrs->name) == 0) {
    attrs->ec = ECho_CM_init(attrs->cm);
    return 0;
  } 
  else if (strcmp("EChannel_create", attrs->name) == 0) {
    attrs->chan = EChannel_create(attrs->ec);
    return 0;
  }
  else if (strcmp("EChannel_create_id", attrs->name) == 0) {
    copy_from_user(conn_id, attrs->conn_id, strlen(attrs->conn_id)+1);
    attrs->chan = EChannel_create_id(attrs->ec, conn_id);
    return 0;
  }
  else if (strcmp("EChannel_create_attr", attrs->name) == 0) {
    attrs->chan = EChannel_create_attr(attrs->ec, attrs->attrs);
    return 0;
  }
  else if (strcmp("EChannel_create_attr_id", attrs->name) == 0) {
    copy_from_user(conn_id, attrs->conn_id, strlen(attrs->conn_id)+1);
    attrs->chan = EChannel_create_attr_id(attrs->ec, attrs->attrs, conn_id);
    return 0;
  }
  else if (strcmp("ECglobal_id", attrs->name) == 0) {
    string = ECglobal_id(attrs->chan);
    copy_to_user(attrs->chan_id, string, strlen(string)+1);
    attrs->ret_string = (char *)attrs->chan_id;
    return 0;
  }
  else if (strcmp("EChannel_open", attrs->name) == 0) {
    copy_from_user(name, attrs->global_name, strlen(attrs->global_name)+1);
    attrs->chan = EChannel_open(attrs->ec, name);
    return 0;
  }
  else if (strcmp("EChannel_open_id", attrs->name) == 0) {
    copy_from_user(name, attrs->global_name, strlen(attrs->global_name)+1);
    copy_from_user(conn_id, attrs->conn_id, strlen(attrs->conn_id)+1);
    attrs->chan = EChannel_open_id(attrs->ec, name, conn_id);
    return 0;
  }
  else if (strcmp("ECsource_subscribe", attrs->name) == 0) {
    attrs->source_handle = ECsource_subscribe(attrs->chan);
    return 0;
  }
  else if (strcmp("ECsource_subscribe_filter", attrs->name) == 0) {
    attrs->source_handle = ECsource_subscribe_filter(attrs->chan, 
	attrs->filter_function);
    return 0;
  }
  else if (strcmp("ECsubmit_event", attrs->name) == 0) {
    kevent[index] = kmalloc(attrs->event_len, GFP_KERNEL);
    copy_from_user(kevent[index], attrs->event, attrs->event_len);
    attrs->cm = ECsubmit_event(attrs->source_handle, kevent[index], attrs->event_len);
    index++; 
    if (index > 19) index = 0;
    if (index < 19) {
	if (kevent[index+1] != NULL) kfree(kevent[index+1]);
    }
    return 0;
  }
  else if (strcmp("ECset_rt_attributes", attrs->name) == 0) {
    ECset_rt_attributes(attrs->source_handle, &(attrs->rt_attrs));
    return 0;
  }
  else if (strcmp("ECsink_subscribe", attrs->name) == 0) {
    attrs->sink_handle = ECsink_subscribe(attrs->chan, attrs->func, 
	attrs->client_data);
    return 0;
  }
  else if (strcmp("ECsink_subscribe_filter", attrs->name) == 0) {
    attrs->sink_handle = ECsink_subscribe_filter(attrs->chan, attrs->func, 
	attrs->client_data, attrs->filter_function);
    return 0;
  }
  else if (strcmp("EChannel_typed_create", attrs->name) == 0) {
    attrs->chan = EChannel_typed_create(attrs->ec, attrs->field_list,
	attrs->format_list); 
    return 0;
  }
  else if (strcmp("ECsource_typed_subscribe", attrs->name) == 0) {
    attrs->source_handle = ECsource_typed_subscribe(attrs->chan,
	attrs->field_list, attrs->format_list);
    return 0;
  }
  else if (strcmp("ECsource_typed_subscribe_filter", attrs->name) == 0) {
    attrs->source_handle = ECsource_typed_subscribe_filter(attrs->chan,
	attrs->field_list, attrs->format_list, attrs->filter_function);
    return 0;
  }
  else if (strcmp("ECsubmit_typed_event", attrs->name) == 0) {
    attrs->cm = ECsubmit_typed_event(attrs->source_handle, attrs->event);
    return 0;
  }
  else if (strcmp("ECsink_typed_subscribe", attrs->name) == 0) {
    attrs->sink_handle = ECsink_typed_subscribe(attrs->chan,
      attrs->field_list, attrs->format_list, 
	(ECTypedHandlerFunction)attrs->func, attrs->client_data);
    return 0;
  }
  else if (strcmp("ECsink_typed_subscribe_filter", attrs->name) == 0) {
    attrs->sink_handle = ECsink_typed_subscribe_filter(attrs->chan,
      attrs->field_list, attrs->format_list, 
	(ECTypedHandlerFunction)attrs->func, attrs->client_data,
	attrs->filter_function);
    return 0;
  }
  else if (strcmp("EChannel_derive", attrs->name) == 0) {
    copy_from_user(name, attrs->chan_id, strlen(attrs->chan_id)+1);
    copy_from_user(filter, attrs->event_filter, strlen(attrs->event_filter)+1);
    attrs->chan = EChannel_derive(attrs->ec, name, filter);
    return 0;
  }
  else if (strcmp("EChannel_typed_derive", attrs->name) == 0) {
    copy_from_user(filter, attrs->event_filter, strlen(attrs->event_filter)+1);
    attrs->chan = EChannel_derive(attrs->ec, name, filter);
    return 0;
  }
  else if (strcmp("EChannel_data_open", attrs->name) == 0) {
    attrs->data_handle = EChannel_data_open(attrs->chan, attrs->field_list,
	attrs->format_list);
    return 0;
  }
  else if (strcmp("EChannel_data_update", attrs->name) == 0) {
    EChannel_data_update(attrs->data_handle, attrs->data);
    return 0;
  }
  else if (strcmp("ECproto_derive_periodic", attrs->name) == 0) {
    /*
    copy_from_user(name, attrs->proto_id, strlen(attrs->proto_id)+1);
    copy_from_user(filter, attrs->filter_function, 
	strlen(attrs->filter_function)+1);
    attrs->chan = ECproto_derive_periodic(attrs->ec, name, filter,
	attrs->field_list, attrs->format_list, attrs->period);
    */
    return 0;
  }
  else
    return -1;
}


/* gethostname function. */
int sys_gethostname(char *name, int len)
{
  int i, errno;

  if (len < 0)
    return -EINVAL;
  down_read(&uts_sem);
  i = 1 + strlen(system_utsname.nodename);
  if (i > len)
    i = len;
  errno = 0;
  memcpy(name, system_utsname.nodename, i);
  up_read(&uts_sem);
  return errno;
}

/* getdomainname function. */
int sys_getdomainname(char *name, int len)
{
  int nlen;
  int err = -EFAULT;

  down_read(&uts_sem);

  nlen =  strlen(system_utsname.domainname) + 1;

  if (nlen < len)
    len = nlen;
  if (len > __NEW_UTS_LEN)
    goto done;
  memcpy(name, system_utsname.domainname, len);
  err = 0;
done:
  up_read(&uts_sem);
  return err;
}

int stringtoint(char *s) {
	int i,n=0,len=strlen(s);
	int j=1;
	for (i=0;i<len;i++) {
		n = n*j + s[len-i-1] - '0';
		j=j*10;
	}
	return n;
}

void inttostring(char *string, int n){
	char s[10],t[10];
	int i=0,j=0;
	while(n!=0){
		s[i++] = n%10 + '0';
		n /= 10;
	}
	i--;
	while(j!=i+1) {
		t[j]=s[i-j];
		j++;
	}
	t[j] = '\0';
	if (i<0)
		strcpy(string, "0");
	else
		strcpy(string, t);
}

/* The following function was added by sandip*/
/* This function is invoked when we try to read the /proc/kproc */
static int kproc_read(char *page, char **start, off_t off, int count, int *eof, void *data) {
	int len=0;
	int i=-1;
	static char my_buffer[1080];
	EChannel chan=NULL;
	if(off>0)
		return 0;
	MOD_INC_USE_COUNT;
	len = sprintf(my_buffer,"Channel Details: Total No. of Channels= %d\n",channel_pool_count);
	
	if(!strcmp(data,"9999")){ 
		for(i=0;i<channel_pool_count;i++){
			chan = channel_pool[i];
			len += sprintf(my_buffer+len,"\nChannel No. %d\n",i+1);
			len += sprintf(my_buffer+len,"Contact_String= %s\nChannel=%s\n",chan->global_id.contact_string,ECglobal_id(chan));
			len += sprintf(my_buffer+len,"Sink_Count= %d\nSource_Count=%d\nLocal_Contact_String= %s\n",chan->sink_count,chan->source_count,chan->local_contact_string);
		}
	}else{
		i = stringtoint(data);
		chan = channel_pool[i];
		len += sprintf(my_buffer+len,"\nChannel No. %d\n",i+1);
		len += sprintf(my_buffer+len,"Contact_String= %s\nChannel=%s\n",chan->global_id.contact_string,ECglobal_id(chan));
		len += sprintf(my_buffer+len,"Sink_Count= %d\nSource_Count=%d\nLocal_Contact_String= %s\n",chan->sink_count,chan->source_count,chan->local_contact_string);

	}

        *start = my_buffer;
	MOD_DEC_USE_COUNT;
	return len;

}

/* The following function was added by sandip*/

extern 
void 
add_to_channel_pool(chan)
EChannel chan;
{
	struct proc_dir_entry *t;
	char channelname[500],*temp,*data1;
	int i=0;
	data1 = (char *)kmalloc(10*sizeof(char),GFP_KERNEL);
	inttostring(data1, channel_pool_count);
	strcpy(channelname , ECglobal_id(chan));
	while(channelname[i]!='\0')
		if(channelname[i++]=='@')
			break;
	temp = channelname + i;
	t = create_proc_read_entry(temp, S_IFREG | S_IRUGO,kecho_dir, kproc_read,data1);
	t->owner=THIS_MODULE;
	//printk("<1>Creating /proc/kecho/%s\n",temp);
	kecho_entries[channel_pool_count] = t;
	channel_pool[channel_pool_count++] = chan;
}

/* 'FAST' sys call. */
int f_kecho_api(int id)
{


  return 0;
}

/* Sys call. */
void s_kecho_api(void)
{

}

/* Initialize module. */
int init_module(void) 
{
  struct proc_dir_entry *t;
  kecho_dir = proc_mkdir("kecho", NULL);
  kecho_dir->owner=THIS_MODULE;
  
  t = create_proc_read_entry("channel_details", S_IFREG | S_IRUGO, kecho_dir, 
	kproc_read, "9999");
  t->owner = THIS_MODULE;

  original_syscall_echo = sys_call_table[__NR_fast_KECho_syscall_echo];
  original_fast_syscall_echo=fast_sys_call_table[__NR_fast_KECho_syscall_echo];

  sys_call_table[__NR_KECho_syscall_echo] = int_syscall_echo;
  fast_sys_call_table[__NR_fast_KECho_syscall_echo] = int_syscall_echo;

  /* ECalls Setup. */
  //register_service("KECho_API", f_kecho_api, s_kecho_api, IMMEDIATE);
  kecho_identifier = register_service("KECho_API", NULL, NULL, IMMEDIATE);

  return 0;
}

/* Remove module. */
void cleanup_module(void) 
{
  int i,j=0;
  char channelname[500], temp[20];

  /* ECalls teardown. */
  unregister_service("KECho_API");
  
  /* /proc teardown. */
  for (i=0; i<channel_pool_count; i++) {
	strcpy(channelname, ECglobal_id(channel_pool[i]));
	j = 0;
	while(channelname[j]!='\0')
		if (channelname[j++]=='@')
			break;
	strcpy(temp,channelname + j);
	//printk("removing /proc/kecho/%s\n", temp);
	kfree(kecho_entries[i]->data);
	remove_proc_entry(temp, kecho_dir);
  }
  remove_proc_entry("channel_details", kecho_dir);
  //printk("removing /proc/kecho/channel_details\n");
  remove_proc_entry("kecho", NULL);
  //printk("removing /proc/kecho\n");
  
  sys_call_table[__NR_KECho_syscall_echo] = original_syscall_echo;
  fast_sys_call_table[__NR_fast_KECho_syscall_echo]=original_fast_syscall_echo;
}

