

#include "config.h"
#ifdef HAVE_WINDOWS_H
#include <windows.h>
#else
#include <sys/time.h>
#endif
#include "io.h"
#include "atl.h"
#include "cm.h"
#include "echo.h"
#include "gen_thread.h"

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/types.h>
#include <signal.h>
#ifdef HAVE_WINDOWS_H
#include <windows.h>
#include <process.h>
#define srand48(s) srand(s)
#define drand48() (((double)rand())/((double)RAND_MAX))
#define lrand48() rand()
#define kill(x,y) 
#else
extern double drand48();
extern long lrand48();
void srand48 ARGS((long seedval));
#include <sys/wait.h>
#endif

static int quiet = 0;

static int do_regression_master_test();

typedef struct _complex_rec {
    double r;
    double i;
} complex, *complex_ptr;

static IOField complex_field_list[] =
{
    {"r", "double", sizeof(double), IOOffset(complex_ptr, r)},
    {"i", "double", sizeof(double), IOOffset(complex_ptr, i)},
    {NULL, NULL, 0, 0}
};

typedef struct _simple_rec {
    int integer_field;
    short short_field;
    long long_field;
    char *string;
    complex complex_field;
    double double_field;
    char char_field;
    long sum;
} simple_rec, *simple_rec_ptr;

static IOField simple_field_list[] =
{
    {"integer_field", "integer",
     sizeof(int), IOOffset(simple_rec_ptr, integer_field)},
    {"short_field", "integer",
     sizeof(short), IOOffset(simple_rec_ptr, short_field)},
    {"long_field", "integer",
     sizeof(long), IOOffset(simple_rec_ptr, long_field)},
    {"string_field", "string",
     sizeof(char *), IOOffset(simple_rec_ptr, string)},
    {"complex_field", "complex",
     sizeof(complex), IOOffset(simple_rec_ptr, complex_field)},
    {"double_field", "float",
     sizeof(double), IOOffset(simple_rec_ptr, double_field)},
    {"char_field", "char",
     sizeof(char), IOOffset(simple_rec_ptr, char_field)},
    {"sum", "integer",
     sizeof(long), IOOffset(simple_rec_ptr, sum)},
    {NULL, NULL, 0, 0}
};

static CMFormatRec simple_format_list[] =
{
    {"complex", complex_field_list},
    {NULL, NULL},
};

#define REPEAT_COUNT 10

static EControlContext cc = NULL;
static simple_rec_ptr stored_event = NULL;

static
void
typed_handler(vevent, client_data, attrs)
void *vevent;
void *client_data;
attr_list attrs;
{
    simple_rec_ptr event = vevent;
    long sum = 0, scan_sum = 0;

    if (stored_event == NULL) {
	if (quiet == 0) {
	    printf("Taking event buffer\n");
	}
	if (ECtake_event_buffer(cc, vevent) == 0) {
	    printf("Take event failed\n");
	}
	stored_event = event;
	return;
    }

    sum += event->integer_field % 100;
    sum += event->short_field % 100;
    sum += event->long_field % 100;
    sum += ((int)(event->complex_field.r * 100.0)) % 100;
    sum += ((int)(event->complex_field.i * 100.0)) % 100;
    sum += ((int)(event->double_field * 100.0)) % 100;
    sum += event->char_field;

    sum += stored_event->integer_field % 100;
    sum += stored_event->short_field % 100;
    sum += stored_event->long_field % 100;
    sum += ((int)(stored_event->complex_field.r * 100.0)) % 100;
    sum += ((int)(stored_event->complex_field.i * 100.0)) % 100;
    sum += ((int)(stored_event->double_field * 100.0)) % 100;
    sum += stored_event->char_field;

    sum = sum % 100;
    if (event->string) 
	sscanf(event->string, "testing, sum is %ld", &scan_sum);
    if (sum != scan_sum) {
	printf("Received record checksum does not match, expected %ld, got %ld.\n",
	       scan_sum, sum);
    }
    if ((quiet == 0) || (sum != scan_sum)) {
	printf("In %d the handler, event data is :\n", quiet);
	printf("	integer_field = %d\n", event->integer_field);
	printf("	short_field = %d\n", event->short_field);
	printf("	long_field = %ld\n", event->long_field);
	printf("	string = %s\n", event->string ? event->string : "NULL");
	printf("	complex_field.r = %g\n", event->complex_field.r);
	printf("	complex_field.i = %g\n", event->complex_field.i);
	printf("	double_field = %g\n", event->double_field);
	printf("	char_field = %c\n", event->char_field);

	printf("In the handler, stored event data is :\n");
	printf("	integer_field = %d\n", stored_event->integer_field);
	printf("	short_field = %d\n", stored_event->short_field);
	printf("	long_field = %ld\n", stored_event->long_field);
	printf("	string = %s\n", stored_event->string ? stored_event->string : "NULL");
	printf("	complex_field.r = %g\n", stored_event->complex_field.r);
	printf("	complex_field.i = %g\n", stored_event->complex_field.i);
	printf("	double_field = %g\n", stored_event->double_field);
	printf("	char_field = %c\n", stored_event->char_field);

    }
    sum = 0;
    if (client_data != NULL) {
	int tmp = *((int*)client_data);
	if (tmp > 0) {
	    *((int*)client_data) = tmp - 1;
	}
    }
    ECreturn_event_buffer(cc, stored_event);
    stored_event = NULL;
}

static
void free_record(event)
simple_rec_ptr event;
{
    free(event->string);
    free(event);
}

static
void generate_record(event)
simple_rec_ptr event;
{
    char *tmp_string = malloc(20);
    long sum = 0;
    event->integer_field = (int) lrand48() % 100;
    sum += event->integer_field;
    event->short_field = (short) lrand48();
    sum += event->short_field % 100;
    event->long_field = (long) lrand48();
    sum += event->long_field % 100;

    event->complex_field.r = drand48();
    sum += ((int)(event->complex_field.r * 100.0)) % 100;
    event->complex_field.i = drand48();
    sum += ((int)(event->complex_field.i * 100.0)) % 100;

    event->double_field = drand48();
    sum += ((int)(event->double_field * 100.0)) % 100;
    event->char_field = lrand48() % 128;
    sum += event->char_field;
    sum = sum % 100;
    sprintf(tmp_string, "testing, sum is %ld\n", sum);
    event->string = tmp_string;
    event->sum = sum;
}

static
void generate_two_records(event1, event2)
simple_rec_ptr event1;
simple_rec_ptr event2;
{
    long sum;
    generate_record(event1);
    generate_record(event2);
    sum = event1->sum + event2->sum;
    event1->sum = 0;
    event2->sum = sum % 100;
    sprintf(event2->string, "testing, sum is %ld\n", sum % 100);
}

static int regression = 0;

static void
do_source_side(cm, chan, done_flag, count)
CManager cm;
EChannel chan;
int *done_flag;
int count;
{
    ECSourceHandle handle;
    int ret = 0;

    if (quiet == 0) printf("chan is %lx\n", (long)chan);
    handle = ECsource_typed_subscribe(chan, simple_field_list, 
					  simple_format_list);
    while (!ret) {
	simple_rec_ptr a, b;
	a = malloc(sizeof(*a));
	b = malloc(sizeof(*b));
	generate_two_records(a, b);
	if (quiet == 0)
	    printf("I'm submitting %ld\n", a->long_field);
	ECsubmit_general_typed_event(handle, a, 
				     (EventFreeFunction)free_record);
	ECsubmit_general_typed_event(handle, b,
				     (EventFreeFunction)free_record);
	if (count != -1) {
	    count--;
	}
	CMusleep(cm, 10000);
	if (done_flag && *done_flag) ret++;
	if (count != -1) {
	    if (count == 0) ret++;
	}	    
    }
    ECcancel_source_subscribe(handle);
}

static pid_t sink_proc = 0;

static void
fail_and_die(signal)
int signal;
{
    fprintf(stderr, "Channel test failed to complete in reasonable time\n");
    if (sink_proc != 0) {
	kill(sink_proc, 9);
    }
    exit(1);
}

int
main(argc, argv)
int argc;
char **argv;
{
    EChannel chan;
    double r = drand48();
    int consumer = 0;
    int both = 0;
    int regression_master = 1;
    CManager cm;

    srand48(getpid());
    regression = 1;
    quiet = 1;
    while (argv[1] && (argv[1][0] == '-')) {
	if (argv[1][1] == 'c') {
	    r = .9;
	    regression_master = 0;
	} else if (argv[1][1] == 's') {
	    r = 0;
	    regression_master = 0;
	} else if (argv[1][1] == 'b') {
	    both++;
	    regression_master = 0;
	} else if (argv[1][1] == 'q') {
	    quiet++;
	} else if (argv[1][1] == 'v') {
	    quiet--;
	} else if (argv[1][1] == 'n') {
	    regression = 0;
	    quiet = 0;
	}
	argv++;
    }
    if (regression && regression_master) {
	return do_regression_master_test();
    }
    cm = CManager_create();
    cc = ECho_CM_init(cm);
    consumer = (r > 0.5);
    if (quiet <= 0) printf("r is %g\n", r);
    chan = EChannel_open(cc, argv[1]);
    if (chan == NULL) {
	printf("Failed to contact channel %s\n", argv[1]);
	exit(1);
    }
    if (both) {
	EChannel chan2;
	chan = EChannel_open(cc, argv[1]);
	chan2 = EChannel_open(cc, argv[1]);
	(void) ECsink_typed_subscribe(chan, simple_field_list, 
				      simple_format_list,
				      typed_handler, NULL);
	if (quiet <= 0) printf("chan is %lx, chan2 is %lx\n", (long)chan, (long)chan2);
	/* do source side forever */
	do_source_side(cm, chan, NULL, -1);
    } else if (consumer) {
	int repeat_count = REPEAT_COUNT;
	if (regression) {
#ifdef HAVE_WINDOWS_H
	  SetTimer(NULL, 5, 1000, (TIMERPROC) fail_and_die);
#else
	  struct sigaction sigact;
	  sigact.sa_flags = 0;
	  sigact.sa_handler = fail_and_die;
	  sigemptyset(&sigact.sa_mask);
	  sigaddset(&sigact.sa_mask, SIGALRM);
	  sigaction(SIGALRM, &sigact, NULL);
	  alarm(300);
#endif
	}
	if (quiet <= 0) printf("I'm a consumer\n");
	(void) ECsink_typed_subscribe(chan, simple_field_list,
				      simple_format_list,
				      typed_handler, 
				      regression ? &repeat_count : NULL);
	while (repeat_count != 0) {
	    CMsleep(cm, 1);
	}
    } else {
	if (quiet <= 0) printf("I'm a supplier\n");
	do_source_side(cm, chan, NULL, -1);
    }
    return 0;
}

static
pid_t
run_subprocess(args, channel_id)
char **args;
{
#ifdef HAVE_WINDOWS_H
    int child;
    child = _spawnv(_P_NOWAIT, "./take_test.exe", args);
    if (child == -1) {
	perror("spawnv");
    }
    return child;
#else
    pid_t child = fork();
    if (child == 0) {
	/* I'm the child */
	execv("./take_test", args);
    }
    return child;
#endif
}

static int
do_regression_master_test()
{
    CManager cm;
    EChannel chan;

#ifdef HAVE_WINDOWS_H
    SetTimer(NULL, 5, 1000, (TIMERPROC) fail_and_die);
#else
    struct sigaction sigact;
    sigact.sa_flags = 0;
    sigact.sa_handler = fail_and_die;
    sigemptyset(&sigact.sa_mask);
    sigaddset(&sigact.sa_mask, SIGALRM);
    sigaction(SIGALRM, &sigact, NULL);
    alarm(300);
#endif

#ifdef USE_PTHREADS
    gen_pthread_init();
#endif

    cm = CManager_create();
    (void) CMfork_comm_thread(cm);
    cc = ECho_CM_init(cm);

    srand48(1);

    chan = EChannel_typed_create(cc, simple_field_list, simple_format_list);
    
    if (quiet <= 0) printf("  channel ID is %s\n", ECglobal_id(chan));

    /* local tests */
    {
	int repeat_count = REPEAT_COUNT;
	int repeat_count1 = REPEAT_COUNT / 2;
	int repeat_count2 = REPEAT_COUNT - repeat_count1;
	ECSinkHandle handle;

	handle = ECsink_typed_subscribe(chan, simple_field_list,
					simple_format_list, typed_handler,
					&repeat_count);
	do_source_side(cm, chan, NULL, repeat_count1);
	do_source_side(cm, chan, NULL, repeat_count2);
	if (repeat_count != 0) {
	    fprintf(stderr, "Not all in-process events delivered, test 1\n");
	} else {
	    if (quiet < 2) 
		fprintf(stderr, "Passed in-process delivery tests\n");
	}	
	ECcancel_sink_subscribe(handle);
    }

    /* multiple process tests */
    {
	int repeat_count1 = REPEAT_COUNT / 2;
	int repeat_count2 = REPEAT_COUNT - repeat_count1;
	char *args[] = {"typed_test", "-c", NULL, NULL};
	int exit_state;

	args[sizeof(args)/sizeof(char*) -2] = ECglobal_id(chan);
	sink_proc = run_subprocess(args);
	free(args[sizeof(args)/sizeof(char*) -2]);
	/* give him time to start */
	CMsleep(cm, 30);
	do_source_side(cm, chan, NULL, repeat_count1);
	do_source_side(cm, chan, NULL, repeat_count2);
#ifdef HAVE_WINDOWS_H
	if (_cwait(&exit_state, sink_proc, 0) == -1) {
	    perror("cwait");
	}
	if (exit_state == 0) {
	    if (quiet < 2) 
		printf("Passed single remote sink test\n");
	} else {
	    printf("Single remote sink exit with status %d\n",
		   exit_state);
	}
#else
	if (waitpid(sink_proc, &exit_state, 0) == -1) {
	    perror("waitpid");
	}
	if (WIFEXITED(exit_state)) {
	    if (WEXITSTATUS(exit_state) == 0) {
		if (quiet < 2) 
		    printf("Passed single remote sink test\n");
	    } else {
		printf("Single remote sink exit with status %d\n",
		       WEXITSTATUS(exit_state));
	    }
	} else if (WIFSIGNALED(exit_state)) {
	    printf("Single remote sink died with signal %d\n",
		   WTERMSIG(exit_state));
	}
#endif
    }
    EChannel_destroy(chan);
/*    DExchange_close(de);
      DExchange_free(de);*/
    return 0;
}
